/*
**    Genray: A Generic Raytracer
**    Copyright (C) 1994  Stephen Chenney (stephen@cs.su.oz.au)
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program; if not, write to the Free Software
**    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/****************************************************************/
/*								*/
/*	Header file for the vector macros.			*/
/*	This file provides macro definitions for most of the	*/
/*	 operations defined in vectorfunc.h.			*/
/*	Include one or the other.				*/
/*	The two DO NOT have the same argument types, and there	*/
/*	 is LIMITED TYPE CHECKING.  The notable change is	*/
/*	 that things which formerly took a pointer to a Vector	*/
/*	 now take the vector itself.				*/
/*	 eg VNew(1,2,3,&v) becomes VNew(1,2,3,v).		*/
/*	Written by Stephen Chenney.  Dec 1993.			*/
/*								*/
/****************************************************************/

#ifndef __VECTOR__
#define __VECTOR__

/* Define a vector.  */
typedef struct Vector
	{
		double  x;
		double  y;
		double  z;
	} Vector;

/* A Matrix is an array of vectors. */
typedef Vector Matrix[3];


/* Takes the modulus of v */
#define VMod(v)		(sqrt((v).x*(v).x + (v).y*(v).y + (v).z*(v).z))

/* Returns the dot product of v1 & v2 */
#define VDot(v1, v2)	((v1).x*(v2).x + (v1).y*(v2).y + (v1).z*(v2).z)

/* Fills the fields of a vector.	*/
#define VNew(a, b, c, r)	((r).x = (a), (r).y = (b), (r).z = (c))

#define VAdd(v1, v2, r)		((r).x = (v1).x + (v2).x , \
							 (r).y = (v1).y + (v2).y , \
							 (r).z = (v1).z + (v2).z )

#define VSub(v1, v2, r)		((r).x = (v1).x - (v2).x , \
							 (r).y = (v1).y - (v2).y , \
							 (r).z = (v1).z - (v2).z )

#define VScalarMul(v, d, r)	((r).x = (v).x * (d) , \
				 			 (r).y = (v).y * (d) , \
				 			 (r).z = (v).z * (d) )

#define VCross(v1, v2, r)	((r).x = (v1).y * (v2).z - (v1).z * (v2).y , \
				 			 (r).y = (v1).z * (v2).x - (v1).x * (v2).z , \
				 			 (r).z = (v1).x * (v2).y - (v1).y * (v2).x )

#define MNew(a, b, c, r)	((r)[0] = (a), (r)[1] = (b) , (r)[2] = (c) )

#define MScalarMul(m, d, r)	(VScalarMul((m)[0], (d), (r)[0]) , \
				 			 VScalarMul((m)[1], (d), (r)[1]) , \
				 			 VScalarMul((m)[2], (d), (r)[2]) )

#define MVMul(m, v, r)		((r).x = VDot((m)[0], (v)) , \
				 			 (r).y = VDot((m)[1], (v)) , \
				 			 (r).z = VDot((m)[2], (v)) )

#define VUnit(v, t, r)		((t) = 1 / VMod(v) , \
				 			 VScalarMul(v, t, r) )

/* Printing functions. */
#define VPrint(v)	(printf("%f %f %f\n", (v).x, (v).y, (v).z))
#define MPrint(m)	(VPrint((m)[0]), VPrint((m)[1]), VPrint((m)[2]))


/* Extern declarations for the rest. */
extern int    VRead(char *s, Vector* v);

/* Matrix functions. */
/* These two don't become macros easily.  Hopefully they're not used much. */
extern Vector* MMMul(Matrix m1, Matrix m2, Matrix r);
extern Vector* MInvert(Matrix m, Matrix r);

#endif
