/*
**    Genray: A Generic Raytracer
**    Copyright (C) 1994  Stephen Chenney (stephen@cs.su.oz.au)
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program; if not, write to the Free Software
**    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
**	genray : a generic raytracer.
**
**	compound.c : File containing all the constructive solid geometry (CSG)
**	 functions.
**
*/

#include "genray.h"


/* Functions for intersecting compound objects. */
/* These differ from their ray_* couterparts in that they return a list	*/
/*  of hit times and normals.											*/
HitList* Comp_Sphere(Ray ray, Instance* obj);
HitList* Comp_Plane(Ray ray, Instance* obj);
HitList* Comp_Square(Ray ray, Instance* obj);
HitList* Comp_Cylinder(Ray ray, Instance* obj);
HitList* Comp_Cone(Ray ray, Instance* obj);
HitList* Comp_Cube(Ray ray, Instance* obj);
HitList* Comp_Torus(Ray ray, Instance* obj);
HitList* Comp_Compound(Ray ray, Instance* obj);



/* Two macros used to free lists. */
#define free_inst_list(list, tmp)	while (list) \
									{ \
										tmp = list; \
										list = list->next; \
										free_inst_elmt(tmp); \
									} \

#define free_hit_list(list, tmp, tmp1)	while (list) \
										{ \
											free_inst_list(list->path, tmp1); \
											tmp = list; \
											list = list->next; \
											free_hit_elmt(tmp); \
										} \


/*	char
**	Ray_Compound(Ray ray, Instance* obj, double* t, Vector* n, char* in_air,
**				InstanceList **derivation)
**	Determines whether ray hit the compound object obj.
**	The function itself returns TRUE / FALSE to indicate a hit or no.
**	t returns the hit time, n the hit vector, in_air the medium.
**	Derivation returns the path needed to reach the basic
**	 object, from top to bottom.
*/
char
Ray_Compound(Ray ray, Instance* obj, double* t, Vector* n, char* in_air,
			 InstanceList **derivation)
{
	HitList	*hit_list;
	HitList	*obj1_list = NULL;
	HitList	*obj2_list = NULL;
	HitList	*part_list;
	HitList	*temp_list;
	HitList	*temp_list1;

	InstanceList	*temp_inst;

	Ray		generic_ray;

	char	in_object, in_object1, in_object2;

	Vector	temp_v;
	Vector	*temp_m;

	/* For each sub-object, transform the ray and determine the hit	*/
	/*  times for that object.										*/

	if ( obj->object1->type == plane || Ray_In_Extent(ray, obj->object1))
	{
		/* Object1. */
		temp_v = (obj->object1->inverse_trans).displacement;
		VAdd(ray.start_point, temp_v, temp_v);
		temp_m = (obj->object1->inverse_trans).matrix;
		MVMul(temp_m, temp_v, generic_ray.start_point);
		MVMul(temp_m, ray.direction, generic_ray.direction);
		switch (obj->object1->type)
		{
			case sphere:
				obj1_list = Comp_Sphere(generic_ray, obj->object1);
				break;
			case plane:
				obj1_list = Comp_Plane(generic_ray, obj->object1);
				break;
			case square:
				obj1_list = Comp_Square(generic_ray, obj->object1);
				break;
			case cone:
				obj1_list = Comp_Cone(generic_ray, obj->object1);
				break;
			case cylinder:
				obj1_list = Comp_Cylinder(generic_ray, obj->object1);
				break;
			case cube:
				obj1_list = Comp_Cube(generic_ray, obj->object1);
				break;
			case torus:
				obj1_list = Comp_Torus(generic_ray, obj->object1);
				break;
			case union_op:
			case intersection_op:
			case difference_op:
				obj1_list = Comp_Compound(generic_ray, obj->object1);
				break;
		}
	}

	/* If necessary, allocate the derivation path.	*/
	if (obj1_list && (obj->object1->type != union_op) &&
		(obj->object1->type != intersection_op) &&
		(obj->object1->type != difference_op))
	{
		for (temp_list = obj1_list ; temp_list ; temp_list = temp_list->next)
		{
			temp_list->path = get_inst_elmt();
			temp_list->path->obj = obj->object1;
			temp_list->path->next = NULL;
		}
	}


	if ( obj->object2->type == plane || Ray_In_Extent(ray, obj->object2))
	{
		/* Object2. */
		temp_v = (obj->object2->inverse_trans).displacement;
		VAdd(ray.start_point, temp_v, temp_v);
		temp_m = (obj->object2->inverse_trans).matrix;
		MVMul(temp_m, temp_v, generic_ray.start_point);
		MVMul(temp_m, ray.direction, generic_ray.direction);
		switch (obj->object2->type)
		{
			case sphere:
				obj2_list = Comp_Sphere(generic_ray, obj->object2);
				break;
			case plane:
				obj2_list = Comp_Plane(generic_ray, obj->object2);
				break;
			case square:
				obj2_list = Comp_Square(generic_ray, obj->object2);
				break;
			case cone:
				obj2_list = Comp_Cone(generic_ray, obj->object2);
				break;
			case cylinder:
				obj2_list = Comp_Cylinder(generic_ray, obj->object2);
				break;
			case cube:
				obj2_list = Comp_Cube(generic_ray, obj->object2);
				break;
			case torus:
				obj2_list = Comp_Torus(generic_ray, obj->object2);
				break;
			case union_op:
			case intersection_op:
			case difference_op:
				obj2_list = Comp_Compound(generic_ray, obj->object2);
				break;
		}
	}
	/* If necessary, allocate the derivation path.	*/
	if (obj2_list && (obj->object2->type != union_op) &&
		(obj->object2->type != intersection_op) &&
		(obj->object2->type != difference_op))
	{
		for (temp_list = obj2_list ; temp_list ; temp_list = temp_list->next)
		{
			temp_list->path = get_inst_elmt();
			temp_list->path->obj = obj->object2;
			temp_list->path->next = NULL;
		}
	}


	/* Combine the lists. */
	if (!obj1_list && !obj2_list)
	{
		/* Both lists empty, no hits, return FALSE. */
		return FALSE;
	}
	if (!obj1_list)
	{
		if (obj->type == union_op)
			hit_list = obj2_list;
		else
		{
			free_hit_list(obj2_list, temp_list, temp_inst);
			hit_list = NULL;
		}
	}
	else if (!obj2_list)
	{
		if (obj->type == intersection_op)
		{
			free_hit_list(obj1_list, temp_list, temp_inst);
			hit_list = NULL;
		}
		else
			hit_list = obj1_list;
	}
	else
	{
		in_object = in_object1 = in_object2 = FALSE;
		part_list = hit_list = NULL;
		
		switch (obj->type)
		{
			/* First the non-cases (to stop gcc complaining) */
			case sphere:
			case cube:
			case plane:
			case square:
			case cone:
			case cylinder:
			case torus : break;

			case union_op:
				while (obj1_list && obj2_list)
				{
					if (obj1_list->t_hit < obj2_list->t_hit)
					{
						if (!in_object ||
							(in_object && in_object1 && !in_object2))
						{
							if (part_list)
							{
								part_list->next = obj1_list;
								part_list = part_list->next;
							}
							else
							{
								part_list = hit_list = obj1_list;
							}
							obj1_list = obj1_list->next;
						}
						else
						{
							temp_list = obj1_list;
							obj1_list = obj1_list->next;
							free_inst_list(temp_list->path, temp_inst);
							free_hit_elmt(temp_list);
						}
						in_object1 = !in_object1;
						in_object = in_object1 || in_object2;
					}
					else
					{
						if (!in_object ||
							(in_object && in_object2 && !in_object1))
						{
							if (part_list)
							{
								part_list->next = obj2_list;
								part_list = part_list->next;
							}
							else
							{
								part_list = hit_list = obj2_list;
							}
							obj2_list = obj2_list->next;
						}
						else
						{
							temp_list = obj2_list;
							obj2_list = obj2_list->next;
							free_inst_list(temp_list->path, temp_inst);
							free_hit_elmt(temp_list);
						}
						in_object2 = !in_object2;
						in_object = in_object1 || in_object2;
					}
				}
				/* Add the end bit */
				if (obj1_list)
					part_list->next = obj1_list;
				else
					part_list->next = obj2_list;
				break;

			case intersection_op:
				while (obj1_list && obj2_list)
				{
					if (obj1_list->t_hit < obj2_list->t_hit)
					{
						if (in_object || in_object2)
						{
							if (part_list)
							{
								part_list->next = obj1_list;
								part_list = part_list->next;
							}
							else
							{
								part_list = hit_list = obj1_list;
							}
							obj1_list = obj1_list->next;
						}
						else
						{
							temp_list = obj1_list;
							obj1_list = obj1_list->next;
							free_inst_list(temp_list->path, temp_inst);
							free_hit_elmt(temp_list);
						}
						in_object1 = !in_object1;
						in_object = in_object1 && in_object2;
					}
					else
					{
						if (in_object || in_object1)
						{
							if (part_list)
							{
								part_list->next = obj2_list;
								part_list = part_list->next;
							}
							else
							{
								part_list = hit_list = obj2_list;
							}
							obj2_list = obj2_list->next;
						}
						else
						{
							temp_list = obj2_list;
							obj2_list = obj2_list->next;
							free_inst_list(temp_list->path, temp_inst);
							free_hit_elmt(temp_list);
						}
						in_object2 = !in_object2;
						in_object = in_object1 && in_object2;
					}
				}

				/* Free the end bit */
				if (part_list)
					part_list->next = NULL;
				free_hit_list(obj1_list, temp_list, temp_inst);
				free_hit_list(obj2_list, temp_list, temp_inst);
				break;

			case difference_op:
				while (obj1_list && obj2_list)
				{
					if (obj1_list->t_hit < obj2_list->t_hit)
					{
						if (!in_object2)
						{
							if (part_list)
							{
								part_list->next = obj1_list;
								part_list = part_list->next;
							}
							else
							{
								part_list = hit_list = obj1_list;
							}
							obj1_list = obj1_list->next;
						}
						else
						{
							temp_list = obj1_list;
							obj1_list = obj1_list->next;
							free_inst_list(temp_list->path, temp_inst);
							free_hit_elmt(temp_list);
						}
						in_object1 = !in_object1;
						in_object = in_object1 && !in_object2;
					}
					else
					{
						if (in_object || in_object1)
						{
							if (part_list)
							{
								part_list->next = obj2_list;
								part_list = part_list->next;
							}
							else
							{
								part_list = hit_list = obj2_list;
							}
							obj2_list = obj2_list->next;
						}
						else
						{
							temp_list = obj2_list;
							obj2_list = obj2_list->next;
							free_inst_list(temp_list->path, temp_inst);
							free_hit_elmt(temp_list);
						}
						in_object2 = !in_object2;
						in_object = in_object1 && !in_object2;
					}
				}
				/* Add the end bit */
				if (obj1_list)
				{
					if (part_list)
						part_list->next = obj1_list;
					else
						part_list = hit_list = obj1_list;
				}
				else
				{
					if (part_list)
						part_list->next = NULL;
					/* Free list 2. */
					free_hit_list(obj2_list, temp_list, temp_inst);
				}
				/* Remove duplicate hits from the list.  These cause errors. */
				while (hit_list &&
					   dequal(hit_list->t_hit, hit_list->next->t_hit))
				{
					temp_list = hit_list->next->next;
					hit_list->next->next = NULL;
					free_hit_list(hit_list, temp_list1, temp_inst);
					hit_list = temp_list;
				}
				if (!hit_list) break;
				part_list = hit_list;
				temp_list = part_list->next->next;
				while (temp_list)
				{
					if (dequal(temp_list->t_hit, temp_list->next->t_hit))
					{
						part_list->next->next = temp_list->next->next;
						temp_list->next->next = NULL;
						free_hit_list(temp_list, temp_list1, temp_inst);
					}
					else
						part_list = part_list->next->next;
					temp_list = part_list->next->next;
				}
				break;

		}
	}

	/* Find the first valid hit. */
	in_object = FALSE;
	for ( temp_list = hit_list ;
		  temp_list && (temp_list->t_hit<DELTA) ;
		  temp_list=temp_list->next )
		in_object = !in_object;

	if (temp_list == NULL)
	{
		free_hit_list(hit_list, temp_list1, temp_inst);
		return FALSE;
	}

	*t = temp_list->t_hit;
	*n = temp_list->hit_normal;
	*in_air = !in_object;
	*derivation = temp_list->path;
	temp_list->path = NULL;	/* To stop it being freed. */

	free_hit_list(hit_list, temp_list, temp_inst);

	return TRUE;
}


/*	HitList*
**	Comp_Sphere(Ray ray, Instance* obj)
**	Intersects a ray with a sphere for combination.
**	The hit data is returned, or NULL if no hit.
*/
HitList*
Comp_Sphere(Ray ray, Instance* obj)
{
	HitList	*result;
	double	AA, BB, CC;
	double	discrim, discrim_root;
	Vector	temp_v1;
	Vector	temp_v2;

	temp_v1 = ray.direction;
	AA = VDot(temp_v1, temp_v1);

	if (AA == 0.0)
	{
		fprintf(stderr, "genray: Comp_Sphere: dud ray.\n");
		exit(1);
	}

	temp_v2 = ray.start_point;
	BB = VDot(temp_v2, temp_v1);
	CC = VDot(temp_v2, temp_v2) - 1;

	discrim = BB * BB - AA * CC;

	if (discrim < 0.0)
		return NULL;
	
	result = get_hit_elmt();
	result->next = get_hit_elmt();
	result->next->next = NULL;

	discrim_root = sqrt(discrim);

	/* A hit on the way in. */
	result->t_hit = (-BB - discrim_root) / AA;
	result->hit_normal.x = ray.start_point.x + result->t_hit * ray.direction.x;
	result->hit_normal.y = ray.start_point.y + result->t_hit * ray.direction.y;
	result->hit_normal.z = ray.start_point.z + result->t_hit * ray.direction.z;

	/* A hit on the way out.  The normal is reversed. */
	result->next->t_hit = (-BB + discrim_root) / AA;
	result->next->hit_normal.x = -(ray.start_point.x +
								  result->next->t_hit * ray.direction.x);
	result->next->hit_normal.y = -(ray.start_point.y +
								  result->next->t_hit * ray.direction.y);
	result->next->hit_normal.z = -(ray.start_point.z +
								  result->next->t_hit * ray.direction.z);

	return result;

}


/*	HitList*
**	Comp_Plane(Ray ray, Instance* obj)
**	Intersects a ray with a plane for combination.
**	The hit data is returned, or NULL if no hit.
*/
HitList*
Comp_Plane(Ray ray, Instance* obj)
{
	HitList	*result;

	if (ray.direction.z == 0.0) return NULL;

	result = get_hit_elmt();
	result->next = get_hit_elmt();
	result->next->next = NULL;

	result->t_hit = result->next->t_hit = -(ray.start_point.z) /
										  ray.direction.z;

	if (ray.direction.z > 0)
	{
		/* Ray is heading up from under plane. */
		VNew(0, 0, -1, result->hit_normal);
		VNew(0, 0, -1, result->next->hit_normal);
	}
	else
	{
		VNew(0, 0, 1, result->hit_normal);
		VNew(0, 0, 1, result->next->hit_normal);
	}

	return result;

}


/*	HitList*
**	Comp_Square(Ray ray, Instance* obj)
**	Intersects a ray with a square for combination.
**	The hit data is returned, or NULL if no hit.
*/
HitList*
Comp_Square(Ray ray, Instance* obj)
{
	HitList	*result;
	double	t;

	if (ray.direction.z == 0.0) return FALSE;

	t = -(ray.start_point.z) / ray.direction.z;

	if ((fabs(ray.start_point.x + ray.direction.x * t) <= 1.0) &&
	    (fabs(ray.start_point.y + ray.direction.y * t) <= 1.0))
	{
		result = get_hit_elmt();
		result->next = get_hit_elmt();
		result->next->next = NULL;

		if (ray.direction.z > 0)
		{
			/* Ray is heading up from under plane. */
			VNew(0, 0, -1, result->hit_normal);
			VNew(0, 0, -1, result->next->hit_normal);
		}
		else
		{
			VNew(0, 0, 1, result->hit_normal);
			VNew(0, 0, 1, result->next->hit_normal);
		}

		return result;
	}

	return NULL;
}


/*	HitList*
**	Comp_Cylinder(Ray ray, Instance* obj)
**	Intersects a ray with a cylinder for combination.
**	The hit data is returned, or NULL if no hit.
*/
HitList*
Comp_Cylinder(Ray ray, Instance* obj)
{
	HitList	*result;
	double	hit1 = -HUGE_VAL;	/* The hit on way in. */
	double	hit2 = -HUGE_VAL;	/* The hit on the way out. */
	double	z1, z2;
	double	AA, BB;
	double	discrim, discrim_root;

	if ((AA = ray.direction.x * ray.direction.x +
		  ray.direction.y * ray.direction.y) != 0.0)
	{
		/* There may be hits with the walls. */
		BB = ray.start_point.x * ray.direction.x +
		     ray.start_point.y * ray.direction.y;
		if ((discrim = BB * BB - AA * (ray.start_point.x * ray.start_point.x +
				ray.start_point.y * ray.start_point.y - 1)) >= 0.0)
		{
			/* There is a hit. */
			discrim_root = sqrt(discrim);

			hit1 = (-BB - discrim_root) / AA;
			z1 = ray.start_point.z + hit1 * ray.direction.z;
			hit2 = (-BB + discrim_root) / AA;
			z2 = ray.start_point.z + hit2 * ray.direction.z;

			if (((z1 > 1.0) && (z2 > 1.0)) || ((z2 < -1.0) && (z1 < -1.0)))
			{
				/* Ray entirely misses cylinder. */
				return NULL;
			}
			else
			{
				/* Must be hits. */
				result = get_hit_elmt();
				result->next = get_hit_elmt();
				result->next->next = NULL;

				if ((z1 < 1.0) && (z1 > -1.0) && (z2 < 1.0) && (z2 > -1.0))
				{
					/* Both hits on walls. */
					result->t_hit = hit1;
					result->hit_normal.x =
						ray.start_point.x + ray.direction.x * hit1;
					result->hit_normal.y =
						ray.start_point.y + ray.direction.y * hit1;

					result->next->t_hit = hit2;
					result->next->hit_normal.x =
						-ray.start_point.x - ray.direction.x * hit2;
					result->next->hit_normal.y =
						-ray.start_point.y - ray.direction.y * hit2;

					result->hit_normal.z = result->next->hit_normal.z = 0.0;

					return result;
				}
				
				if ((z1 < 1.0) && (z1 > -1.0))
				{
					/* hit1 is valid. */
					result->t_hit = hit1;
					result->hit_normal.x =
						ray.start_point.x + ray.direction.x * hit1;
					result->hit_normal.y =
						ray.start_point.y + ray.direction.y * hit1;
					result->hit_normal.z = 0.0;

					/* Find out where other hit is. Must be one end. */
					if (z2 > 1.0)
					{
						/* Ray must hit top. */
						hit2 = (1 - ray.start_point.z) / ray.direction.z;
						result->next->t_hit = hit2;
						result->next->hit_normal.x = 0.0;
						result->next->hit_normal.y = 0.0;
						result->next->hit_normal.z = -1.0;

						return result;
					}
					else
					{
						/* Ray must hit bottom. */
						hit2 = (-1.0 - ray.start_point.z) / ray.direction.z;

						result->next->t_hit = hit2;
						result->next->hit_normal.x = 0.0;
						result->next->hit_normal.y = 0.0;
						result->next->hit_normal.z = 1.0;

						return result;
					}
				}
				else if ((z2 < 1.0) && (z2 > -1.0))
				{
					/* hit2 is valid. */
					result->next->t_hit = hit2;
					result->next->hit_normal.x =
						-ray.start_point.x - ray.direction.x * hit2;
					result->next->hit_normal.y =
						-ray.start_point.y - ray.direction.y * hit2;
					result->next->hit_normal.z = 0.0;

					/* Find out where other hit is. Must be one end. */
					if (z1 > 1.0)
					{
						/* Ray must hit top. */
						hit1 = (1 - ray.start_point.z) / ray.direction.z;
						result->t_hit = hit1;
						result->hit_normal.x = 0.0;
						result->hit_normal.y = 0.0;
						result->hit_normal.z = 1.0;

						return result;
					}
					else
					{
						/* Ray must hit bottom. */
						hit1 = (-1.0 - ray.start_point.z) / ray.direction.z;

						result->t_hit = hit1;
						result->hit_normal.x = 0.0;
						result->hit_normal.y = 0.0;
						result->hit_normal.z = -1.0;

						return result;
					}
				}
				else
				{
					/* Both hits on ends. */
					/* Find out where other hit is. Must be one end. */
					if (z1 > 1.0)
					{
						/* Ray must hit top first. */
						hit1 = (1 - ray.start_point.z) / ray.direction.z;
						result->t_hit = hit1;
						result->hit_normal.x = 0.0;
						result->hit_normal.y = 0.0;
						result->hit_normal.z = 1.0;

						hit2 = (-1.0 - ray.start_point.z) / ray.direction.z;
						result->next->t_hit = hit2;
						result->next->hit_normal.x = 0.0;
						result->next->hit_normal.y = 0.0;
						result->next->hit_normal.z = 1.0;

						return result;
					}
					else
					{
						/* Ray must hit bottom first. */
						hit1 = (-1.0 - ray.start_point.z) / ray.direction.z;
						result->t_hit = hit1;
						result->hit_normal.x = 0.0;
						result->hit_normal.y = 0.0;
						result->hit_normal.z = -1.0;

						hit2 = (1 - ray.start_point.z) / ray.direction.z;
						result->next->t_hit = hit2;
						result->next->hit_normal.x = 0.0;
						result->next->hit_normal.y = 0.0;
						result->next->hit_normal.z = -1.0;

						return result;
					}
				}
			}
		}
		else return NULL;		/* Ray misses entirely. */
	}
	else
	{
		/* Ray is parallel to sides. */
		hit1 = (1 - ray.start_point.z) / ray.direction.z;

		/* Test for inside end circle. */
		AA = ray.start_point.x + hit2 * ray.direction.x;
		BB = ray.start_point.y + hit2 * ray.direction.y;
		if ((AA*AA + BB*BB) > 1.0)
			return NULL;

		/* One hit is valid so must the other one be. */
		hit2 = (-1.0 - ray.start_point.z) / ray.direction.z;

		result = get_hit_elmt();
		result->next = get_hit_elmt();
		result->next->next = NULL;

		if (hit1 < hit2)
		{
			result->t_hit = hit1;
			result->hit_normal.x = result->hit_normal.y = 0;
			result->hit_normal.z = 1.0;

			result->next->t_hit = hit2;
			result->next->hit_normal.x = result->next->hit_normal.y = 0;
			result->next->hit_normal.z = -1.0;

		}
		else
		{
			result->t_hit = hit2;
			result->hit_normal.x = result->hit_normal.y = 0;
			result->hit_normal.z = -1.0;

			result->next->t_hit = hit1;
			result->next->hit_normal.x = result->next->hit_normal.y = 0;
			result->next->hit_normal.z = 1.0;
		}

		return result;
	}
}


/*	HitList*
**	Comp_Cone(Ray ray, Instance* obj)
**	Intersects a ray with a cone for combination.
**	The hit data is returned, or NULL if no hit.
**	The code is entirely non obvious.  I arrived at it by
**	 drawing a pretty picture and examining cases. There are 9
**	 or so.
*/
HitList*
Comp_Cone(Ray ray, Instance* obj)
{
	HitList	*result;
	double	hit1;	/* The side hit on way in. */
	double	hit2;	/* The side hit on the way out. */
	double	z1, z2;
	double	AA, BB;
	double	temp_d;
	double	discrim, discrim_root;

	/* AA = 0 I think implies the ray lies in one of the walls of the cone. */
	if ((AA = ray.direction.x * ray.direction.x +
		  ray.direction.y * ray.direction.y -
		  ray.direction.z * ray.direction.z / 4) != 0.0)
	{
		/* There may be hits with the walls. */
		BB = ray.start_point.x * ray.direction.x +
		     ray.start_point.y * ray.direction.y +
		     ray.direction.z * ( 1 - ray.start_point.z) / 4;
		temp_d = 1 - ray.start_point.z;
		if ((discrim = BB * BB - AA * (ray.start_point.x * ray.start_point.x +
			ray.start_point.y * ray.start_point.y - temp_d*temp_d / 4)) >= 0.0)
		{
			/* There is a hit. */
			discrim_root = sqrt(discrim);

			if (AA > 0)
			{
				hit1 = (-BB - discrim_root) / AA;
				hit2 = (-BB + discrim_root) / AA;
			}
			else
			{
				hit1 = (-BB + discrim_root) / AA;
				hit2 = (-BB - discrim_root) / AA;
			}
			z1 = ray.start_point.z + hit1 * ray.direction.z;
			z2 = ray.start_point.z + hit2 * ray.direction.z;

			if (((z1 > 1.0) && ((z2 > 1.0) || (z2 < -1.0))) ||
				((z1 < -1.0) && ((z2 > 1.0) || (z2 < -1.0))))
			{
				/* Ray entirely misses cone. */
				return NULL;
			}
			else
			{
				/* Must be hits. */
				/* Must be hits. */
				result = get_hit_elmt();
				result->next = get_hit_elmt();
				result->next->next = NULL;

				if (((z1 < 1.0) && (z1 > -1.0)) && ((z2 < 1.0) && (z2 > -1.0)))
				{
					/* Both hits on walls. */
					result->t_hit = hit1;
					result->hit_normal.x =
						ray.start_point.x + ray.direction.x * hit1;
					result->hit_normal.y =
						ray.start_point.y + ray.direction.y * hit1;
					result->hit_normal.z = 0.125 * (1 - z1) * (1- z1);

					result->next->t_hit = hit2;
					result->next->hit_normal.x =
						-ray.start_point.x - ray.direction.x * hit2;
					result->next->hit_normal.y =
						-ray.start_point.y - ray.direction.y * hit2;
					result->hit_normal.z = -0.125 * (1 - z2) * (1- z2);

					return result;
				}
				else if ((z1 < 1.0) && (z1 > -1.0))
				{
					/* Valid base hit as hit2. */
					hit2 = (-1.0 - ray.start_point.z) / ray.direction.z;
					if (hit2 < hit1)
					{
						result->t_hit = hit2;
						result->hit_normal.x = result->hit_normal.y = 0.0;
						result->hit_normal.z = -1.0;

						result->next->t_hit = hit1;
						result->next->hit_normal.x =
							-ray.start_point.x - ray.direction.x * hit1;
						result->next->hit_normal.y =
							-ray.start_point.y - ray.direction.y * hit1;
						result->next->hit_normal.z =
							-0.125 * (1 - z1) * (1- z1);
					}
					else
					{
						result->next->t_hit = hit2;
						result->next->hit_normal.x = 0.0;
						result->next->hit_normal.y = 0.0;
						result->next->hit_normal.z = 1.0;

						result->t_hit = hit1;
						result->hit_normal.x =
							ray.start_point.x + ray.direction.x * hit1;
						result->hit_normal.y =
							ray.start_point.y + ray.direction.y * hit1;
						result->hit_normal.z = 0.125 * (1 - z1) * (1- z1);
					}

					return result;
				}
				else
				{
					/* hit2 must be good and a valid base hit. */
					hit1 = (-1.0 - ray.start_point.z) / ray.direction.z;
					if (hit1 < hit2)
					{
						result->t_hit = hit1;
						result->hit_normal.x = result->hit_normal.y = 0.0;
						result->hit_normal.z = -1.0;

						result->next->t_hit = hit2;
						result->next->hit_normal.x =
							-ray.start_point.x - ray.direction.x * hit2;
						result->next->hit_normal.y =
							-ray.start_point.y - ray.direction.y * hit2;
						result->next->hit_normal.z =
							-0.125 * (1 - z2) * (1- z2);
					}
					else
					{
						result->next->t_hit = hit1;
						result->next->hit_normal.x = 0.0;
						result->next->hit_normal.y = 0.0;
						result->next->hit_normal.z = 1.0;

						result->t_hit = hit2;
						result->hit_normal.x =
							ray.start_point.x + ray.direction.x * hit2;
						result->hit_normal.y =
							ray.start_point.y + ray.direction.y * hit2;
						result->hit_normal.z = 0.125 * (1 - z2) * (1- z2);
					}

					return result;
				}
			}
		}
		else return NULL;
	}
	else return NULL;
}



/*	HitList*
**	Comp_Cube(Ray ray, Instance* obj)
**	Intersects a ray with a cube for combination.
**	The hit data is returned, or NULL if no hit.
*/
HitList*
Comp_Cube(Ray ray, Instance* obj)
{
	HitList	*result;
	double	tin = -HUGE;
	double	tout = HUGE;
	double	thit;
	int	fin = 0;
	int	fout = 0;

	/* Since there are so many 0's in all this, I'm going	*/
	/*  to do it linearly.  This makes for poor readability	*/
	/*  and lots of code, but much greater speed.		*/

	/* x = 1 plane: n = {1,0,0}, D = 1 */
	if (ray.direction.x == 0.0)	/* Ray parallel.	*/
	{
		if (1 - ray.start_point.x < 0.0) return NULL;	/* Ray outside */
		/* Check x = -1 plane while we're at it. */
		if (1 + ray.start_point.x < 0.0) return NULL;
	}
	else
	{
		thit = (1 - ray.start_point.x) / ray.direction.x;
		if (ray.direction.x > 0.0)
		{
			tout = thit;
			fout = 0;
		}
		else
		{
			tin = thit;
			fin = 0;
		}
	}
	/* Not possible to eliminate ray yet */

	/* y = 1 plane: n = {0,1,0}, D = 1 */
	if (ray.direction.y == 0.0)     /* Ray parallel.        */
	{
		if (1 - ray.start_point.y < 0.0) return NULL;  /* Ray outside */
		/* Check y = -1 plane too. */
		if (1 + ray.start_point.y < 0.0) return NULL;
	}
	else
	{
		thit = (1 - ray.start_point.y) / ray.direction.y;
		if (ray.direction.y > 0.0)
		{
			if (thit < tout)
			{
				fout = 1;
				tout = thit;
			}
		}
		else
		{
			if (thit > tin)
			{
				fin = 1;
				tin = thit;
			}
		}
	}
	/* Try to eliminate intersection. */
	if (tin > tout) return NULL;

	/* z = 1 plane: n = {0,0,1}, D = 1 */
	if (ray.direction.z == 0.0)     /* Ray parallel.        */
	{
		if (1 - ray.start_point.z < 0.0) return NULL;  /* Ray outside */
		/* Check y = -1 plane too. */
		if (1 + ray.start_point.z < 0.0) return NULL;
	}
	else
	{
		thit = (1 - ray.start_point.z) / ray.direction.z;
		if (ray.direction.z > 0.0)
		{
			if (thit < tout)
			{
				fout = 2;
				tout = thit;
			}
		}
		else
		{
			if (thit > tin)
			{
				fin = 2;
				tin = thit;
			}
		}
	}
	/* Try to eliminate intersection. */
	if (tin > tout) return NULL;


	/* x = -1 plane : n = {-1,0,0}, D = 1 */
	/* Already eliminated parallel rays above. */
	thit = (1 + ray.start_point.x) / -ray.direction.x;
	if (ray.direction.x <  0.0)
	{
		if (thit < tout)
		{
			fout = 3;
			tout = thit;
		}
	}
	else
	{
		if (thit > tin)
		{
			fin = 3;
			tin = thit;
		}
	}
	if (tin > tout) return NULL;

	/* y = -1 plane : n = {0,-1,0}, D = 1 */
	/* Already eliminated parallel rays above. */
	thit = (1 + ray.start_point.y) / -ray.direction.y;
	if (ray.direction.y <  0.0)
	{
		if (thit < tout)
		{
			fout = 4;
			tout = thit;
		}
	}
	else
	{
		if (thit > tin)
		{
			fin = 4;
			tin = thit;
		}
	}
	if (tin > tout) return NULL;

	/* z = -1 plane : n = {0,0,-1}, D = 1 */
	/* Already eliminated parallel rays above. */
	thit = (1 + ray.start_point.z) / -ray.direction.z;
	if (ray.direction.z <  0.0)
	{
		if (thit < tout)
		{
			fout = 5;
			tout = thit;
		}
	}
	else
	{
		if (thit > tin)
		{
			fin = 5;
			tin = thit;
		}
	}
	if (tin > tout) return NULL;

	result = get_hit_elmt();
	result->next = get_hit_elmt();
	result->next->next = NULL;

	result->t_hit = tin;
	switch (fin)
	{
		case 0 : VNew(1, 0, 0, result->hit_normal);
			break;
		case 1 : VNew(0, 1, 0, result->hit_normal);
			break;
		case 2 : VNew(0, 0, 1, result->hit_normal);
			break;
		case 3 : VNew(-1, 0, 0, result->hit_normal);
			break;
		case 4 : VNew(0, -1, 0, result->hit_normal);
			break;
		case 5 : VNew(0, 0, -1, result->hit_normal);

	}

	result->next->t_hit = tout;
	switch (fout)
	{
		case 0 : VNew(-1, 0, 0, result->next->hit_normal);
			break;
		case 1 : VNew(0, -1, 0, result->next->hit_normal);
			break;
		case 2 : VNew(0, 0, -1, result->next->hit_normal);
			break;
		case 3 : VNew(1, 0, 0, result->next->hit_normal);
			break;
		case 4 : VNew(0, 1, 0, result->next->hit_normal);
			break;
		case 5 : VNew(0, 0, 1, result->next->hit_normal);
	}

	return result;

}


/*	HitList*
**	Comp_Compound(Ray ray, Instance* obj)
**	Intersects a ray with a compound object.
**	The hit data is returned, or NULL if no hit.
*/
HitList*
Comp_Compound(Ray ray, Instance* obj)
{
	HitList	*hit_list;
	HitList	*obj1_list = NULL;
	HitList	*obj2_list = NULL;
	HitList	*part_list;
	HitList	*temp_list;
	HitList	*temp_list1;

	InstanceList	*temp_inst;
	InstanceList	*temp_inst1;

	Ray		generic_ray;

	char	in_object, in_object1, in_object2;

	Vector	temp_v;
	Vector	*temp_m;

	/* For each sub-object, transform the ray and determine the hit	*/
	/*  times for that object.										*/

	if ( obj->object1->type == plane || Ray_In_Extent(ray, obj->object1))
	{
		/* Object1. */
		temp_v = (obj->object1->inverse_trans).displacement;
		VAdd(ray.start_point, temp_v, temp_v);
		temp_m = (obj->object1->inverse_trans).matrix;
		MVMul(temp_m, temp_v, generic_ray.start_point);
		MVMul(temp_m, ray.direction, generic_ray.direction);
		switch (obj->object1->type)
		{
			case sphere:
				obj1_list = Comp_Sphere(generic_ray, obj->object1);
				break;
			case plane:
				obj1_list = Comp_Plane(generic_ray, obj->object1);
				break;
			case square:
				obj1_list = Comp_Square(generic_ray, obj->object1);
				break;
			case cone:
				obj1_list = Comp_Cone(generic_ray, obj->object1);
				break;
			case cylinder:
				obj1_list = Comp_Cylinder(generic_ray, obj->object1);
				break;
			case cube:
				obj1_list = Comp_Cube(generic_ray, obj->object1);
				break;
			case torus:
				obj1_list = Comp_Torus(generic_ray, obj->object1);
				break;
			case union_op:
			case intersection_op:
			case difference_op:
				obj1_list = Comp_Compound(generic_ray, obj->object1);
				break;
		}
	}
	/* If necessary, allocate the derivation path.	*/
	if (obj1_list && (obj->object1->type != union_op) &&
		(obj->object1->type != intersection_op) &&
		(obj->object1->type != difference_op))
	{
		for (temp_list = obj1_list ; temp_list ; temp_list = temp_list->next)
		{
			temp_list->path = get_inst_elmt();
			temp_list->path->obj = obj->object1;
			temp_list->path->next = NULL;
		}
	}

	if ( obj->object2->type == plane || Ray_In_Extent(ray, obj->object2))
	{
		/* Object2. */
		temp_v = (obj->object2->inverse_trans).displacement;
		VAdd(ray.start_point, temp_v, temp_v);
		temp_m = (obj->object2->inverse_trans).matrix;
		MVMul(temp_m, temp_v, generic_ray.start_point);
		MVMul(temp_m, ray.direction, generic_ray.direction);
		switch (obj->object2->type)
		{
			case sphere:
				obj2_list = Comp_Sphere(generic_ray, obj->object2);
				break;
			case plane:
				obj2_list = Comp_Plane(generic_ray, obj->object2);
				break;
			case square:
				obj2_list = Comp_Square(generic_ray, obj->object2);
				break;
			case cone:
				obj2_list = Comp_Cone(generic_ray, obj->object2);
				break;
			case cylinder:
				obj2_list = Comp_Cylinder(generic_ray, obj->object2);
				break;
			case cube:
				obj2_list = Comp_Cube(generic_ray, obj->object2);
				break;
			case torus:
				obj2_list = Comp_Torus(generic_ray, obj->object2);
				break;
			case union_op:
			case intersection_op:
			case difference_op:
				obj2_list = Comp_Compound(generic_ray, obj->object2);
				break;
		}
	}
	/* If necessary, allocate the derivation path.	*/
	if (obj2_list && (obj->object2->type != union_op) &&
		(obj->object2->type != intersection_op) &&
		(obj->object2->type != difference_op))
	{
		for (temp_list = obj2_list ; temp_list ; temp_list = temp_list->next)
		{
			temp_list->path = get_inst_elmt();
			temp_list->path->obj = obj->object2;
			temp_list->path->next = NULL;
		}
	}


	/* Combine the lists. */
	if (!obj1_list && !obj2_list)
	{
		/* Both lists empty, no hits, return FALSE. */
		return NULL;
	}
	if (!obj1_list)
	{
		if (obj->type == union_op)
			hit_list = obj2_list;
		else
		{
			free_hit_list(obj2_list, temp_list, temp_inst);
			hit_list = NULL;
		}
	}
	else if (!obj2_list)
	{
		if (obj->type == intersection_op)
		{
			free_hit_list(obj1_list, temp_list, temp_inst);
			hit_list = NULL;
		}
		else
			hit_list = obj1_list;
	}
	else
	{
		in_object = in_object1 = in_object2 = FALSE;
		part_list = hit_list = NULL;
		
		switch (obj->type)
		{
			/* First the non-cases (to stop gcc complaining) */
			case sphere:
			case cube:
			case plane:
			case square:
			case cone:
			case cylinder:
			case torus : break;

			case union_op:
				while (obj1_list && obj2_list)
				{
					if (obj1_list->t_hit < obj2_list->t_hit)
					{
						if (!in_object ||
							(in_object && in_object1 && !in_object2))
						{
							if (part_list)
							{
								part_list->next = obj1_list;
								part_list = part_list->next;
							}
							else
							{
								part_list = hit_list = obj1_list;
							}
							obj1_list = obj1_list->next;
						}
						else
						{
							temp_list = obj1_list;
							obj1_list = obj1_list->next;
							free_inst_list(temp_list->path, temp_inst);
							free_hit_elmt(temp_list);
						}
						in_object1 = !in_object1;
						in_object = in_object1 || in_object2;
					}
					else
					{
						if (!in_object ||
							(in_object && in_object2 && !in_object1))
						{
							if (part_list)
							{
								part_list->next = obj2_list;
								part_list = part_list->next;
							}
							else
							{
								part_list = hit_list = obj2_list;
							}
							obj2_list = obj2_list->next;
						}
						else
						{
							temp_list = obj2_list;
							obj2_list = obj2_list->next;
							free_inst_list(temp_list->path, temp_inst);
							free_hit_elmt(temp_list);
						}
						in_object2 = !in_object2;
						in_object = in_object1 || in_object2;
					}
				}
				/* Add the end bit */
				if (obj1_list)
					part_list->next = obj1_list;
				else
					part_list->next = obj2_list;
				break;

			case intersection_op:
				while (obj1_list && obj2_list)
				{
					if (obj1_list->t_hit < obj2_list->t_hit)
					{
						if (in_object || in_object2)
						{
							if (part_list)
							{
								part_list->next = obj1_list;
								part_list = part_list->next;
							}
							else
							{
								part_list = hit_list = obj1_list;
							}
							obj1_list = obj1_list->next;
						}
						else
						{
							temp_list = obj1_list;
							obj1_list = obj1_list->next;
							free_inst_list(temp_list->path, temp_inst);
							free_hit_elmt(temp_list);
						}
						in_object1 = !in_object1;
						in_object = in_object1 && in_object2;
					}
					else
					{
						if (in_object || in_object1)
						{
							if (part_list)
							{
								part_list->next = obj2_list;
								part_list = part_list->next;
							}
							else
							{
								part_list = hit_list = obj2_list;
							}
							obj2_list = obj2_list->next;
						}
						else
						{
							temp_list = obj2_list;
							obj2_list = obj2_list->next;
							free_inst_list(temp_list->path, temp_inst);
							free_hit_elmt(temp_list);
						}
						in_object2 = !in_object2;
						in_object = in_object1 && in_object2;
					}
				}

				/* Free the end bit */
				if (part_list)
					part_list->next = NULL;
				free_hit_list(obj1_list, temp_list, temp_inst);
				free_hit_list(obj2_list, temp_list, temp_inst);
				break;

			case difference_op:
				while (obj1_list && obj2_list)
				{
					if (obj1_list->t_hit < obj2_list->t_hit)
					{
						if (!in_object2)
						{
							if (part_list)
							{
								part_list->next = obj1_list;
								part_list = part_list->next;
							}
							else
							{
								part_list = hit_list = obj1_list;
							}
							obj1_list = obj1_list->next;
						}
						else
						{
							temp_list = obj1_list;
							obj1_list = obj1_list->next;
							free_inst_list(temp_list->path, temp_inst);
							free_hit_elmt(temp_list);
						}
						in_object1 = !in_object1;
						in_object = in_object1 && !in_object2;
					}
					else
					{
						if (in_object || in_object1)
						{
							if (part_list)
							{
								part_list->next = obj2_list;
								part_list = part_list->next;
							}
							else
							{
								part_list = hit_list = obj2_list;
							}
							obj2_list = obj2_list->next;
						}
						else
						{
							temp_list = obj2_list;
							obj2_list = obj2_list->next;
							free_inst_list(temp_list->path, temp_inst);
							free_hit_elmt(temp_list);
						}
						in_object2 = !in_object2;
						in_object = in_object1 && !in_object2;
					}
				}
				/* Add the end bit */
				if (obj1_list)
				{
					if (part_list)
						part_list->next = obj1_list;
					else
						part_list = hit_list = obj1_list;
				}
				else
				{
					if (part_list)
						part_list->next = NULL;
					/* Free list 2. */
					free_hit_list(obj2_list, temp_list, temp_inst);
				}
				/* Remove duplicate hits from the list.  These cause errors. */
				while (hit_list &&
					   dequal(hit_list->t_hit, hit_list->next->t_hit))
				{
					temp_list = hit_list->next->next;
					hit_list->next->next = NULL;
					free_hit_list(hit_list, temp_list1, temp_inst);
					hit_list = temp_list;
				}
				if (!hit_list) break;
				part_list = hit_list;
				temp_list = part_list->next->next;
				while (temp_list)
				{
					if (dequal(temp_list->t_hit, temp_list->next->t_hit))
					{
						part_list->next->next = temp_list->next->next;
						temp_list->next->next = NULL;
						free_hit_list(temp_list, temp_list1, temp_inst);
						temp_list = NULL;
					}
					else
						part_list = part_list->next->next;
					temp_list = part_list->next->next;
				}
				break;

		}
	}

	/* Append this object to each valid path is the hit list. */
	for ( temp_list = hit_list ; temp_list && (temp_list->t_hit < 0.0) ;
		  temp_list = temp_list->next );

	for ( ; temp_list ; temp_list = temp_list->next )
	{
		/* Append this object to path. */
		temp_inst = get_inst_elmt();
		for ( temp_inst1 = temp_list->path ;
			  temp_inst1->next ;
			  temp_inst1 = temp_inst1->next );
		temp_inst1->next = temp_inst;
		temp_inst->obj = obj;
		temp_inst->next = NULL;
	}

	return hit_list;
}


/*	HitList*
**	Comp_Torus(Ray ray, Instance* obj)
**	Intersects ray with a generic torus and returns a list of hit
**	 points and normals.
**	The quartic root solving code is adapted from Roots3And4.c,
**	 from Graphics Gems.
**	The hit data is returned, or NULL if no hit.
*/
HitList*
Comp_Torus(Ray ray, Instance* obj)
{
	HitList	*result;
	HitList	*temp_elmt;
	double	AA, BB, CC;	/* Coefficients for sphere eqn.*/
	double	hit[4];
	double	temp1, temp2, temp3, temp4, temp5, temp6;
	double	line_scale;	/* The scaling factor for the ray direction. */
	double	coeffs[4];	/* coefficients for quartic eqn. */
	Vector	hit_pt;
	int		num = 0;
	int		i, j;

	/* Test for some special cases.								*/
	/* First, test for rays which don't intersect the top and	*/
	/*  bottom planes of the torus.  ie the plane z=1 and z=-1.	*/
	if ((ray.start_point.z > 1.0) && (ray.direction.z > 0.0))
		return NULL;

	if ((ray.start_point.z < -1.0) && (ray.direction.z < 0.0))
		return NULL;

	/* Test for rays which don't intersect the bounding sphere, r=d+1. */
	AA = VDot(ray.direction, ray.direction);
	BB = VDot(ray.start_point, ray.direction);
	CC = VDot(ray.start_point, ray.start_point) -
			(obj->torus_rad + 1.0) * (obj->torus_rad + 1.0);
	if ((BB * BB - AA * CC) < 0.0) return NULL;

	VUnit(ray.direction, line_scale, ray.direction);

	temp1 = obj->torus_rad * obj->torus_rad;
	temp2 = ray.start_point.z * ray.start_point.z;
	temp3 = ray.direction.z * ray.direction.z;
	temp4 = ray.direction.z * ray.start_point.z;
	temp5 = ray.start_point.x * ray.start_point.x + ray.start_point.y *
			ray.start_point.y + temp2 - temp1 - 1.0;
	temp6 = ray.start_point.x * ray.direction.x + ray.start_point.y *
			ray.direction.y + ray.start_point.z * ray.direction.z;

	coeffs[3] = 4.0 * temp6;
	coeffs[2] = 2.0 * (temp5 + 2.0 * temp6 * temp6 + 2.0 * temp1 * temp3);
	coeffs[1] = 4.0 * (temp6 * temp5 + 2.0 * temp1 * temp4);
	coeffs[0] = temp5 * temp5 + 4.0 * temp1 * (temp2 - 1.0);

	/* Solve the resulting quartic equation.				*/
	num = SolveQuartic(coeffs, hit);

	if (num == 0)
		return NULL;

	/* Sort the hits. */
	for ( i = 0 ; i < num-1 ; i++ )
	{
		for ( j = i + 1 ; j < num ; j++ )
		{
			if (hit[i] > hit[j])
			{
				temp1 = hit[i];
				hit[i] = hit[j];
				hit[j] = temp1;
			}
		}
	}

	/* Allocate space. */
	if (num == 2)
	{
		result = get_hit_elmt();
		result->t_hit = hit[0] * line_scale;
		result->next = get_hit_elmt();
		result->next->t_hit = hit[1] * line_scale;
		result->next->next = NULL;
	}
	else
	{
		result = get_hit_elmt();
		result->t_hit = hit[0] * line_scale;
		result->next = get_hit_elmt();
		result->next->t_hit = hit[1] * line_scale;
		result->next->next = get_hit_elmt();
		result->next->next->t_hit = hit[2] * line_scale;
		result->next->next->next = get_hit_elmt();
		result->next->next->next->t_hit = hit[3] * line_scale;
		result->next->next->next->next = NULL;
	}

	/* Work out the normals. */
	for ( temp_elmt = result, i = 0 ;
		  temp_elmt ;
		  temp_elmt = temp_elmt->next, i++ )
	{
		/* Work out the normal. */
		hit_pt = ray.direction;
		VScalarMul(hit_pt, hit[i], hit_pt);
		VAdd(hit_pt, ray.start_point, hit_pt);

		temp1 = sqrt(hit_pt.x * hit_pt.x + hit_pt.y * hit_pt.y);
		temp2 = obj->torus_rad * hit_pt.x / temp1;
		temp3 = obj->torus_rad * hit_pt.y / temp1;
		if (i % 2)
		{
			/* inside hit. */
			temp_elmt->hit_normal.x = temp2 - hit_pt.x;
			temp_elmt->hit_normal.y = temp3 - hit_pt.y;
			temp_elmt->hit_normal.z = -hit_pt.z;
		}
		else
		{
			temp_elmt->hit_normal.x = hit_pt.x - temp2;
			temp_elmt->hit_normal.y = hit_pt.y - temp3;
			temp_elmt->hit_normal.z = hit_pt.z;
		}
	}

	return result;

}

