/*
**    Genray: A Generic Raytracer
**    Copyright (C) 1994  Stephen Chenney (stephen@cs.su.oz.au)
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program; if not, write to the Free Software
**    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/****************************************************************/
/*																*/
/*	Main program for genray.									*/
/*	Written by Stephen Chenney.  Dec 1993.						*/
/*																*/
/****************************************************************/

#include "xstuff.h"
#include "genray.h"

#ifdef HAVE_TIFF
#include <tiff.h>
#include <tiffio.h>
#endif



/* Declaration of global variables. */
double  win_width;		/* The leftmost x value of the viewing window   */
double  win_height;     /* The uppermost y value of the viewing window  */
Vector	min_extent, max_extent;	/* The min and max extents for drawing	*/


Vector  view_reference; /* The view reference point.                    */
Vector  eye_position;   /* The eye in world co-ordinates. */
double	eyedist;		/* The distance the eye is behind the viewplane.*/


Matrix  convert_matrix; /* The matrix to take a view point to a world point. */
						/* Note this is the transpose of the matrix to take  */
						/* world points to view points.                      */


Instance*       objects = NULL; /* The list of objects to raytrace.     */


Light*	light_srcs;		/* The light sources in the scene.	*/
int	light_number = 0;	/* The number of such light sources.	*/
Colour	ambient_light = 0x001F1F1F;	/*  Default. */



char	do_colour = FALSE;	/* TRUE for colour monitor.	*/
char	silent = FALSE;		/* TRUE for no visual output. */
char	use_true_colours = FALSE;	/* If true, will attempt to use the	*/
									/* actual colour of an object for	*/
									/* run time rendering, rather than	*/
									/* grey.							*/
#ifdef HAVE_TIFF
char	write_tiff = TRUE;
#else
char    write_tiff = FALSE;
#endif
char	*out_filename;		/* Name of the output file. */


unsigned long grey[256];	/* Array of values for grey shades.	*/
/* Declarations for greyscale dithers.*/
char greyshade[17][4][4] = {
	{{0,0,0,0}, {0,0,0,0}, {0,0,0,0}, {0,0,0,0}},
	{{0,0,0,0}, {0,0,0,0}, {0,0,0,0}, {1,0,0,0}},
	{{0,0,0,0}, {0,0,1,0}, {0,0,0,0}, {1,0,0,0}},
	{{0,0,0,0}, {0,0,1,0}, {0,0,0,0}, {1,0,1,0}},
	{{0,0,0,0}, {1,0,1,0}, {0,0,0,0}, {1,0,1,0}},
	{{0,0,0,0}, {1,0,1,0}, {0,1,0,0}, {1,0,1,0}},
	{{0,0,0,1}, {1,0,1,0}, {0,1,0,0}, {1,0,1,0}},
	{{0,0,0,1}, {1,0,1,0}, {0,1,0,1}, {1,0,1,0}},
	{{0,1,0,1}, {1,0,1,0}, {0,1,0,1}, {1,0,1,0}},
	{{0,1,0,1}, {1,0,1,0}, {0,1,0,1}, {1,1,1,0}},
	{{0,1,0,1}, {1,0,1,1}, {0,1,0,1}, {1,1,1,0}},
	{{0,1,0,1}, {1,0,1,1}, {0,1,0,1}, {1,1,1,1}},
	{{0,1,0,1}, {1,1,1,1}, {0,1,0,1}, {1,1,1,1}},
	{{0,1,0,1}, {1,1,1,1}, {1,1,0,1}, {1,1,1,1}},
	{{0,1,1,1}, {1,1,1,1}, {1,1,0,1}, {1,1,1,1}},
	{{0,1,1,1}, {1,1,1,1}, {1,1,1,1}, {1,1,1,1}},
	{{1,1,1,1}, {1,1,1,1}, {1,1,1,1}, {1,1,1,1}}};


Colour	*intensity = NULL;	/* The array of pixel colours.	*/



/* Variables associated with the display. */
XtAppContext	app_context;
Widget			top_level;
Widget			view_window;
Widget			menu_widget;
Widget			menu_children[2];
XRectangle		screen_rect;
Pixmap			drawing_pixmap;
XGCValues		gc_values;
GC				graphic_cont;


/* Function prototypes. */
void Colour_Init();
void Do_Raytrace();
Colour Raytrace_Primary(Ray scene_ray, double c, double r);
void Reshaped(Widget w, XtPointer cl_data, XtPointer ca_data);
void Redraw(Widget w, XtPointer cl_data, XtPointer ca_data);
void Save_PPM_Format();
#ifdef HAVE_TIFF
void Save_TIFF_Format();
#endif


void
main(int argc, char* argv[])
/****************************************************************/
/*																*/
/*																*/
/****************************************************************/
{

	/* Initialise various things. */
	Ray_Init(argc, argv);
	Extents_Init(&min_extent, &max_extent);
	Calculate_S_Bounds(3);

	/* Initialise colours, if necessary. */
	if (do_colour && !silent) Colour_Init();

	if (!silent)
	{
		/* Clear the background. */
		gc_values.function = GXcopy;
		gc_values.foreground = XBlackPixel(XtDisplay(top_level),
										XDefaultScreen(XtDisplay(top_level)));

		graphic_cont = XtGetGC(top_level, GCForeground|GCFunction, &gc_values);

		XFillRectangle(XtDisplay(top_level), XtWindow(view_window),
					graphic_cont, 0, 0, screen_rect.width, screen_rect.height);


		/* If in black and white, only need one GC (white foreground). */
		if (!do_colour)
		{
			gc_values.function = GXcopy;
			gc_values.foreground = XWhitePixel(XtDisplay(top_level),
										XDefaultScreen(XtDisplay(top_level)));
			gc_values.background = XBlackPixel(XtDisplay(top_level),
										XDefaultScreen(XtDisplay(top_level)));
			graphic_cont = XtGetGC(top_level, GCForeground|GCBackground|GCFunction,
								&gc_values);
		}

		/* Add resize and expose callbacks. */
		XtAddCallback(view_window, XtNresizeCallback, Reshaped, NULL);
		XtAddCallback(view_window, XtNexposeCallback, Redraw, NULL);
	}

	/* Redraw calls it if we aren't silent. */
	if (silent)
		Do_Raytrace();


	if (!silent)
	{
		XtAppMainLoop(app_context);
	}
	else
	{
		Save_Function(top_level, NULL, NULL);
	}

	exit(0);

}


void
Do_Raytrace()
/****************************************************************/
/*																*/
/*	Does the actual raytrace.  It needs to be a function so		*/
/*	 that ereshaped can call it.								*/
/*																*/
/****************************************************************/
{
	Colour	hit_colour;	/* The colour to draw.	*/
	short	hit_shade;	/* The greyshade to draw.	*/

	Ray	the_ray;			/* The ray we are currently dealing with. */
	double	row, column;	/* Current position in the window.	*/
	XPoint	current_pt;		/* The current point on the screen.	*/
	double	row_inc, colm_inc;	/* The increments for above.	*/
	Vector	temp_v;
	double	min_row, max_row, min_col, max_col;	/* bounding extent for scene. */
	int	startx, starty;
	int	i, base;



	/* Make space for the intensity array.  */
	if ((intensity = (Colour*)malloc(screen_rect.width*screen_rect.height*sizeof(Colour)))==NULL)
	{
		fprintf(stderr,"genray: ERROR: Could not malloc sufficient space.\n");
		exit(1);
	}

	/* All rays emanate from the eye. */
	the_ray.start_point = eye_position;

	/* Set up the increment variables. */
	row_inc = win_height / (double)screen_rect.height;
	colm_inc = win_width / (double)screen_rect.width;

	/* Set the actual extents for the raytrace. */
	min_row = (-0.5 * win_height > min_extent.y) ?
		-0.5 * win_height : min_extent.y;
	max_row = (0.5 * win_height < max_extent.y) ?
		0.5 * win_height : max_extent.y;
	min_col = (-0.5 * win_width > min_extent.x) ?
		-0.5 * win_width : min_extent.x;
	max_col = (0.5 * win_width < max_extent.x) ?
		0.5 * win_width : max_extent.x;

	/* Set the start x and y points for the screen. */
	starty = screen_rect.height - (min_row + 0.5 * win_height) / row_inc - 1;
	startx = (min_col + 0.5 * win_width) / colm_inc;

	/* These loops maintain 2 sets of loop variables.	*/
	/* The row and colm refer to the viewplane window.	*/
	/* Current point refers to the mapping from the window	*/
	/*  to the screen.  It is the point which is actually	*/
	/*  drawn.						*/
	for ( row = min_row , current_pt.y = starty ;
	      (row < max_row) && (current_pt.y >= 0) ;
	      row += row_inc , (current_pt.y)-- )
	{
		base = current_pt.y * screen_rect.width;
		for ( column = min_col , current_pt.x = startx ;
		      (column < max_col) &&
		       ((unsigned)current_pt.x < screen_rect.width) ;
		      column += colm_inc , (current_pt.x)++ )
		{
			/* Calculate the new ray direction.  pg 617 TB */
			VNew(column, row, eyedist, temp_v);
			MVMul(convert_matrix, temp_v, the_ray.direction);

			/* Actually do the trace for this point. */
			hit_colour = Raytrace_Primary(the_ray, column, row);
			intensity[base + current_pt.x] =
				hit_colour;

			if (!silent)
			{
				/* Convert the colour to a greyscale for output. */
				/* I got the formula from the news board comp.graphics.	*/
				/* It does pay to read news.							*/
				hit_shade = (short)(0.299 * (hit_colour & ColMask) +
									0.587 * ((hit_colour >> 8 ) & ColMask) +
									0.114 * ((hit_colour >> 16 ) & ColMask));

				/* Put the point on the screen. */
				if (do_colour)
				{
					gc_values.function = GXcopy;
					gc_values.foreground = grey[GetDitheredColor(hit_colour & ColMask, (hit_colour >> 8 ) & ColMask, (hit_colour >> 16 ) & ColMask, current_pt.x, current_pt.y)];

					graphic_cont = XtGetGC(top_level, GCForeground, &gc_values);
					XDrawPoint(XtDisplay(top_level), XtWindow(view_window),
							   graphic_cont, current_pt.x, current_pt.y);
				}
				else
				{
					/* Only draw white points.	*/
					if (greyshade[(int)((double)hit_shade * 16.0 / 255.0)]
									[(current_pt.x)&03][(current_pt.y)&03])
					{
						XDrawPoint(XtDisplay(top_level), XtWindow(view_window),
								   graphic_cont, current_pt.x, current_pt.y);
					}
				}
			}
		}
	}

}



Colour
Raytrace_Primary(Ray scene_ray, double c, double r)
/****************************************************************/
/*																*/
/*	Determines the first hit of ray with an object, and			*/
/*	 returns the colour for that object.						*/
/*																*/
/****************************************************************/
{
	Instance*	obj;		/* The current object of interest.	*/
	Ray	generic_ray;		/* The scene ray transformed.		*/
	char	have_hit = FALSE;
	double	t_hit = HUGE;	/* A really big first hit.	*/
	double	t_this;
	Instance*	first = NULL;	/* The object to be drawn.	*/
	Vector	hit_normal;		/* The normal at the hit point.		*/
	Vector	best_normal;	/* The normal for the best hit.		*/
	Colour	result;
	char	in_air;			/* Whether the ray was passing through	*/
							/*  "air" when it hit.			*/
	char	best_air = TRUE;
	InstanceList	*comp_path;	/* The path used to find the hit normal.	*/
	InstanceList	*best_path = NULL;
	InstanceList *tmp;

	Vector	temp_v;
	Vector*	temp_m;	/* Actually a matrix. */

	for ( obj = objects ; obj != NULL ; obj = obj->next )
	{
		/* test against projection extent */
		if  ((obj->proj_extent.min.x > c) || (obj->proj_extent.min.y > r) ||
		     (obj->proj_extent.max.x < c) || (obj->proj_extent.max.y < r))
			continue;

		/* Untransform the ray back to generic co-ords. */
		VAdd(scene_ray.start_point, (obj->inverse_trans).displacement, temp_v);
		temp_m = (obj->inverse_trans).matrix;
		MVMul(temp_m, temp_v, generic_ray.start_point);
		MVMul(temp_m, scene_ray.direction, generic_ray.direction);

		comp_path = NULL;

		switch (obj->type)
		{
			case sphere :
				have_hit = Ray_Sphere(generic_ray, obj, &t_this, &hit_normal,
										&in_air);
				break;
			case plane :
				have_hit = Ray_Plane(generic_ray, obj, &t_this, &hit_normal,
										&in_air);
				break;
			case square :
				have_hit = Ray_Square(generic_ray, obj, &t_this, &hit_normal,
										&in_air);
				break;
			case cylinder :
				have_hit = Ray_Cylinder(generic_ray, obj, &t_this, &hit_normal,
										&in_air);
				break;
			case cone :
				have_hit = Ray_Cone(generic_ray, obj, &t_this, &hit_normal,
									&in_air);
				break;
			case cube :
				have_hit = Ray_Cube(generic_ray, obj, &t_this, &hit_normal,
									&in_air);
				break;
			case torus :
				have_hit = Ray_Torus(generic_ray, obj, &t_this, &hit_normal,
									&in_air);
				break;
			case union_op :
			case difference_op :
			case intersection_op :
				have_hit = Ray_Compound(generic_ray, obj, &t_this, &hit_normal,
										&in_air, &comp_path);
				break;
		}

		if (have_hit && (t_this < t_hit))
		{
			t_hit = t_this;
			best_normal = hit_normal;
			best_air = in_air;
			while (best_path)
			{
				tmp = best_path;
				best_path = best_path->next;
				free_inst_elmt(tmp);
			}
			best_path = comp_path;
			first = obj;
		}
		else
		{
			while (comp_path)
			{
				tmp = comp_path;
				comp_path = comp_path->next;
				free_inst_elmt(tmp);
			}
		}

	}

	if (first == NULL)
	{
		result = 0;
		return result; /* black */
	}
	else
		return Find_Colour(scene_ray, first, t_hit, best_normal, 0, 1.0, best_air,
						best_path);

}


Colour
RaytraceSecondary(Ray scene_ray, Instance *origin, int depth, double intensity)
/****************************************************************/
/*																*/
/*	Raytraces a secondary ray.  This differs from the primary	*/
/*	 only in that rays are tested against an object's 3D		*/
/*	 space_extent, rather than a 2D projection extent.			*/
/*	It also takes into account how significant the ray is to	*/
/*	 the final colour value (intensity) and how deep the light	*/
/*	 tree has gone (depth).										*/
/*																*/
/****************************************************************/
{
	Instance*	obj;	/* The current object of interest.	*/
	Ray	generic_ray;	/* The scene ray transformed.		*/
	char	have_hit = FALSE;
	double	t_hit = HUGE;	/* A really big first hit.	*/
	double	t_this;
	Instance*	first = NULL;	/* The object to be drawn.	*/
	Vector	hit_normal;		/* The normal at the hit point.		*/
	Vector	best_normal;	/* The normal for the best hit.		*/
	Colour	result;
	char	in_air;
	char	best_air = TRUE;
	InstanceList	*comp_path;
	InstanceList	*best_path = NULL;
	InstanceList	*tmp;

	Vector	temp_v;
	Vector*	temp_m;	/* Actually a matrix. */

	for ( obj = objects ; obj != NULL ; obj = obj->next )
	{
		/* test against spacial extent */
		if ((obj->type != plane) && ! Ray_In_Extent(scene_ray, obj))
			continue;

		/* Untransform the ray back to generic co-ords.  pg 625 TB */
		temp_v = (obj->inverse_trans).displacement;
		VAdd(scene_ray.start_point, temp_v, temp_v);
		temp_m = (obj->inverse_trans).matrix;
		MVMul(temp_m, temp_v, generic_ray.start_point);
		MVMul(temp_m, scene_ray.direction, generic_ray.direction);

		comp_path = NULL;

		switch (obj->type)
		{
			case sphere :
				have_hit = Ray_Sphere(generic_ray, obj, &t_this, &hit_normal,
										&in_air);
				break;
			case plane :
				have_hit = Ray_Plane(generic_ray, obj, &t_this, &hit_normal,
										&in_air);
				break;
			case square :
				have_hit = Ray_Square(generic_ray, obj, &t_this, &hit_normal,
										&in_air);
				break;
			case cylinder :
				have_hit = Ray_Cylinder(generic_ray, obj, &t_this, &hit_normal,
										&in_air);
				break;
			case cone :
				have_hit = Ray_Cone(generic_ray, obj, &t_this, &hit_normal,
										&in_air);
				break;
			case cube :
				have_hit = Ray_Cube(generic_ray, obj, &t_this, &hit_normal,
										&in_air);
				break;
			case torus :
				have_hit = Ray_Torus(generic_ray, obj, &t_this, &hit_normal,
										&in_air);
				break;
			case union_op :
			case difference_op :
			case intersection_op :
				have_hit = Ray_Compound(generic_ray, obj, &t_this, &hit_normal,
										&in_air, &comp_path);
				break;
		}

		if (have_hit && (t_this < t_hit))
		{
			t_hit = t_this;
			best_normal = hit_normal;
			best_air = in_air;
			while (best_path)
			{
				tmp = best_path;
				best_path = best_path->next;
				free_inst_elmt(tmp);
			}
			best_path = comp_path;
			first = obj;
		}
		else
		{
			while (comp_path)
			{
				tmp = comp_path;
				comp_path = comp_path->next;
				free_inst_elmt(tmp);
			}
		}
	}

	if (first == NULL)
	{
		result = 0;
		return result; /* black */
	}
	else
		return Find_Colour(scene_ray, first, t_hit, best_normal, depth,
					  intensity, best_air, best_path);

}



void
Colour_Init()
/****************************************************************/
/*																*/
/*	Initialises 256 greys for the application.					*/
/*																*/
/****************************************************************/
{
	short	i;
	XColor	rgb;
	Display	*disp;
	Colormap	colmap;
	Arg		arg[5];

	disp = XtDisplay(top_level);
	XtSetArg(arg[0], XtNcolormap, &colmap);
	XtGetValues(top_level, arg, 1);

	rgb.flags = DoRed|DoGreen|DoBlue;
	/*
	for ( i = 0 ; i < 256 ; i++ )
	{
		rgb.red = ((512 * i - i * i) / 256) << 8;
		rgb.green = ((512 * i - i * i) / 256) << 8;
		rgb.blue = ((512 * i - i * i) / 256) << 8;
		*/
		/*rgb.red = i << 8;
		rgb.green = i << 8;
		rgb.blue = i << 8;*/
		/*
		XAllocColor(disp, colmap, &rgb);
		grey[i] = rgb.pixel;
	}
	*/

	for ( i = 0 ; i < 216 ; i++ )
	{
		rgb.red = ( ( i / 36 ) * 51 ) << 8;
		rgb.green = ( ( ( i % 36 ) / 6 ) * 51 ) << 8;
		rgb.blue = ( ( i % 6 ) * 51 ) << 8;
		XAllocColor(disp, colmap, &rgb);
		grey[i] = rgb.pixel;
	}
}


void
Reshaped(Widget w, XtPointer cl_data, XtPointer ca_data)
/****************************************************************/
/*																*/
/*	Called upon a resize for the window.  Frees the intensity	*/
/*	array, resets the screen size and redoes the raytrace.		*/
/*																*/
/****************************************************************/
{
	XPoint	new_size = *(XPoint*)ca_data;
	int i;

	/* Free the old intensity array.  */
	if (intensity)
	{
		free(intensity);
	}

	screen_rect.width = (unsigned)new_size.x;
	screen_rect.height = (unsigned)new_size.y;

	Do_Raytrace();
}



void
Redraw(Widget w, XtPointer cl_data, XtPointer ca_data)
/****************************************************************/
/*																*/
/*	Called upon exposure for the window.  ca_data holds the		*/
/*	region exposed.												*/
/*																*/
/****************************************************************/
{
	XRectangle	expose_rect;
	XPoint		current_pt;
	short		hit_shade;
	register int	col, base;
	static Boolean	first_time = TRUE;

	XClipBox(*(Region*)ca_data, &expose_rect);

	if (first_time)
	{
		Dimension	win_width, win_height;
		Arg			args[2];
		XPoint	corner;

		XtSetArg(args[0], XtNwidth, &win_width);
		XtSetArg(args[1], XtNheight, &win_height);
		XtGetValues(w, args, 2);

		corner.x = (short)win_width;
		corner.y = (short)win_height;
		first_time = FALSE;
		Reshaped(w, NULL, (XtPointer)&corner);
	}
	else
	{
		/* Redraw those points that fall within the area. */
		for ( current_pt.y = expose_rect.y ;
			  current_pt.y < expose_rect.y + expose_rect.height ;
			  current_pt.y++ )
		{
			base = current_pt.y * screen_rect.width;
			for ( current_pt.x = expose_rect.x ;
				  current_pt.x < expose_rect.x + expose_rect.width ;
				  current_pt.x++ )
			{
				col = intensity[base + current_pt.x];
				hit_shade = (short)(0.299 * (col & ColMask) +
									0.587 * ((col >> 8 ) & ColMask) +
									0.114 * ((col >> 16 ) & ColMask));


				if (do_colour)
				{
					gc_values.function = GXcopy;
					gc_values.foreground = grey[hit_shade];
					graphic_cont = XtGetGC(top_level, GCForeground, &gc_values);
					XDrawPoint(XtDisplay(top_level), XtWindow(view_window),
							   graphic_cont, current_pt.x, current_pt.y);
				}
				else
				{
						/* Only draw white points.	*/
					if (greyshade[(int)((double)hit_shade * 16.0 / 255.0)]
									[(current_pt.x)&03][(current_pt.y)&03])
					{
						XDrawPoint(XtDisplay(top_level), XtWindow(view_window),
								   graphic_cont, current_pt.x, current_pt.y);
					}
				}
			}
		}
	}

}




void
Save_Function(Widget w, XtPointer cl_data, XtPointer ca_data)
/****************************************************************/
/*																*/
/*	The callback function for the save menu item.				*/
/*																*/
/****************************************************************/
{
#ifdef HAVE_TIFF
	if (write_tiff)
		Save_TIFF_Format();
	else
#endif
		Save_PPM_Format();
}

void
Save_PPM_Format()
/****************************************************************/
/*																*/
/*	Saves the image in ppm format.								*/
/*																*/
/****************************************************************/
{
	FILE	*outfile;	/* File pointer for output file.*/
	register int	col, base;
	int		i, j;


	/* Write the output file. */
	if ((outfile = fopen(out_filename, "w")) == NULL)
	{
		fprintf(stderr, "genray : ERROR : Couldn't open output file %s\n", out_filename);
		return;
	}

	/* Create output file in raw ppm format. */
	fprintf(outfile, "P6\n");
	fprintf(outfile, "%d %d\n", screen_rect.width, screen_rect.height);
	fprintf(outfile, "255\n");
	for ( i = 0 ; i < screen_rect.height ; i++ )
	{
		base = i * screen_rect.width;
  		for ( j = 0 ; j < screen_rect.width ; j++ )
		{
			col = intensity[base + j];
			fprintf(outfile, "%c%c%c", col & ColMask, (col >> 8) & ColMask,
				(col >> 16) & ColMask);
		}
	}

	fclose(outfile);

}

#ifdef HAVE_TIFF
void
Save_TIFF_Format()
/****************************************************************/
/*																*/
/*	Saves the image in tiff format.								*/
/*																*/
/****************************************************************/
{
#define ushort unsigned short
#define uchar unsigned char

	TIFF	*outfile;
	uchar	*data;
	register int col, base;
	int		i, j, k;

	/* Open the output file. */
	if ((outfile = TIFFOpen(out_filename, "w")) == NULL)
	{
		fprintf(stderr, "genray : ERROR : Couldn't open output file %s\n", out_filename);
		return;
	}

	data = malloc(screen_rect.width * 3 * sizeof(uchar));

	/* Set the various image parameters. */
	TIFFSetField(outfile, TIFFTAG_BITSPERSAMPLE, (ushort)8);
	TIFFSetField(outfile, TIFFTAG_XRESOLUTION, 1.0);
	TIFFSetField(outfile, TIFFTAG_YRESOLUTION, 1.0);
	TIFFSetField(outfile, TIFFTAG_RESOLUTIONUNIT, 1);
	TIFFSetField(outfile, TIFFTAG_IMAGELENGTH, (ushort)screen_rect.height);
	TIFFSetField(outfile, TIFFTAG_IMAGEWIDTH, (ushort)screen_rect.width);
	TIFFSetField(outfile, TIFFTAG_PLANARCONFIG, (ushort)1);
	TIFFSetField(outfile, TIFFTAG_SAMPLESPERPIXEL, (ushort)3);
	TIFFSetField(outfile, TIFFTAG_PHOTOMETRIC, 2);
	TIFFSetField(outfile, TIFFTAG_COMPRESSION, 5);

	/* Write it out. */
	for ( i = 0 ; i < screen_rect.height ; i++ )
	{
		k = 0;
		base = i * screen_rect.width;
		for ( j = 0 ; j < screen_rect.width ; j++ )
		{
			col = intensity[base + j];
			data[k++] = (uchar)(col & ColMask);
			data[k++] = (uchar)((col >> 8 ) & ColMask);
			data[k++] = (uchar)((col >> 16 ) & ColMask);
		}
		TIFFWriteScanline(outfile, data, i, 0);
	}

	TIFFClose(outfile);

#undef u_short
}
#endif

/* The exit callback function. */
void
Exit_Function(Widget w, XtPointer cl_data, XtPointer ca_data)
{
	exit(0);
}

/* Colour getting function. */
int
GetDitheredColor(int red, int green, int blue, int x, int y)
{
	int	red_val, green_val, blue_val;
	int	red_int, green_int, blue_int;
	int	mod;

	red_val = floor(red * 0.082);
	green_val = floor(green * 0.082);
	blue_val = floor(blue * 0.082);

	mod = red_val % 4;
	red_int = red_val / 4;
	switch ( mod )
	{
		case 0:
			break;
		case 1:
			if ( x % 2 && y % 2 )
				red_int++;
			break;
		case 2:
			if ( x % 2 == y % 2 )
				red_int++;
			break;
		case 3:
			if ( x % 2 || y % 2 )
				red_int++;
	}

	mod = green_val % 4;
	green_int = green_val / 4;
	switch ( mod )
	{
		case 0:
			break;
		case 1:
			if ( x % 2 && y % 2 )
				green_int++;
			break;
		case 2:
			if ( x % 2 == y % 2 )
				green_int++;
			break;
		case 3:
			if ( x % 2 || y % 2 )
				green_int++;
	}

	mod = blue_val % 4;
	blue_int = blue_val / 4;
	switch ( mod )
	{
		case 0:
			break;
		case 1:
			if ( x % 2 && y % 2 )
				blue_int++;
			break;
		case 2:
			if ( x % 2 == y % 2 )
				blue_int++;
			break;
		case 3:
			if ( x % 2 || y % 2 )
				blue_int++;
	}

	return ((red_int * 6 + green_int) * 6 + blue_int);
}

