/*
**    Genray: A Generic Raytracer
**    Copyright (C) 1994  Stephen Chenney (stephen@cs.su.oz.au)
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program; if not, write to the Free Software
**    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/****************************************************************/
/*																*/
/*	Header file for raytrace.									*/
/*	Written by Stephen Chenney.  Dec 1993.						*/
/*																*/
/****************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <malloc.h>
#include <string.h>
#include <math.h>

#include "config.h"

#include "Vector.h"

#define TRUE 1
#define FALSE 0

/* Define a DELTA, used to allow rays to escape from surfaces. */
#define DELTA 1.e-6

/* Define an EPSILON, used to compare floats to 0.	*/
#define EPSILON 1.e-14

/* Macros for comparing doubles. */
#define dequal(a, b)    (((a) > (b) - EPSILON) && ((a) < (b) + EPSILON))
#define iszero(x)       ((x) < EPSILON && (x) > -EPSILON)


#define max(a,b)	((a) > (b) ? (a) : (b))
#define min(a,b)	((a) < (b) ? (a) : (b))


/* An linear transformation type, consisting of an affine	*/
/*  transformation matrix (for scaling, rotating etc) and a	*/
/*  displacement vector for positioning in space.		*/
typedef struct Transformation
	{
		Matrix	matrix;
		Vector	displacement;
	} Transformation;


/* A ray starts somewhere and goes in a certain direction.	*/
typedef struct Ray
	{
		Vector	start_point;
		Vector	direction;
	} Ray;


/* The various type of objects to raytrace.	*/
typedef enum ObjectType
	{
		sphere,
		cube,
		plane,
		square,
		cone,
		cylinder,
		torus,
	 	union_op,
		intersection_op,
		difference_op
	} ObjectType;

typedef int Colour;
#define ColMask	0x000000FF

/* A type to hold information about pixel mappings for surface colours.	*/
typedef struct PixelMap
	{
		short	width;
		short	height;
		Colour	*intensity;
	} PixelMap;

typedef struct Light
	{
		Vector	position;
		double	red;	/* Using doubles saves a couple of type	*/
		double	green;	/*  conversions and divides.			*/
		double	blue;
	} Light;

/* An attribute type.	*/
typedef struct Attributes
	{
		Colour		colour;			/* Obvious. */
		PixelMap	*pix_map;		/* The pixel mapping to use. NULL if none.*/
		double		diff_coef;		/* Diffuse lighting coeficient. */
		double		spec_coef;		/* Specular lighting coeficient. */
		double		spec_power;		/* Specular power. */
		double		reflect_coef;	/* Coeficient of reflection. */
		double		transparency;	/* Degree of transparency. */
		double		refr_index_sq;	/* Index of refraction squared. */
	} Attributes;

/* A Cuboid is the 3D equivalent of a Rectangle.	*/
typedef struct Cuboid
	{
		Vector min;
		Vector max;
	} Cuboid;

/* This instance definition comes straight from pg 693 of the book.	*/
typedef struct Instance
	{
		ObjectType		type;
		double			torus_rad;
		Transformation	transform;
		Transformation	inverse_trans;
		Attributes		nature;
		Cuboid			proj_extent;
		Cuboid			space_extent;
		struct Instance	*next;
		struct Instance *object1;
		struct Instance *object2;
	} Instance;


/* Need a list of instances for deriving hit normals for compound objects. */
typedef struct InstanceList
	{
		Instance	*obj;
		struct InstanceList	*next;
	} InstanceList;

/* A list of hit times for compound objects. */
typedef struct HitList
	{
		double  t_hit;          /* the hit time. */
		Vector  hit_normal;     /* The hit normal. */
		InstanceList   *path;  /* The path to follow to find the hit.  */
		struct HitList *next;
	} HitList;

/* Global variables. */
extern double	win_width;
extern double	win_height;
extern Vector	eye_position;
extern double	eyedist;
extern Matrix	convert_matrix;
extern Vector	view_reference;
extern Instance*	objects;
extern Light*	light_srcs;
extern int		light_number;
extern Colour	ambient_light;
extern unsigned long grey[];
extern char		do_colour;
extern char		silent;
extern char		write_tiff;
extern char		*out_filename;



/* Extern declarations. */
extern void Ray_Init(int, char**);
extern void Extents_Init(Vector*, Vector*);
extern int Ray_In_Extent(Ray, Instance*);
extern void Calculate_S_Bounds(int);

/* Intersection functions. */
extern char Ray_Sphere(Ray, Instance*, double*, Vector*, char*);
extern char Ray_Plane(Ray, Instance*, double*, Vector*, char*);
extern char Ray_Square(Ray, Instance*, double*, Vector*, char*);
extern char Ray_Cylinder(Ray, Instance*, double*, Vector*, char*);
extern char Ray_Cone(Ray, Instance*, double*, Vector*, char*);
extern char Ray_Cube(Ray, Instance*, double*, Vector*, char*);
extern char Ray_Torus(Ray, Instance*, double*, Vector*, char*);
extern
	char Ray_Compound(Ray, Instance*, double*, Vector*, char*, InstanceList**);

/* The shading function. */
extern Colour Find_Colour(Ray, Instance*, double, Vector, int, double, char,
							InstanceList*);

/* A shadowing function. */
extern char In_Shadow(Instance*, Ray);

/* The secondary ray caster. */
extern Colour RaytraceSecondary(Ray, Instance*, int, double);

/* A cubic equation solver. */
extern int SolveCubic(double c[3], double s[3]);
extern int SolveQuartic(double c[4], double s[4]);


/* My own memory management functions. */
extern	HitList*		get_hit_elmt();
extern	InstanceList*	get_inst_elmt();
extern	void			free_hit_elmt(HitList*);
extern	void			free_inst_elmt(InstanceList*);

