/*
**    Genray: A Generic Raytracer
**    Copyright (C) 1994  Stephen Chenney (stephen@cs.su.oz.au)
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program; if not, write to the Free Software
**    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/****************************************************************/
/*																*/
/*	File containing code to read a scene file for genray.		*/
/*	Also sets up any variables determined by such a file.		*/
/*	Written by Stephen Chenney.  Dec 1993.						*/
/*																*/
/****************************************************************/

#define USAGE "USAGE : genray [-geom widthxheight] [-ppm] [-s] <filename>\n"

#include "xstuff.h"

#include "genray.h"
#include "rayin.h"
#include <ctype.h>

#define printerror()	fprintf(stderr, "genray : error in file %s line %d : %s\n",\
								 input_file_name, yylineno, yytext)

#define read_float(x) \
	current_tok = yylex(); \
	switch (current_tok) \
	{ \
		case INT: \
			(x) = (double)lex_int; \
			break; \
		case FLOAT: \
			(x) = lex_float; \
			break; \
		default: \
			fprintf(stderr, "genray: ERROR file %s line %d : %s : float expected\n",\
							input_file_name, yylineno, yytext); \
			exit(1); \
	}

typedef struct NamedList {
	char				*name;
	Instance			*obj;
	struct NamedList	*next;
	} NamedList;

char	*file_dir;

input_token		current_tok;

NamedList	*named_list = NULL;

Vector	look_from = {10, 0, 0}; /* Various vectors with usual meanings. */
Vector	look_at = {0, 0, 0};
Vector	look_up = {0, 0, 1};
Vector	u;
Vector	v;
Vector	n;

char	have_win = FALSE;


Colour		def_col = 0x00FFFFFF;	/* Attribute defaults. */
PixelMap	*def_mapping = NULL;
double		def_diff_c = 1.0;
double		def_spec_c = 0.0;
double		def_refl_c = 0.0;
double		def_transp = 0.0;
double		def_refr_c = 1.0;

Vector		col_vect;	/* A vector used to hold rgb colour info. */

/* Fallback resources */
String	fallbacks[] = {
	"Raytrace.geometry: 400x400",
	"Raytrace.background: black",
	"Raytrace.viewWindow.background: black",
	NULL
	};

String	translations =
		"<Btn1Down>: XawPositionSimpleMenu(Menu1) XtMenuPopup(Menu1)";


/* Functions. */
void		read_named();
Instance	*create_instance(char *name);
Instance	*read_instance(ObjectType obj);
Instance	*read_composite(ObjectType obj);
void		read_light();
void		read_defs();
PixelMap*	read_mapping(char *file_name);
void		init_vars();
void		print_objects(Instance *start);
void		InitColRead(Vector *v);

extern void Compound_Extent(Instance *obj, Vector* w_to_v, Vector disp);
extern void Square_Extent(Instance *obj, Vector* w_to_v, Vector disp);
extern void Cube_Extent(Instance *obj, Vector* w_to_v, Vector disp);
extern void Pyramid_Extent(Instance *obj, Vector* w_to_v, Vector disp);
extern void Torus_Extent(Instance *obj, Vector* w_to_v, Vector disp);


/*
**	InitVRead: Reads in a vector.
*/
void
InitVRead(Vector *v)
{
	read_float(v->x);
	read_float(v->y);
	read_float(v->z);
}


/*
**	InitColRead: Reads in a colour.
*/
void
InitColRead(Vector *v)
{
	InitVRead(v);

	if ((v->x > 1.0) || (v->y > 1.0) || (v->z > 1.0))
	{
		fprintf(stderr, "genray: ERROR in file %s line %d : colour out of range\n",
						input_file_name, yylineno);
		exit(1);
	}
}

void
Ray_Init(int argc, char* argv[])
/****************************************************************/
/*																*/
/*	Opens the input file, reads what's in it then assigns		*/
/*	the global variables.										*/
/*	On error returns FALSE, TRUE otherwise.						*/
/*																*/
/****************************************************************/
{
	Instance	*current;
	short		scr_width, scr_height;	/* Width and height for the screen. */
	int			i, j;
	char		*temp_arg = NULL;
	char		*temp_c;
	Arg			args[3];
	int			screen_depth;
	Boolean		use_stdin = FALSE;
	char		command_string[128];

	/* Check for (probable) correct usage. */
	if (argc < 2)
	{
		fprintf(stderr, USAGE);
		exit(1);
	}

	/* Parse the command line. */
	for ( i = 1 ; i < argc - 1; i++ )
	{
		if (argv[i][0] == '-')
		{
			if ( argv[i][1] == '-' )
			{
				use_stdin = TRUE;
			}
			else if (argv[i][1] == 's')
			{
				silent = TRUE;
			}
			else if (argv[i][1] == 'p')
			{
				write_tiff = FALSE;
			}
			else if ((argv[i][1] != 'g') && (argv[i][1] != 'd'))
			{
				fprintf(stderr, USAGE);
				exit(1);
			}
			else
			{
				/* Skip over anything else. */
				i++;
			}
		}
		else
		{
			fprintf(stderr, USAGE);
			exit(1);
		}

	}

	if (silent)
	{
		/* Set default screen size. */
		screen_rect.x = screen_rect.y = 0;
		screen_rect.width = screen_rect.height = 400;
	}

	/* Try to open the file for reading. */
	if ( use_stdin )
		sprintf(command_string, CPP);
	else
	{
		input_file_name = argv[i];
		sprintf(command_string, CPP" < %s", argv[i]);
	}

	if ((yyin = popen(command_string, "r")) == NULL)
	{
		fprintf(stderr, "genray : Cannot open file %s\n", argv[i]);
		exit(1);
	}

	/* Set up output file name. */
	out_filename = malloc(strlen(argv[i]) + 8);
	strcpy(out_filename, argv[i]);
	/* Look for a .gen suffix in the filename. */
	temp_c = strrchr(out_filename, '.');
	if ( temp_c == NULL )
	{
		if (write_tiff)
			sprintf(out_filename, "%s.tiff", argv[i]);
		else
			sprintf(out_filename, "%s.ppm", argv[i]);
	}
	else
	{
		temp_c[1] = '\0';
		if (write_tiff)
			strcat(out_filename, "tiff");
		else
			strcat(out_filename, "ppm");
	}

	/* Set up the input file directory. */
	/* This directory will be searched for include files and pixmaps. */
	file_dir = strdup(argv[i]);
	if ((temp_c = strrchr(file_dir, '/')) != NULL)
	{
		*temp_c = '\0';
	}
	else
		file_dir = ".";


	while ((current_tok = yylex()) != EOF_TOKEN)
	{
		switch (current_tok)
		{
			case LOOKFROM:
				/* Look_from viewpoint vector. */
				InitVRead(&look_from);
				break;

			case LOOKAT:
				/* Look_at viewpoint vector. */
				InitVRead(&look_at);
				break;

			case LOOKUP:
				/* Look_up viewpoint vector. */
				InitVRead(&look_up);
				break;

			case EYEDIST:
				/* Distance from eye to viewplane. */
				read_float(eyedist);
				break;

			case WINDOW:
				/* Size of viewing window. */
				read_float(win_width);
				read_float(win_height);
				have_win = TRUE;
				break;

			case SCREEN:
				/* Actual screen size. */
				current_tok = yylex();
				switch (current_tok)
				{
					case INT:
						scr_width = lex_int;
						break;
					case FLOAT:
						scr_width = (int)lex_float;
						break;
					default:
						fprintf(stderr, "genray: ERROR file %s line %d : %s : number expected\n",
										input_file_name, yylineno, yytext);
						exit(1);
				}
				current_tok = yylex();
				switch (current_tok)
				{
					case INT:
						scr_height = lex_int;
						break;
					case FLOAT:
						scr_height = (int)lex_float;
						break;
					default:
						fprintf(stderr, "genray: ERROR file %s line %d : %s : number expected\n",
										input_file_name, yylineno, yytext);
						exit(1);
				}

				if (silent)
				{
					/* Just reset the screen size.*/
					screen_rect.width = scr_width;
					screen_rect.height = scr_height;
				}
				else
				{
					/* Set fallback resources. The command line will override.*/
					fallbacks[0] = malloc(50);
					sprintf(fallbacks[0], "Raytrace.geometry: %0dx%0d",
											scr_width, scr_height);
				}
				break;

			case AMBIENT:
				InitColRead(&col_vect);
				ambient_light = 0;
				ambient_light |= ((int)(col_vect.x * 255)) & ColMask;
				ambient_light |= (((int)(col_vect.y * 255)) & ColMask) << 8;
				ambient_light |= (((int)(col_vect.z * 255)) & ColMask) << 16;
				break;

			case LIGHT:
				/* Light source information. */
				read_light();
				break;

			case SET:
				read_defs();
				break;

			case NAME:
				read_named();
				break;

			case INSTANCE: {
				Instance *new_inst;
				if ((current_tok = yylex()) != STRING)
				{
					fprintf(stderr, "genray: ERROR file %s line %d : %s : string expected\n",
							input_file_name, yylineno, yytext);
					break;
				}
				new_inst = create_instance(lex_string);
				new_inst->next = objects;
				objects = new_inst;
				break;
				}

			case UNION:
				current = read_composite(union_op);
				if (current != NULL)
				{
					current->next = objects;
					objects = current;
				}
				break;

			case INTERSECTION:
				current = read_composite(intersection_op);
				if (current != NULL)
				{
					current->next = objects;
					objects = current;
				}
				break;

			case DIFFERENCE:
				current = read_composite(difference_op);
				if (current != NULL)
				{
					current->next = objects;
					objects = current;
				}
				break;

			case SPHERE:
				current = read_instance(sphere);
				if (current)
				{
					current->next = objects;
					objects = current;
				}
				break;

			case CUBE:
				current = read_instance(cube);
				if (current)
				{
					current->next = objects;
					objects = current;
				}
				break;

			case CYLINDER:
				current = read_instance(cylinder);
				if (current)
				{
					current->next = objects;
					objects = current;
				}
				break;

			case CONE:
				current = read_instance(cone);
				if (current)
				{
					current->next = objects;
					objects = current;
				}
				break;

			case PLANE:
				current = read_instance(plane);
				if (current)
				{
					current->next = objects;
					objects = current;
				}
				break;

			case SQUARE:
				current = read_instance(square);
				if (current)
				{
					current->next = objects;
					objects = current;
				}
				break;

			case TORUS:
				current = read_instance(torus);
				if (current)
				{
					current->next = objects;
					objects = current;
				}
				break;

			default:
				printerror();
		}

	}

	if (!silent)
	{
		top_level = XtAppInitialize(&app_context, "Raytrace",
									(XrmOptionDescList)NULL, 0,
									&argc, argv, fallbacks, NULL, 0);

		view_window = XtCreateManagedWidget("viewWindow", viewWidgetClass,
											top_level, NULL, 0);

		/* Set up the menu. */
		XtSetArg(args[0], XtNbackground, XWhitePixel(XtDisplay(top_level),
									XDefaultScreen(XtDisplay(top_level))));
		menu_widget = XtCreatePopupShell("Menu1", simpleMenuWidgetClass,
									 top_level, args, 1);

		menu_children[0] = XtCreateManagedWidget("SAVE", smeBSBObjectClass,
											menu_widget, args, 1);
		menu_children[1] = XtCreateManagedWidget("QUIT", smeBSBObjectClass,
											menu_widget, args, 1);
	
		/* Set up the callbacks. */
		XtAddCallback(menu_children[0], XtNcallback, Save_Function, NULL);
		XtAddCallback(menu_children[1], XtNcallback, Exit_Function, NULL);

		/* Modify the translation table for the top_level widget. */
		XtOverrideTranslations(view_window,
									XtParseTranslationTable(translations));

		XtRealizeWidget(top_level);
		XFlush(XtDisplay(top_level));

		/* Get the screen data. */
		screen_rect.x = screen_rect.y = 0;
		j = 0;
		XtSetArg(args[j], XtNdepth, &screen_depth);			j++;
		XtSetArg(args[j], XtNwidth, &screen_rect.width);	j++;
		XtSetArg(args[j], XtNheight, &screen_rect.height);	j++;
		XtGetValues(top_level, args, j);
		do_colour = (screen_depth == 8) ? TRUE : FALSE;

		/* Define and set the pixmap to draw into. */
		drawing_pixmap = XCreatePixmap( XtDisplay(top_level),
										XtWindow(top_level),
										screen_rect.width,
										screen_rect.height,
										screen_depth);
	}

	init_vars();

	return;

}


void
read_named()
{
	Instance	*current = NULL;
	NamedList	*named_obj;

	if ((current_tok = yylex()) != STRING)
	{
		fprintf(stderr, "genray: ERROR file %s line %d : %s : string expected\n",
				input_file_name, yylineno, yytext);
		return;
	}

	named_obj = (NamedList*)malloc(sizeof(NamedList));
	named_obj->name = strdup(lex_string);

	current_tok = yylex();
	switch (current_tok)
	{
		case UNION:
			current = read_composite(union_op);
			break;

		case INTERSECTION:
			current = read_composite(intersection_op);
			break;

		case DIFFERENCE:
			current = read_composite(difference_op);
			break;

		case SPHERE:
			current = read_instance(sphere);
			break;

		case CUBE:
			current = read_instance(cube);
			break;

		case CYLINDER:
			current = read_instance(cylinder);
			break;

		case CONE:
			current = read_instance(cone);
			break;

		case PLANE:
			current = read_instance(plane);
			break;

		case SQUARE:
			current = read_instance(square);
			break;

		case TORUS:
			current = read_instance(torus);
			break;

		default:
			printerror();
	}

	if (current)
	{
		named_obj->obj = current;
		named_obj->next = named_list;
		named_list = named_obj;
	}
	else
	{
		free(named_obj->name);
		free(named_obj);
	}

	return;
}



Instance*
create_instance(char *name)
{
	NamedList	*named_obj;
	Instance	*new_inst;
	Instance	*extra_stuff;

	for ( named_obj = named_list ;
		  (named_obj != NULL) && (strcmp(named_obj->name, name)) ;
		  named_obj = named_obj->next );

  	if (named_obj == NULL)
  	{
		fprintf(stderr, "genray: ERROR file %s line %d : %s : unknown name\n",
				input_file_name, yylineno, name);
  		return;
  	}

  	extra_stuff = read_instance(named_obj->obj->type);

	if ((new_inst = (Instance*)malloc(sizeof(Instance))) == NULL)
	{
		fprintf(stderr, "genray : Unable to allocate memory.\n");
		exit(1);
	}

	new_inst->type = named_obj->obj->type;

	MMMul(extra_stuff->transform.matrix, named_obj->obj->transform.matrix,
		  new_inst->transform.matrix);
  	VAdd(extra_stuff->transform.displacement,
  		 named_obj->obj->transform.displacement,
  		 new_inst->transform.displacement);

	MInvert(new_inst->transform.matrix, new_inst->inverse_trans.matrix);
	VScalarMul(new_inst->transform.displacement, -1,
				new_inst->inverse_trans.displacement);

	new_inst->torus_rad =
		(extra_stuff->torus_rad == 2.0) ?
		named_obj->obj->torus_rad :
		extra_stuff->torus_rad;
	new_inst->nature.colour =
		(extra_stuff->nature.colour == def_col) ?
		named_obj->obj->nature.colour :
		extra_stuff->nature.colour;
	new_inst->nature.pix_map =
		(extra_stuff->nature.pix_map == def_mapping) ?
		named_obj->obj->nature.pix_map :
		extra_stuff->nature.pix_map;
	new_inst->nature.diff_coef =
		(extra_stuff->nature.diff_coef == def_diff_c) ?
		named_obj->obj->nature.diff_coef :
		extra_stuff->nature.diff_coef;
	new_inst->nature.spec_coef =
		(extra_stuff->nature.spec_coef == def_spec_c) ?
		named_obj->obj->nature.spec_coef :
		extra_stuff->nature.spec_coef;
	new_inst->nature.spec_power =
		(extra_stuff->nature.spec_power == def_spec_c * 100) ?
		named_obj->obj->nature.spec_power :
		extra_stuff->nature.spec_power;
	new_inst->nature.reflect_coef =
		(extra_stuff->nature.reflect_coef == def_refl_c) ?
		named_obj->obj->nature.reflect_coef :
		extra_stuff->nature.reflect_coef;
	new_inst->nature.refr_index_sq =
		(extra_stuff->nature.refr_index_sq == def_refr_c) ?
		named_obj->obj->nature.refr_index_sq :
		extra_stuff->nature.refr_index_sq;
	new_inst->nature.transparency =
		(extra_stuff->nature.transparency == def_transp) ?
		named_obj->obj->nature.transparency :
		extra_stuff->nature.transparency;

	new_inst->object1 = named_obj->obj->object1;
	new_inst->object2 = named_obj->obj->object2;

	return new_inst;

}




Instance*
read_instance(ObjectType obj)
/****************************************************************/
/*																*/
/*	Reads in the data for an object and puts it on the list		*/
/*	of objects.													*/
/*																*/
/****************************************************************/
{
	Instance	*current;
	Vector		temp_disp;
	Matrix		temp_mat;
	Matrix		trans_mat;
	Vector		temp_v1;
	Vector		temp_v2;
	Vector		temp_v3;
	Matrix		temp_m;
	input_token	rot_axis;
	double		angle;
	double		sin_a;
	double		cos_a;
	double		tor_rad = 2.0;	/* The mean radius for a torus. */
	Colour		col = def_col;	/* Attribute defaults. */
	PixelMap	*mapping = def_mapping;
	double		diff_c = def_diff_c;
	double		spec_c = def_spec_c;
	double		refl_c = def_refl_c;
	double		transp = def_transp;
	double		refr_c = def_refr_c;

	if ((current = (Instance*)malloc(sizeof(Instance))) == NULL)
	{
		fprintf(stderr, "genray : Unable to allocate memory.\n");
		exit(1);
	}

	/* The initial displacement is <0,0,0> */
	VNew(0, 0, 0, temp_disp);

	/* The initial transformation matrix is the identity. */
	VNew(1, 0, 0, temp_v1);
	VNew(0, 1, 0, temp_v2);
	VNew(0, 0, 1, temp_v3);
	MNew(temp_v1, temp_v2, temp_v3, temp_mat);

	/* Read in a complete instance. */
	while ((current_tok = yylex()) != EOF_TOKEN)
	{
		if (current_tok == END) break;

		switch (current_tok)
		{
			case COLOUR:
				InitColRead(&col_vect);
				col = 0;
				col |= ((int)(col_vect.x * 255)) & ColMask;
				col |= (((int)(col_vect.y * 255)) & ColMask) << 8;
				col |= (((int)(col_vect.z * 255)) & ColMask) << 16;
				break;

			case PIXMAP:
				if ((current_tok = yylex()) != STRING)
				{
					printerror();
					exit(1);
				}

				if ((obj != square) && (obj != plane))
				{
					fprintf(stderr, "genray: ERROR: pixmap only for squares or planes, line %d",
						yylineno);
					break;
				}

				mapping = read_mapping(lex_string);
				break;

			case DIFFUSE:
				read_float(diff_c);
				break;

			case SPECULAR:
				read_float(spec_c);
				break;

			case REFLECT:
				read_float(refl_c);
				break;

			case REFRACT:
				read_float(refr_c);
				break;

			case TRANSPARENCY:
				read_float(transp);
				break;

			case TRANSLATE:
				InitVRead(&temp_v1);
				VAdd(temp_v1, temp_disp, temp_disp);
				break;

			case SCALE:
				InitVRead(&temp_v1);
				/* Use a shortcut. */
				/* This is equivalent to multiplying temp_mat */
				/*  by the scaling matrix implied by temp_v1. */
				VScalarMul(temp_mat[0], temp_v1.x, temp_mat[0]);
				VScalarMul(temp_mat[1], temp_v1.y, temp_mat[1]);
				VScalarMul(temp_mat[2], temp_v1.z, temp_mat[2]);
				break;

			case ROTATE:
				rot_axis = yylex();
				read_float(angle);

				/* Transform to radians. */
				angle *= M_PI / 180.0;
				sin_a = sin(angle);
				cos_a = cos(angle);

				switch (rot_axis)
				{
					case X_AXIS:
						/* rotation about x. */
						VNew(1, 0, 0, temp_v1);
						VNew(0, cos_a, -sin_a, temp_v2);
						VNew(0, sin_a, cos_a, temp_v3);
						MNew(temp_v1, temp_v2, temp_v3, temp_m);
						MMMul(temp_m, temp_mat, temp_mat);
						break;

					case Y_AXIS:
						/* rotation about y. */
						VNew(cos_a, 0, sin_a, temp_v1);
						VNew(0, 1, 0, temp_v2);
						VNew(-sin_a, 0, cos_a, temp_v3);
						MNew(temp_v1, temp_v2, temp_v3, temp_m);
						MMMul(temp_m, temp_mat, temp_mat);
						break;

					case Z_AXIS:
						/* rotation about z. */
						VNew(cos_a, -sin_a, 0, temp_v1);
						VNew(sin_a, cos_a, 0, temp_v2);
						VNew(0, 0, 1, temp_v3);
						MNew(temp_v1, temp_v2, temp_v3, temp_m);
						MMMul(temp_m, temp_mat, temp_mat);
						break;

					default :
						fprintf(stderr, "genray: ERROR file %s line %d : %s : axis expected\n",
										input_file_name, yylineno, yytext);
						exit(1);
				}
				break;

			case TRANSFORM:
				InitVRead(trans_mat);
				InitVRead(trans_mat + 1);
				InitVRead(trans_mat + 2);
				MMMul(trans_mat, temp_mat, temp_mat);
				break;

			case RADIUS:
				if (obj != torus)
				{
					fprintf(stderr, "genray: ERROR file %s line %d : radius only for torus\n",
										input_file_name, yylineno);
					break;
				}
				read_float(tor_rad);
				break;

			default:
				printerror();
		}
		
	}

	/* All data has been read, now store it. */
	current->type = obj;

	(current->transform).matrix[0] = temp_mat[0];
	(current->transform).matrix[1] = temp_mat[1];
	(current->transform).matrix[2] = temp_mat[2];
	(current->transform).displacement = temp_disp;

	/* Calculate the inverse transformation. */
	if (MInvert(temp_mat, (current->inverse_trans).matrix) == NULL)
	{
		fprintf(stderr, "genray : ERROR : file %s singular transformation around line %d\n",
			input_file_name, yylineno);
		return NULL;	/* Don't add it to the object list. */
	}
	VScalarMul(temp_disp, -1, (current->inverse_trans).displacement);

	current->torus_rad = tor_rad;
	current->nature.colour = col;
	current->nature.pix_map = mapping;
	current->nature.diff_coef = diff_c;
	current->nature.spec_coef = spec_c;
	current->nature.spec_power = spec_c * 100;
	current->nature.reflect_coef = refl_c;
	current->nature.refr_index_sq = refr_c * refr_c;
	current->nature.transparency = transp;

	/* The other fields will be assigned later, */

	return current;

}


Instance*
read_composite(ObjectType obj)
/****************************************************************/
/*																*/
/*	Reads in a composite object.								*/
/*	Fills in whatever fields the uses specifies.				*/
/*	This function is sort of a mix between raytrace_init and	*/
/*	read_instance.												*/
/*																*/
/****************************************************************/
{
	Instance	*current;
	Instance	*operand = NULL;

	Vector		temp_disp;
	Matrix		temp_mat;
	Matrix		trans_mat;
	Vector		temp_v1;
	Vector		temp_v2;
	Vector		temp_v3;
	Matrix		temp_m;

	input_token	rot_axis;
	double		angle;
	double		sin_a;
	double		cos_a;

	Colour		col = def_col;	/* Attribute defaults. */
	PixelMap	*mapping = def_mapping;
	double		tor_rad = 2.0;	/* The mean radius for a torus. */
	double		diff_c = def_diff_c;
	double		spec_c = def_spec_c;
	double		refl_c = def_refl_c;
	double		transp = def_transp;
	double		refr_c = def_refr_c;

	char		have_left = FALSE;
	char		have_right = FALSE;


#define	fill_field(op)	if (!have_left)\
						{ \
							current->object1 = op; \
							have_left = TRUE; \
						} \
						else if (!have_right) \
						{ \
							current->object2 = op; \
							have_right = TRUE; \
						} \
						else \
							printerror(); \
						op->next = NULL

	if ((current = (Instance*)malloc(sizeof(Instance))) == NULL)
	{
		fprintf(stderr, "genray : Unable to allocate memory.\n");
		exit(1);
	}

	/* The initial displacement is <0,0,0> */
	VNew(0, 0, 0, temp_disp);

	/* The initial transformation matrix is the identity. */
	VNew(1, 0, 0, temp_v1);
	VNew(0, 1, 0, temp_v2);
	VNew(0, 0, 1, temp_v3);
	MNew(temp_v1, temp_v2, temp_v3, temp_mat);

	/* Read in a complete instance. */
	while ((current_tok = yylex()) != EOF_TOKEN)
	{
		if (current_tok == END) break;

		switch (current_tok)
		{
			case SET:
				read_defs();
				break;

			case INSTANCE:
				if ((current_tok = yylex()) != STRING)
				{
					fprintf(stderr, "genray: ERROR file %s line %d : %s : string expected\n",
							input_file_name, yylineno, yytext);
					break;
				}
				operand = create_instance(lex_string);
				fill_field(operand);
				break;

			case UNION:
				operand = read_composite(union_op);
				fill_field(operand);
				break;

			case INTERSECTION:
				operand = read_composite(intersection_op);
				fill_field(operand);
				break;

			case DIFFERENCE:
				operand = read_composite(difference_op);
				fill_field(operand);
				break;

			case SPHERE:
				operand = read_instance(sphere);
				fill_field(operand);
				break;

			case CUBE:
				operand = read_instance(cube);
				fill_field(operand);
				break;

			case CYLINDER:
				operand = read_instance(cylinder);
				fill_field(operand);
				break;

			case CONE:
				operand = read_instance(cone);
				fill_field(operand);
				break;

			case PLANE:
				operand = read_instance(plane);
				fill_field(operand);
				break;

			case SQUARE:
				operand = read_instance(square);
				fill_field(operand);
				break;

			case TORUS:
				operand = read_instance(torus);
				fill_field(operand);
				break;

			case COLOUR:
				InitColRead(&col_vect);
				col = 0;
				col |= ((int)(col_vect.x * 255)) & ColMask;
				col |= (((int)(col_vect.y * 255)) & ColMask) << 8;
				col |= (((int)(col_vect.z * 255)) & ColMask) << 16;
				break;

			case DIFFUSE:
				read_float(diff_c);
				break;

			case SPECULAR:
				read_float(spec_c);
				break;

			case REFLECT:
				read_float(refl_c);
				break;

			case REFRACT:
				read_float(refr_c);
				break;

			case TRANSPARENCY:
				read_float(transp);
				break;

			case TRANSLATE:
				InitVRead(&temp_v1);
				VAdd(temp_v1, temp_disp, temp_disp);
				break;

			case SCALE:
				InitVRead(&temp_v1);
				/* Use a shortcut. */
				/* This is equivalent to multiplying temp_mat */
				/*  by the scaling matrix implied by temp_v1. */
				VScalarMul(temp_mat[0], temp_v1.x, temp_mat[0]);
				VScalarMul(temp_mat[1], temp_v1.y, temp_mat[1]);
				VScalarMul(temp_mat[2], temp_v1.z, temp_mat[2]);
				break;

			case ROTATE:
				rot_axis = yylex();
				read_float(angle);

				/* Transform to radians. */
				angle *= M_PI / 180.0;
				sin_a = sin(angle);
				cos_a = cos(angle);

				switch (rot_axis)
				{
					case X_AXIS:
						/* rotation about x. */
						VNew(1, 0, 0, temp_v1);
						VNew(0, cos_a, -sin_a, temp_v2);
						VNew(0, sin_a, cos_a, temp_v3);
						MNew(temp_v1, temp_v2, temp_v3, temp_m);
						MMMul(temp_m, temp_mat, temp_mat);
						break;

					case Y_AXIS:
						/* rotation about y. */
						VNew(cos_a, 0, sin_a, temp_v1);
						VNew(0, 1, 0, temp_v2);
						VNew(-sin_a, 0, cos_a, temp_v3);
						MNew(temp_v1, temp_v2, temp_v3, temp_m);
						MMMul(temp_m, temp_mat, temp_mat);
						break;

					case Z_AXIS:
						/* rotation about z. */
						VNew(cos_a, -sin_a, 0, temp_v1);
						VNew(sin_a, cos_a, 0, temp_v2);
						VNew(0, 0, 1, temp_v3);
						MNew(temp_v1, temp_v2, temp_v3, temp_m);
						MMMul(temp_m, temp_mat, temp_mat);
						break;

					default :
						fprintf(stderr, "genray: ERROR file %s line %d : %s : axis expected\n",
										input_file_name, yylineno, yytext);
						exit(1);
				}
				break;

			case TRANSFORM:
				InitVRead(trans_mat);
				InitVRead(trans_mat + 1);
				InitVRead(trans_mat + 2);
				MMMul(trans_mat, temp_mat, temp_mat);
				break;

			case RADIUS:
				if (obj != torus)
				{
					fprintf(stderr, "genray: ERROR file %s line %d : radius only for torus\n",
										input_file_name, yylineno);
					break;
				}
				read_float(tor_rad);
				break;

			default:
				printerror();

		}

	}

	/* All data has been read, now store it. */
	current->type = obj;

	(current->transform).matrix[0] = temp_mat[0];
	(current->transform).matrix[1] = temp_mat[1];
	(current->transform).matrix[2] = temp_mat[2];
	(current->transform).displacement = temp_disp;

	/* Calculate the inverse transformation. */
	if (MInvert(temp_mat, (current->inverse_trans).matrix) == NULL)
	{
		fprintf(stderr, "genray : ERROR : singular transformation around line %d\n",
			yylineno);
		/* Give it the identity transform. */
		VNew(1,0,0,temp_v1);
		VNew(0,1,0,temp_v2);
		VNew(0,0,1,temp_v3);
		MNew(temp_v1, temp_v2, temp_v3, temp_mat);
	}
	VScalarMul(temp_disp, -1, (current->inverse_trans).displacement);

	current->torus_rad = tor_rad;
	current->nature.colour = col;
	current->nature.pix_map = mapping;
	current->nature.diff_coef = diff_c;
	current->nature.spec_coef = spec_c;
	current->nature.spec_power = spec_c * 100;
	current->nature.reflect_coef = refl_c;
	current->nature.refr_index_sq = refr_c * refr_c;
	current->nature.transparency = transp;

	/* The other fields will be assigned later, */

	return current;

}


void
read_light()
/****************************************************************/
/*																*/
/*	Reads in a light source and puts it in the array of			*/
/*	 light sources.												*/
/*																*/
/****************************************************************/
{
	char	have_col = FALSE;
	char	have_posn = FALSE;
	Light	temp_l;

	/* read in a light position and intensity.	*/
	while ((current_tok = yylex()) != EOF_TOKEN)
	{
		if (current_tok == END) break;

		if (current_tok == POSITION)
		{
			InitVRead(&(temp_l.position));
			have_posn = TRUE;
		}
		else if (current_tok == COLOUR)
		{
			InitColRead(&col_vect);
			temp_l.red = col_vect.x;
			temp_l.green = col_vect.y;
			temp_l.blue = col_vect.z;
			have_col = TRUE;
		}
		else
			printerror();

	}

	if (!have_col || !have_posn)
	{
		fprintf(stderr, "genray : error : file %s line %d : missing light information\n",
						input_file_name, yylineno);
		return;
	}

	if (light_number++ == 0)
		light_srcs = (Light*)malloc(sizeof(Light));
	else
		light_srcs = (Light*)realloc(light_srcs, sizeof(Light) * light_number);

	light_srcs[light_number - 1] = temp_l;

	return;

}


void
read_defs()
/****************************************************************/
/*																*/
/*	Reads in new default properties for all objects.			*/
/*																*/
/****************************************************************/
{
	while ((current_tok = yylex()) != EOF_TOKEN)
	{
		if (current_tok == END) break;

		switch (current_tok)
		{
			case COLOUR:
				InitColRead(&col_vect);
				def_col = 0;
				def_col |= ((int)(col_vect.x * 255)) & ColMask;
				def_col |= (((int)(col_vect.y * 255)) & ColMask) << 8;
				def_col |= (((int)(col_vect.z * 255)) & ColMask) << 16;
				break;

			case PIXMAP:
				if ((current_tok = yylex()) != STRING)
				{
					printerror();
					exit(1);
				}
				def_mapping = read_mapping(lex_string);
				break;

			case DIFFUSE:
				read_float(def_diff_c);
				break;

			case SPECULAR:
				read_float(def_spec_c);
				break;

			case REFLECT:
				read_float(def_refl_c);
				break;

			case REFRACT:
				read_float(def_refr_c);
				break;

			case TRANSPARENCY:
				read_float(def_transp);
				break;

			default:
				printerror();

		}
		
	}
}


PixelMap*
read_mapping(char *file_name)
/****************************************************************/
/*																*/
/*	Reads in a ppm file to use as a pixel mapping for surface	*/
/*	 colours.  Returns the map or NULL on error.				*/
/*																*/
/****************************************************************/
{
	FILE		*map_file;
	char		*pix_file;
	short		pixwidth, pixheight;
	short		max_val;
	PixelMap	*result;
	int			ch;
	int			i, j;
	register int	base;

	/* Prepend the file_directory. */
	pix_file = malloc(strlen(file_dir) + strlen(file_name) + 5);
	sprintf(pix_file, "%s/%s", file_dir, file_name);


	if ((map_file = fopen(pix_file, "r")) == NULL)
	{
		fprintf(stderr, "genray : ERROR : Unable to open pixmap file %s\n",
				pix_file);
		return NULL;
	}

	if ((fgetc(map_file) != 'P') || (fgetc(map_file) != '6'))
	{
		fprintf(stderr, "genray : ERROR : %s not in raw ppm format\n", file_name);
		return NULL;
	}

	/* Find width. */
	ch = fgetc(map_file);
	pixwidth = 0;
	while (TRUE)
	{
		while (isspace(ch))
			ch = fgetc(map_file);
		if (ch == EOF)
		{
			fprintf(stderr, "genray : ERROR : %s not in raw ppm format - no width\n",
					file_name);
			return NULL;
		}
		if (ch == '#')
		{
			while ((ch = fgetc(map_file)) != '\n');
		}
		else
		{
			if (!isdigit(ch))
			{
				fprintf(stderr, "genray : ERROR : %s not in raw ppm format - extra chars\n",
					file_name);
				return NULL;
			}
			else
			{
				while (isdigit(ch))
				{
					pixwidth = pixwidth * 10 + ch - '0';
					ch = fgetc(map_file);
				}
				break;
			}
		}
	}


	/* find height. */
	ch = fgetc(map_file);
	pixheight = 0;
	while (TRUE)
	{
		while (isspace(ch))
			ch = fgetc(map_file);
		if (ch == EOF)
		{
			fprintf(stderr, "genray : ERROR : %s not in raw ppm format - no width\n",
					file_name);
			return NULL;
		}
		if (ch == '#')
		{
			while ((ch = fgetc(map_file)) != '\n');
		}
		else
		{
			if (!isdigit(ch))
			{
				fprintf(stderr, "genray : ERROR : %s not in raw ppm format - extra chars\n",
					file_name);
				return NULL;
			}
			else
			{
				while (isdigit(ch))
				{
					pixheight = pixheight * 10 + ch - '0';
					ch = fgetc(map_file);
				}
				break;
			}
		}
	}


	/* Find the max_val. */
	ch = fgetc(map_file);
	max_val = 0;
	while (TRUE)
	{
		while (isspace(ch))
			ch = fgetc(map_file);
		if (ch == EOF)
		{
			fprintf(stderr, "genray : ERROR : %s not in raw ppm format - no width\n",
					file_name);
			return NULL;
		}
		if (ch == '#')
		{
			while ((ch = fgetc(map_file)) != '\n');
		}
		else
		{
			if (!isdigit(ch))
			{
				fprintf(stderr, "genray : ERROR : %s not in raw ppm format - extra chars\n",
					file_name);
				return NULL;
			}
			else
			{
				while (isdigit(ch))
				{
					max_val = max_val * 10 + ch - '0';
					ch = fgetc(map_file);
				}
				break;
			}
		}
	}


	/* Find the data. */
	/* Assuming the data starts on the next line. */
	while (ch != '\n')
	{
		if (ch == EOF)
		{
			fprintf(stderr, "genray : ERROR : %s not in raw ppm format - no data\n",
				file_name);
			return NULL;
		}
		ch = fgetc(map_file);
	}

	/* Allocate all the space. */
	if ((result = (PixelMap*)malloc(sizeof(PixelMap))) == NULL)
	{
		fprintf(stderr, "genray  : TERMINATING : Unable to allocate PixelMap space\n");
		exit(1);
	}

	if ((result->intensity = (Colour*)malloc(sizeof(Colour) * pixwidth * pixheight)) == NULL)
	{
		fprintf(stderr, "genray  : TERMINATING : Unable to allocate PixelMap space\n");
		exit(1);
	}

	result->width = pixwidth;
	result->height = pixheight;

	for ( j = 0 ; j < pixheight ; j++ )
	{
		base = j * pixwidth;
		for ( i = 0 ; i < pixwidth ; i++ )
		{
			if ((ch = fgetc(map_file)) == EOF)
			{
				fprintf(stderr, "genray : ERROR : %s not in raw ppm format - missing data\n",
					file_name);
				free(result->intensity);
				free(result);
				return NULL;
			}
			result->intensity[base+i] = 0;
			result->intensity[base+i] |=
										((int)(ch * 255 / max_val)) & ColMask;
			if ((ch = fgetc(map_file)) == EOF)
			{
				fprintf(stderr, "genray : ERROR : %s not in raw ppm format - missing data\n",
					file_name);
				free(result->intensity);
				free(result);
				return NULL;
			}
			result->intensity[base+i] |=
								(((int)(ch * 255 / max_val)) & ColMask) << 8;
			if ((ch = fgetc(map_file)) == EOF)
			{
				fprintf(stderr, "genray : ERROR : %s not in raw ppm format - missing data\n",
					file_name);
				free(result->intensity);
				free(result);
				return NULL;
			}
			result->intensity[base+i] |=
								(((int)(ch * 255 / max_val)) & ColMask) << 16;
		}
	}

	fclose(map_file);

	return result;

}


void
init_vars()
/****************************************************************/
/*																*/
/*	Initialises most global variables.  In particular those		*/
/*	to do with the transformations from and to view co-ords.	*/
/*																*/
/****************************************************************/
{
	double	temp_d;
	Vector	temp_v;

	/* First get u, v, n */

	/* n = (norm)(look_at - look_from) */
	VSub(look_at, look_from, n);
	VUnit(n, temp_d, n);

	/* v = (norm)(look_up - (look_up.n)n) */
	temp_d = VDot(look_up, n);
	VScalarMul(n, temp_d, temp_v);
	VSub(look_up, temp_v, v);
	VUnit(v, temp_d, v);

	/* u = n x v */
	VCross(n, v, u);
	if (((temp_d = VMod(u)) < EPSILON) && (temp_d > -EPSILON))
	{
		fprintf(stderr, "genray : ERROR : up vector and viewplane normal parallel.\n");
		exit(1);
	}

	/* The matrix required to transform from view to world is :	*/
	/*	[ u v n]						*/
	/* Note this is the transpose of that used to go the other way.	*/
	/* The creation of this matrix is a little messy, since we	*/
	/*  already have the column ( as opposed to row) vectors.	*/
	convert_matrix[0].x = u.x;
	convert_matrix[0].y = v.x;
	convert_matrix[0].z = n.x;
	convert_matrix[1].x = u.y;
	convert_matrix[1].y = v.y;
	convert_matrix[1].z = n.y;
	convert_matrix[2].x = u.z;
	convert_matrix[2].y = v.z;
	convert_matrix[2].z = n.z;

	/* The view reference point is eye_dist in from look_from along n. */
	VScalarMul(n, eyedist, temp_v);
	VAdd(temp_v, look_from, view_reference);
	eye_position = look_from;


	if (!have_win)
	{
		fprintf(stderr, "genray: ERROR: no window specified\n");
		exit(1);
	}

#ifdef VIEW_DEBUG
	printf("u :"); VPrint(u);
	printf("v :"); VPrint(v);
	printf("n :"); VPrint(n);
	printf("\nTransform matrix:\n"); MPrint(convert_matrix);
	printf("\nView reference point: "); VPrint(view_reference);
	printf("\nEye position: "); VPrint(eye_position);
	printf("\n");
	printf("\nLight sources:\n");
	for ( i = 0 ; i < light_number ; i++ )
	{
		VPrint(light_srcs[i].position);
		printf("Red : %d ; Green: %d ; Blue: %d\n",
			light_srcs[i].colour.red,
			light_srcs[i].colour.green,
			light_srcs[i].colour.blue);
		printf("\n");
	}
#endif

	return;
}


void
Extents_Init(Vector* min, Vector* max)
/****************************************************************/
/*																*/
/*	Initialises the projection extents for each shape.			*/
/*																*/
/****************************************************************/
{
	Instance*	obj;
	Matrix		world_to_view;
	Vector		disp;

	VNew(HUGE, HUGE, 0, (*min));
	VNew(-HUGE, -HUGE, 0, (*max));

	/* Make up the transformation to convert world coords to view coords. */
	MNew(u, v, n, world_to_view);
	VNew(-(VDot(view_reference, u)),
		 -(VDot(view_reference,v)),
		 -(VDot(view_reference,n)), disp);

	for ( obj = objects ; obj != NULL ; obj = obj->next )
	{
		switch (obj->type)
		{
			case plane :
				/* do nothing.  Planes will generally cover	*/
				/*  most of the field of view, and checking	*/
				/*  for intersections is trivial., so it's not	*/
				/*  worth the effort to find a projection	*/
				/*  extent.					*/
				VNew(-0.5 * win_width, -0.5 * win_height, 0,
					 obj->proj_extent.min);
				VNew(0.5 * win_width, 0.5 * win_height,0,obj->proj_extent.max);
				VNew(HUGE, HUGE, HUGE, obj->space_extent.max);
				VNew(-HUGE, -HUGE, -HUGE, obj->space_extent.min);
				break;
			case square :
				Square_Extent(obj, world_to_view, disp);
				break;
			case sphere:
			case cube:
			case cylinder:
				Cube_Extent(obj, world_to_view, disp);
				break;
			case cone:
				Pyramid_Extent(obj, world_to_view, disp);
				break;
			case torus:
				Torus_Extent(obj, world_to_view, disp);
				break;
			case union_op:
			case intersection_op:
			case difference_op:
				Compound_Extent(obj, world_to_view, disp);
				break;
		}

#ifdef EXTENT_DEBUG
		VPrint(obj->proj_extent.min);
		VPrint(obj->proj_extent.max);
		VPrint(obj->space_extent.min);
		VPrint(obj->space_extent.max);
#endif

		/* Update global projection extents. */
		if (obj->proj_extent.min.x < (*min).x)
			(*min).x = obj->proj_extent.min.x;
		if (obj->proj_extent.max.x > (*max).x)
			(*max).x = obj->proj_extent.max.x;
		if (obj->proj_extent.min.y < (*min).y)
			(*min).y = obj->proj_extent.min.y;
		if (obj->proj_extent.max.y > (*max).y)
			(*max).y = obj->proj_extent.max.y;
			
	}

#ifdef FILE_DEBUG
	print_objects(objects);
#endif

}



void
print_objects(Instance *start)
/****************************************************************/
/*																*/
/*	Prints out the object data.  Be warned, this can get awful	*/
/*	 big.														*/
/*																*/
/****************************************************************/
{
	Instance	*obj;

	for (  obj = start ; obj != NULL ; obj = obj->next )
	{

		printf("Type : %d\n", obj->type);
		printf("Address : %x\n", (unsigned long)obj);
		printf("Transform matrix:\n");
		MPrint(obj->transform.matrix);
		printf("Transform vector: ");
		VPrint(obj->transform.displacement);
		printf("Inverse matrix:\n");
		MPrint((obj->inverse_trans).matrix);
		printf("Inverse vector: ");
		VPrint((obj->inverse_trans).displacement);
		printf("\n");
		printf("Projection extents:\n");
		VPrint(obj->proj_extent.max);
		VPrint(obj->proj_extent.min);
		printf("Spacial extents:\n");
		VPrint(obj->space_extent.max);
		VPrint(obj->space_extent.min);
		printf("\n");
		printf("Colour : %hd %hd %hd\n", obj->nature.colour|ColMask,
			(obj->nature.colour>>8)|ColMask,
			(obj->nature.colour>>16)|ColMask);
		printf("Diffuse coef: %f\n", obj->nature.diff_coef);
		printf("Specular coef: %f\n", obj->nature.spec_coef);
		printf("Reflection coef: %f\n", obj->nature.reflect_coef);
		printf("Transparency: %f\n", obj->nature.transparency);
		printf("Refractive index: %f\n", obj->nature.refr_index_sq);

		if ((obj->type == union_op) || (obj->type == intersection_op) ||
			(obj->type == difference_op))
		{
			printf("\n");
			print_objects(obj->object1);
			print_objects(obj->object2);
		}
		printf("\n\n");
	}

}

