/*
**    Genray: A Generic Raytracer
**    Copyright (C) 1994  Stephen Chenney (stephen@cs.su.oz.au)
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program; if not, write to the Free Software
**    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
**	sbounds.c : Functions to calculate S-bounds (Cameron 1989)
**
*/

#include "genray.h"


#define max_min_to_points(pts, max, min) \
	pts[0] = max; \
	VNew(max.x, max.y, min.z, pts[1]); \
	VNew(max.x, min.y, max.z, pts[2]); \
	VNew(max.x, min.y, min.z, pts[3]); \
	VNew(min.x, max.y, max.z, pts[4]); \
	VNew(min.x, max.y, min.z, pts[5]); \
	VNew(min.x, min.y, max.z, pts[6]); \
	pts[7] = min

void Down_S_Bounds(Instance *obj, Vector *upper_bound);
void Up_S_Bounds(Instance *obj);
Cuboid Union_Bounds(Vector *a, Vector *b);
Cuboid Intersect_Bounds(Vector *a, Vector *b);
Cuboid Points_To_Bound(Vector *pts);
void Print_Bounds(Instance *obj, int indent);


/*	void
**	Calculate_S_Bounds(int num_iterations)
**	Runs the S-Bounds algoritm for all the objects.
**	num_iterations says how many times to propagate values up and down
**	the tree.
*/
void
Calculate_S_Bounds(int num_iterations)
{
	Instance 	*obj;
	int			i;

	for ( obj = objects ; obj != NULL ; obj = obj->next )
	{
		for ( i = 0 ; i < num_iterations ; i++ )
		{
			Down_S_Bounds(obj, NULL);
			Up_S_Bounds(obj);
		}
	}

}


/*	void
**	Down_S_Bounds(Instance *obj, Vector *upper_bound)
**	Given the bounds for the parent object, calculates the intersection
**	of those bounds with the object's bounds.  This is propagating the
**	S-Bounds down the tree.
*/
void
Down_S_Bounds(Instance *obj, Vector *upper_bound)
{
	Vector	obj_points[8];
	Vector	temp_v;
	int		i;

	max_min_to_points(obj_points, obj->space_extent.max, obj->space_extent.min);
	if ( upper_bound )
		obj->space_extent = Intersect_Bounds(upper_bound, obj_points);

	if ( obj->type == intersection_op ||
		 obj->type == union_op ||
		 obj->type == difference_op )
 	{
 		/* Convert the points back into this objects generic space. */
 		/* This is because the childs bounds are in this space. */
 		for ( i = 0 ; i < 8 ; i++ )
 		{
			VAdd(obj->inverse_trans.displacement, obj_points[i], temp_v);
			MVMul(obj->inverse_trans.matrix, temp_v, obj_points[i]);
 		}

 		Down_S_Bounds(obj->object1, obj_points);
 		Down_S_Bounds(obj->object2, obj_points);
 	}

}


/*	void
**	Up_S_Bounds(Instance *obj)
**	Calculates an objects bounds based on those of its children.
**	This is propagating the S-Bounds up the tree.
*/
void
Up_S_Bounds(Instance *obj)
{
	Vector	obj_points[8];
	Vector	obj1_points[8];
	Vector	obj2_points[8];
	Cuboid	child_bound;
	Vector	child_points[8];
	Vector	temp_v;
	int		i;


	if ( obj->type == sphere ||
		 obj->type == cube ||
		 obj->type == plane ||
		 obj->type == square ||
		 obj->type == cone ||
		 obj->type == cylinder ||
		 obj->type == torus ) return;

	Up_S_Bounds(obj->object1);
	Up_S_Bounds(obj->object2);

	max_min_to_points(obj_points, obj->space_extent.max, obj->space_extent.min);
	max_min_to_points(obj1_points, obj->object1->space_extent.max,
									obj->object1->space_extent.min);
	max_min_to_points(obj2_points, obj->object2->space_extent.max,
									obj->object2->space_extent.min);

	/* Convert the childs points to this objects bound space. */
	for ( i = 0 ; i < 8 ; i++ )
	{
		MVMul(obj->transform.matrix, obj1_points[i], temp_v);
		VAdd(obj->transform.displacement, temp_v, obj1_points[i]);
		MVMul(obj->transform.matrix, obj2_points[i], temp_v);
		VAdd(obj->transform.displacement, temp_v, obj2_points[i]);
	}

	switch ( obj->type )
	{
		case union_op:
			child_bound = Union_Bounds(obj1_points, obj2_points);
			max_min_to_points(child_points, child_bound.max, child_bound.min);
			break;
		case intersection_op:
			child_bound = Intersect_Bounds(obj1_points, obj2_points);
			max_min_to_points(child_points, child_bound.max, child_bound.min);
			break;
		case difference_op:
			for ( i = 0 ; i < 8 ; i++ )
				child_points[i] = obj1_points[i];
			break;
	}

	obj->space_extent = Intersect_Bounds(child_points, obj_points);

}


/*	Cuboid
**	Union_Bounds(Vector *a, Vector *b)
**	Takes the union of 2 bounding boxes.
*/
Cuboid
Union_Bounds(Vector *a, Vector *b)
{
	Cuboid	a_bound, b_bound;
	Cuboid	result;

	/* Get the extents for both the uniting bodies. */
	a_bound = Points_To_Bound(a);
	b_bound = Points_To_Bound(b);

	result.max.x = max(a_bound.max.x, b_bound.max.x);
	result.max.y = max(a_bound.max.y, b_bound.max.y);
	result.max.z = max(a_bound.max.z, b_bound.max.z);
	result.min.x = min(a_bound.min.x, b_bound.min.x);
	result.min.y = min(a_bound.min.y, b_bound.min.y);
	result.min.z = min(a_bound.min.z, b_bound.min.z);

	return result;
}



/*	Cuboid
**	Intersect_Bounds(Vector *a, Vector *b)
**	Takes the intersection of 2 bounding boxes.
*/
Cuboid
Intersect_Bounds(Vector *a, Vector *b)
{
	Cuboid	a_bound, b_bound;
	Cuboid	result;

	/* Get the extents for both the intersecting bodies. */
	a_bound = Points_To_Bound(a);
	b_bound = Points_To_Bound(b);

	result.max.x = min(a_bound.max.x, b_bound.max.x);
	result.max.y = min(a_bound.max.y, b_bound.max.y);
	result.max.z = min(a_bound.max.z, b_bound.max.z);
	result.min.x = max(a_bound.min.x, b_bound.min.x);
	result.min.y = max(a_bound.min.y, b_bound.min.y);
	result.min.z = max(a_bound.min.z, b_bound.min.z);

	return result;

}


/*	Cuboid
**	Points_To_Bound(Vector *pts)
**	Converts a set of points into a bounding box representation.
*/
Cuboid
Points_To_Bound(Vector *pts)
{
	Cuboid	result;
	int		i;

	result.min = result.max = pts[0];
	for ( i = 1 ; i < 8 ; i++ )
	{
		if (pts[i].x > result.max.x)
			result.max.x = pts[i].x;
		else if (pts[i].x < result.min.x)
			result.min.x = pts[i].x;
		if (pts[i].y > result.max.y)
			result.max.y = pts[i].y;
		else if (pts[i].y < result.min.y)
			result.min.y = pts[i].y;
		if (pts[i].z > result.max.z)
			result.max.z = pts[i].z;
		else if (pts[i].z < result.min.z)
			result.min.z = pts[i].z;
	}

	return result;

}


void
Print_Bounds(Instance *obj, int indent)
{
#define spaces(n) \
	for ( i = 0 ; i < n ; i++ ) printf(" ");

	int	i;

	if ( obj == NULL ) return;

	spaces(indent);
	switch ( obj->type )
	{
		case sphere:		printf("sphere\n"); break;
		case cube:			printf("cube\n"); break;
		case plane:			printf("plane\n"); break;
		case square:		printf("square\n"); break;
		case cylinder:		printf("cylinder\n"); break;
		case cone:			printf("cone\n"); break;
		case torus:			printf("torus\n"); break;
		case difference_op:	printf("difference_op\n"); break;
		case intersection_op:	printf("intersection_op\n"); break;
		case union_op:		printf("union_op\n"); break;
	}
	spaces(indent);
	VPrint(obj->space_extent.max);
	spaces(indent);
	VPrint(obj->space_extent.min);

	Print_Bounds(obj->object1, indent + 4);
	Print_Bounds(obj->object2, indent + 4);
}


