/*
**    Genray: A Generic Raytracer
**    Copyright (C) 1994  Stephen Chenney (stephen@cs.su.oz.au)
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program; if not, write to the Free Software
**    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/****************************************************************/
/*																*/
/*	Shading / Colouring functions for genray.					*/
/*	Written by Stephen Chenney.  Dec 1993.						*/
/*																*/
/****************************************************************/

#include "genray.h"

#define MAXDEPTH 50


Colour
Find_Colour(Ray ray, Instance* obj, double t, Vector n, int depth,
		double intensity, char in_air, InstanceList *path)
/****************************************************************/
/*																*/
/*	Determines and returns the appropriate shade for obj		*/
/*	 with ray Ray, hit time t and normal vector n.				*/
/*	Intensity is some measure of the weight this result has in	*/
/*	 the final result.  It is used to decide whether secondary	*/
/*	 rays should be cast.										*/
/*	In_air states whether the hit was in the open or inside the	*/
/*	 object.													*/
/*																*/
/****************************************************************/
{
	Vector	hit_pt;		/* Where the hit occurred. */
	Vector	src;		/* A vector to the source. */
	Vector	reflect;	/* A vector in the reflection direction.*/
	Vector	outgoing;	/* A unit vector back in the direction of ray.*/
	Vector	incoming;	/* A unit vector in the same direction as ray.*/
	Vector	norm;		/* The object's normal in world coords */
	Matrix	inv_transp;	/* the inverse transpose needed to take	*/
						/* the normal to generic.		*/

	Colour	result;
	short	result_red, result_green, result_blue;
	double	specular_contrib;
	double	diffuse_contrib;
	Ray		shadow_ray;
	Ray		reflected_ray;
	Ray		refracted_ray;
	Colour	reflected_colour;
	Colour	refracted_colour;
	double	delta;	/* This is the delta on pg 656 of Hill. */

	int		i;
	double	temp_d;
	Vector	temp_v1, temp_v2;
	InstanceList	*tmp_elmt;

	/* Transpose the hit normal.  This means running down the path	*/
	/*  performing the applicable transform.						*/
	while (path)
	{
		/* Build up the transpose matrix. */
		/* It's messy. */
		inv_transp[0].x = path->obj->inverse_trans.matrix[0].x;
		inv_transp[0].y = path->obj->inverse_trans.matrix[1].x;
		inv_transp[0].z = path->obj->inverse_trans.matrix[2].x;
		inv_transp[1].x = path->obj->inverse_trans.matrix[0].y;
		inv_transp[1].y = path->obj->inverse_trans.matrix[1].y;
		inv_transp[1].z = path->obj->inverse_trans.matrix[2].y;
		inv_transp[2].x = path->obj->inverse_trans.matrix[0].z;
		inv_transp[2].y = path->obj->inverse_trans.matrix[1].z;
		inv_transp[2].z = path->obj->inverse_trans.matrix[2].z;

		MVMul(inv_transp, n, norm);
		n = norm;
		tmp_elmt = path;
		path = path->next;
		free_inst_elmt(tmp_elmt);
		tmp_elmt = NULL;
	}
	/* Do the final object (not included in path.)	*/
	inv_transp[0].x = obj->inverse_trans.matrix[0].x;
	inv_transp[0].y = obj->inverse_trans.matrix[1].x;
	inv_transp[0].z = obj->inverse_trans.matrix[2].x;
	inv_transp[1].x = obj->inverse_trans.matrix[0].y;
	inv_transp[1].y = obj->inverse_trans.matrix[1].y;
	inv_transp[1].z = obj->inverse_trans.matrix[2].y;
	inv_transp[2].x = obj->inverse_trans.matrix[0].z;
	inv_transp[2].y = obj->inverse_trans.matrix[1].z;
	inv_transp[2].z = obj->inverse_trans.matrix[2].z;
	MVMul(inv_transp, n, norm);
	VUnit(norm, temp_d, norm);

	hit_pt.x = ray.start_point.x + ray.direction.x * t;
	hit_pt.y = ray.start_point.y + ray.direction.y * t;
	hit_pt.z = ray.start_point.z + ray.direction.z * t;


	shadow_ray.start_point = hit_pt;
	reflected_ray.start_point = hit_pt;
	refracted_ray.start_point = hit_pt;

	VUnit(ray.direction, temp_d, incoming);
	VScalarMul(incoming, -1, outgoing);

	result = 0;
	result_red = 0;
	result_green = 0;
	result_blue = 0;

	if (in_air)
	{
		result_red = (short)((ambient_light & ColMask) *
						obj->nature.diff_coef *
						(obj->nature.colour & ColMask) / 255);
		result_green = (short)(((ambient_light >> 8) & ColMask ) *
						obj->nature.diff_coef *
						((obj->nature.colour >> 8) & ColMask) / 255);
		result_blue = (short)(((ambient_light >> 16) & ColMask) *
						obj->nature.diff_coef *
						((obj->nature.colour >> 16) & ColMask) / 255);

		for ( i = 0 ; i < light_number ; i++ )
		{
			/* Calculate vector to light source. */
			VSub(light_srcs[i].position, hit_pt, shadow_ray.direction);
			VUnit(shadow_ray.direction, temp_d, src);

			if ((diffuse_contrib = VDot(src, norm)) < 0)
				continue; /* light behind surface */

			/* Check for shadowing. */
			if (In_Shadow(obj, shadow_ray)) continue;

			temp_d = 2 * VDot(src, norm);
			VScalarMul(norm, temp_d, reflect);
			VSub(reflect, src, reflect);

			if ((temp_d = VDot(outgoing, reflect)) > 0.0)
			{
				specular_contrib = 255 * obj->nature.spec_coef *
									pow(temp_d, obj->nature.spec_power);
			}
			else
			{
				specular_contrib = 0;
			}

			diffuse_contrib *= obj->nature.diff_coef;

			result_red += light_srcs[i].red * (specular_contrib +
				(obj->nature.colour & ColMask) * diffuse_contrib );
			result_green += light_srcs[i].green * (specular_contrib +
				((obj->nature.colour >> 8) & ColMask) * diffuse_contrib );
			result_blue += light_srcs[i].blue * (specular_contrib +
				((obj->nature.colour >> 16) & ColMask) * diffuse_contrib );
		
		}
	}

	if ((obj->nature.reflect_coef * intensity > 0.1) && (depth < MAXDEPTH))
	{
		/* Calculate reflected ray. */
		temp_d = 2 * VDot(ray.direction, norm);
		VScalarMul(norm, temp_d, reflected_ray.direction);
		VSub(ray.direction,reflected_ray.direction,reflected_ray.direction);

		reflected_colour =
			RaytraceSecondary(reflected_ray, obj, depth+1,
								obj->nature.reflect_coef * intensity);

		result_red += (short)(obj->nature.reflect_coef *
			(reflected_colour & ColMask));
		result_green += (short)(obj->nature.reflect_coef *
			((reflected_colour >> 8) & ColMask));
		result_blue += (short)(obj->nature.reflect_coef *
			((reflected_colour >> 16) & ColMask));

	}

	if ((obj->nature.transparency * intensity > 0.1) && (depth < MAXDEPTH))
	{
		/* Calculate refracted ray. pg656 Hill. */
		/* Assume for the moment that no 2 objects interpenetrate.	*/
		temp_d = 1.0 /  fabs(VDot(incoming, norm));
		VScalarMul(incoming, temp_d, temp_v1);
		VAdd(temp_v1, norm, temp_v2);

		if (in_air)
		{
			/* Passing into object. */
			temp_d = obj->nature.refr_index_sq * VDot(temp_v1, temp_v1) -
					 VDot(temp_v2, temp_v2);
	 	}
	 	else
	 	{
	 		/* Exiting object.*/
			temp_d = 1 / obj->nature.refr_index_sq * VDot(temp_v1, temp_v1) -
					 VDot(temp_v2, temp_v2);
	 	}

	 	if (temp_d > 0.0)		/* NOT total reflection. */
	 	{
	 		delta = sqrt(temp_d);
	 		VScalarMul(temp_v1, delta, temp_v1);
			temp_d = delta - 1;
			VScalarMul(norm, temp_d, temp_v2);
			VAdd(temp_v1, temp_v2, refracted_ray.direction);

			refracted_colour =
				RaytraceSecondary(refracted_ray, obj, depth+1,
									obj->nature.transparency * intensity);

			result_red += (short)(obj->nature.transparency *
				(refracted_colour & ColMask) *
				(obj->nature.colour & ColMask) / 255);
			result_green += (short)(obj->nature.transparency *
				((refracted_colour >> 8) & ColMask) *
				((obj->nature.colour >> 8) & ColMask) / 255);
			result_blue += (short)(obj->nature.transparency *
				((refracted_colour >> 16) & ColMask) *
				((obj->nature.colour >> 16) & ColMask) / 255);
	 	}
	}


	result = result_red > 255 ? 255 : result_red;
	result |= ( result_green > 255 ? 255 : result_green ) << 8;
	result |= ( result_blue > 255 ? 255 : result_blue ) << 16;

	return	result;

}

