/*
**    Genray: A Generic Raytracer
**    Copyright (C) 1994  Stephen Chenney (stephen@cs.su.oz.au)
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program; if not, write to the Free Software
**    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/****************************************************************/
/*																*/
/*	File containing all the shadow determining functions		*/
/*	 for genray.												*/
/*	Written by Stephen Chenney.  Jan 1994.						*/
/*																*/
/****************************************************************/

#include "genray.h"

/* Functions for intersecting compound objects. */
HitList* Sh_Comp_Sphere(Ray ray, Instance* obj);
HitList* Sh_Comp_Plane(Ray ray, Instance* obj);
HitList* Sh_Comp_Square(Ray ray, Instance* obj);
HitList* Sh_Comp_Cylinder(Ray ray, Instance* obj);
HitList* Sh_Comp_Cone(Ray ray, Instance* obj);
HitList* Sh_Comp_Cube(Ray ray, Instance* obj);
HitList* Sh_Comp_Torus(Ray ray, Instance* obj);
HitList* Sh_Comp_Compound(Ray ray, Instance* obj);


/****************************************************************/
/*																*/
/*	All the functions intersect a ray with something.			*/
/*	They return the hit time of the first hit, 2 if none.		*/
/*																*/
/****************************************************************/

double Shadow_Sphere(Ray ray, Instance* obj);
double Shadow_Cylinder(Ray ray, Instance* obj);
double Shadow_Cone(Ray ray, Instance* obj);
double Shadow_Cube(Ray ray, Instance* obj);
double Shadow_Torus(Ray ray, Instance* obj);
double Shadow_Compound(Ray ray, Instance* obj);

char
In_Shadow(Instance* obj, Ray ray)
/****************************************************************/
/*																*/
/*	Returns true if obj is shadowed by something along ray.		*/
/*																*/
/****************************************************************/
{
	Instance*	current;/* The current object of interest.	*/
	Ray	generic_ray;	/* The scene ray transformed.		*/
	double	t;

	Vector	temp_v;
	Vector*	temp_m;	/* Actually a matrix. */

	for ( current = objects ; current != NULL ; current = current->next )
	{
		if ((current == obj) && (current->type != union_op) &&
								(current->type != intersection_op) &&
								(current->type != difference_op) &&
								(current->type != torus)) continue;

		/* test against spacial extent */
		if ((current->type != plane) && ! Ray_In_Extent(ray, current))
			continue;

		/* Untransform the ray back to generic co-ords.  pg 625 TB */
		temp_v = (current->inverse_trans).displacement;
		VAdd(ray.start_point, temp_v, temp_v);
		temp_m = (current->inverse_trans).matrix;
		MVMul(temp_m, temp_v, generic_ray.start_point);
		MVMul(temp_m, ray.direction, generic_ray.direction);

		switch (current->type)
		{
			case sphere :
				if ((t = Shadow_Sphere(generic_ray, current)) && (t < 1.0))
					return TRUE;
				break;
			case plane :
				if ((generic_ray.direction.z != 0.0) &&
					((t = -(generic_ray.start_point.z) /
							generic_ray.direction.z) > DELTA) &&
					( t < 1.0))
					return TRUE;
				break;
			case square :
				if ((generic_ray.direction.z != 0.0) &&
					((t = -(generic_ray.start_point.z) /
							generic_ray.direction.z) > DELTA) &&
					( t < 1.0) &&
					(fabs(generic_ray.start_point.x +
							generic_ray.direction.x * t) <= 1.0) &&
					(fabs(generic_ray.start_point.y +
							generic_ray.direction.y * t) <= 1.0))
					return TRUE;
				break;
			case cylinder :
				if ((t = Shadow_Cylinder(generic_ray, current)) && (t < 1.0))
					return TRUE;
				break;
			case cone :
				if ((t = Shadow_Cone(generic_ray, current)) && (t < 1.0))
					return TRUE;
				break;
			case cube :
				if ((t = Shadow_Cube(generic_ray, current)) && (t < 1.0))
					return TRUE;
				break;
			case torus :
				if ((t = Shadow_Torus(generic_ray, current)) && (t < 1.0))
					return TRUE;
				break;
			case union_op :
			case difference_op :
			case intersection_op :
				if ((t = Shadow_Compound(generic_ray, current)) && (t < 1.0))
					return TRUE;
				break;
		}
	}
	return FALSE;

}


double
Shadow_Sphere(Ray ray, Instance* obj)
/****************************************************************/
/*																*/
/*	Intersects a ray with a sphere.								*/
/*	The hit time is returned in t.  The return value of the		*/
/*	 function is whether it hit at all.							*/
/*	Taken from Hill pg 627.										*/
/*																*/
/****************************************************************/
{
	register	double	AA, BB, CC;
	register	double	discrim, discrim_root;
	Vector	temp_v1;
	Vector	temp_v2;

	temp_v1 = ray.direction;
	AA = VDot(temp_v1, temp_v1);

	if (AA == 0.0)
	{
		fprintf(stderr, "genray: Shadow_Sphere: dud ray.\n");
		exit(1);
	}

	temp_v2 = ray.start_point;
	BB = VDot(temp_v2, temp_v1);
	CC = VDot(temp_v2, temp_v2) - 1;

	discrim = BB * BB - AA * CC;

	if (discrim < 0.0)
		return 2.0;
	
	discrim_root = sqrt(discrim);

	if (-BB > discrim_root + DELTA)
	{
		/* A hit on the way in. */
		return ((-BB - discrim_root) / AA);
	}
	else if (BB < discrim_root - DELTA)
	{
		/* A hit on the way out.  The normal is reversed. */
		return ((-BB + discrim_root) / AA);
	}
	return 2.0;

}


double
Shadow_Cylinder(Ray ray, Instance* obj)
/****************************************************************/
/*																*/
/*	Intersects a ray with the generic cylinder.					*/
/*	The hit time is returned in t.  The return value of the		*/
/*	 function is whether it hit at all.							*/
/*	As per outline on Hill pg 631.								*/
/*																*/
/****************************************************************/
{
	double	hit1;	/* The hit on way in. */
	double	hit2;	/* The hit on the way out. */
	double	z1, z2;
	double	AA, BB;
	double	discrim, discrim_root;

	if ((AA = ray.direction.x * ray.direction.x +
		  ray.direction.y * ray.direction.y) != 0.0)
	{
		/* There may be hits with the walls. */
		BB = ray.start_point.x * ray.direction.x +
		     ray.start_point.y * ray.direction.y;
		if ((discrim = BB * BB - AA * (ray.start_point.x * ray.start_point.x +
				ray.start_point.y * ray.start_point.y - 1)) >= 0.0)
		{
			/* There is a hit. */
			discrim_root = sqrt(discrim);

			hit1 = (-BB - discrim_root) / AA;
			z1 = ray.start_point.z + hit1 * ray.direction.z;

			if ((hit1 > DELTA) && (z1 <= 1.0) && (z1 >= -1.0))
			{
				/* hit1 is valid */
				/* Convince yourself that this is the best hit. */
				return hit1;
			}

			/* Try hit2. */
			hit2 = (-BB + discrim_root) / AA;

			if (hit2 > DELTA)
			{
				z2 = ray.start_point.z + hit2 * ray.direction.z;
				if (((z1 > 1.0) && (z2 > 1.0)) ||
					((z2 < -1.0) && (z1 < -1.0)))
				{
					/* Ray entirely misses cylinder. */
					return 2.0;
				}
				else if ((z2 <= 1.0) && (z2 >= -1.0))
				{
					/* hit2 is valid. */
					/* Test whether ray will hit either end. */
					if (z1 < -1.0)
					{
						/* Ray must hit bottom. */
						hit1 = (-1.0 - ray.start_point.z) / ray.direction.z;
						if (hit1 < DELTA)
							return hit2;
						else
							return hit1;
						
					}
					else if (z1 > 1.0)
					{
						/* Ray must hit top. */
						hit1 = (1 - ray.start_point.z) / ray.direction.z;
						if (hit1 < DELTA)
							return hit2;
						else
							return hit1;
					}
					else
						return hit2;

				}
				else
				{
					/* Need to test ends depending on z1. */
					if (z1 < -1)
					{
						/* Ray coming from bottom up. */
						hit2 = (1.0 - ray.start_point.z) / ray.direction.z;
						if (hit2 < DELTA) return 2.0;
						hit1 = (-1.0 - ray.start_point.z) / ray.direction.z;
						if (hit1 < DELTA)
							return hit2;
						else
							return hit1;
					}
					else
					{
						/* Ray coming from top down. */
						hit2 = (-1.0 - ray.start_point.z) / ray.direction.z;
						if (hit2 < DELTA)
							return 2.0;
						hit1 = (1.0 - ray.start_point.z) / ray.direction.z;
						if (hit1 < DELTA)
							return hit2;
						else
							return hit1;
					}
				}
			}
			else
				return 2.0;	/* If hit2 is behind eye, whole thing is. */
		}
		else return 2.0;		/* Ray misses entirely. */
	}
	else
	{
		/* Ray is parallel to sides. */
		/* Test top and bottom. */
		if (ray.direction.z > 0)
		{
			/* ray is heading up. */
			/* Get intersection with top plane. */
			hit2 = (1 - ray.start_point.z) / ray.direction.z;

			/* Test for behind eye. */
			if (hit2 < DELTA) return 2.0;

			/* Test for inside end circle. */
			AA = ray.start_point.x + hit2 * ray.direction.x;
			BB = ray.start_point.y + hit2 * ray.direction.y;
			if ((AA*AA + BB*BB) > 1.0) return 2.0;

			/* Make a decision. */
			hit1 = (-1.0 - ray.start_point.z) / ray.direction.z;
			if (hit1 < DELTA)
				return hit1;
			else
				return hit2;
		}
		else
		{
			/* ray is going down. */
			/* Get intersection with bottom plane. */
			hit2 = (-1.0 - ray.start_point.z) / ray.direction.z;

			/* Test for behind eye. */
			if (hit2 < DELTA) return 2.0;

			/* Test for inside end circle. */
			AA = ray.start_point.x + hit2 * ray.direction.x;
			BB = ray.start_point.y + hit2 * ray.direction.y;
			if ((AA*AA + BB*BB) > 1.0) return 2.0;

			/* Make a decision. */
			hit1 = (1.0 - ray.start_point.z) / ray.direction.z;
			if (hit1 < DELTA)
				return hit2;
			else
				return hit1;
		}
	}
}


double
Shadow_Cone(Ray ray, Instance* obj)
/****************************************************************/
/*																*/
/*	Intersects a ray with the generic cone.						*/
/*	The hit time is returned in t.  The return value of the		*/
/*	 function is whether it hit at all.							*/
/*	As per outline on Hill pg 631-2.							*/
/*	The code is entirely non obvious.  I arrived at it by		*/
/*	 drawing a pretty picture and examining cases. There are 9	*/
/*	 or so.														*/
/*																*/
/****************************************************************/
{
	double	hit1;	/* The hit on way in. */
	double	hit2;	/* The hit on the way out. */
	double	z1, z2;
	double	AA, BB;
	double	temp_d;
	double	discrim, discrim_root;

	/* AA = 0 I think implies the ray lies in one of the walls of the cone. */
	if ((AA = ray.direction.x * ray.direction.x +
		  ray.direction.y * ray.direction.y -
		  ray.direction.z * ray.direction.z / 4) != 0.0)
	{
		/* There may be hits with the walls. */
		BB = ray.start_point.x * ray.direction.x +
		     ray.start_point.y * ray.direction.y +
		     ray.direction.z * ( 1 - ray.start_point.z) / 4;
		temp_d = 1 - ray.start_point.z;
		if ((discrim = BB * BB - AA * (ray.start_point.x * ray.start_point.x +
			ray.start_point.y * ray.start_point.y - temp_d*temp_d / 4)) >= 0.0)
		{
			/* There is a hit. */
			discrim_root = sqrt(discrim);

			if (AA > 0)
				hit1 = (-BB - discrim_root) / AA;
			else
				hit1 = (-BB + discrim_root) / AA;
			z1 = ray.start_point.z + hit1 * ray.direction.z;

			if ((hit1 > DELTA) && (z1 <= 1.0) && (z1 >= -1.0))
			{
				/* hit1 is valid */
				/* Possibly need to test against the base. */
				if (ray.direction.z > 0)
				{
					/* May be a valid hit on the bottom. */
					hit2 = (-1.0 - ray.start_point.z) / ray.direction.z;
					AA = ray.start_point.x + ray.direction.x * hit2;
					BB = ray.start_point.y + ray.direction.y * hit2;
					AA = AA*AA + BB*BB;

					if ((hit2 < DELTA) || (AA > 1.0))
						/* Doesn't intersect base. */
						return hit1;
					else
						return hit2;
				}
				else
					return hit1;
			}

			/* Try hit2. */
			if (AA > 0)
				hit2 = (-BB + discrim_root) / AA;
			else
				hit2 = (-BB - discrim_root) / AA;

			if (hit2 > DELTA)
			{
				z2 = ray.start_point.z + hit2 * ray.direction.z;
				if ((z2 <= 1.0) && (z2 >= -1.0))
				{
					/* hit2 is valid. */
					/* Test whether ray will hit the end. */
					if (z1 < -1.0)
					{
						/* Ray must hit bottom. */
						hit1 = (-1.0 - ray.start_point.z) / ray.direction.z;
						if (hit1 < DELTA)
							return hit2;
						else
							return hit1;
					}
					else if (z1 > 1.0)
						return hit2;
					else
						return hit2;
				}
				else return FALSE;
				
			}
			else if (z1 > 1.0)
			{
				/* May be a valid hit on the bottom. */
				hit2 = (-1.0 - ray.start_point.z) / ray.direction.z;
				if (hit2 < DELTA) return 2.0;
				return hit2;
			}
			else return 2.0;
		}
		else return 2.0;
	}
	else return 2.0;
}



double
Shadow_Cube(Ray ray, Instance* obj)
/****************************************************************/
/*																*/
/*	Intersects a ray with the generic cube.						*/
/*	The hit time is returned in t.  The return value of the		*/
/*	 function is whether it hit at all.							*/
/*	As per outline on Hill pg 632-5.							*/
/*	I use an array of planes for the cube, globally defined.	*/
/*																*/
/****************************************************************/
{
	double	tin = -HUGE;
	double	tout = HUGE;
	double	thit;

	/* Since there are so many 0's in all this, I'm going	*/
	/*  to do it linearly.  This makes for poor readability	*/
	/*  and lots of code, but much greater speed.		*/

	/* x = 1 plane: n = {1,0,0}, D = 1 */
	if (ray.direction.x == 0.0)	/* Ray parallel.	*/
	{
		if (1 - ray.start_point.x < 0.0) return 2.0;	/* Ray outside */
		/* Check x = -1 plane while we're at it. */
		if (1 + ray.start_point.x < 0.0) return 2.0;
	}
	else
	{
		thit = (1 - ray.start_point.x) / ray.direction.x;
		if (ray.direction.x > 0.0)
			tout = thit;
		else
			tin = thit;
	}
	/* Not possible to eliminate ray yet */

	/* y = 1 plane: n = {0,1,0}, D = 1 */
	if (ray.direction.y == 0.0)     /* Ray parallel.        */
	{
		if (1 - ray.start_point.y < 0.0) return 2.0;  /* Ray outside */
		/* Check y = -1 plane too. */
		if (1 + ray.start_point.y < 0.0) return 2.0;
	}
	else
	{
		thit = (1 - ray.start_point.y) / ray.direction.y;
		if (ray.direction.y > 0.0)
		{
			if (thit < tout)
				tout = thit;
		}
		else
		{
			if (thit > tin)
				tin = thit;
		}
	}
	/* Try to eliminate intersection. */
	if (tin > tout) return FALSE;

	/* z = 1 plane: n = {0,0,1}, D = 1 */
	if (ray.direction.z == 0.0)     /* Ray parallel.        */
	{
		if (1 - ray.start_point.z < 0.0) return 2.0;  /* Ray outside */
		/* Check y = -1 plane too. */
		if (1 + ray.start_point.z < 0.0) return 2.0;
	}
	else
	{
		thit = (1 - ray.start_point.z) / ray.direction.z;
		if (ray.direction.z > 0.0)
		{
			if (thit < tout)
				tout = thit;
		}
		else
		{
			if (thit > tin)
				tin = thit;
		}
	}
	/* Try to eliminate intersection. */
	if (tin > tout) return 2.0;


	/* x = -1 plane : n = {-1,0,0}, D = 1 */
	/* Already eliminated parallel rays above. */
	thit = (1 + ray.start_point.x) / -ray.direction.x;
	if (ray.direction.x <  0.0)
	{
		if (thit < tout)
			tout = thit;
	}
	else
	{
		if (thit > tin)
			tin = thit;
	}
	if (tin > tout) return FALSE;

	/* y = -1 plane : n = {0,-1,0}, D = 1 */
	/* Already eliminated parallel rays above. */
	thit = (1 + ray.start_point.y) / -ray.direction.y;
	if (ray.direction.y <  0.0)
	{
		if (thit < tout)
			tout = thit;
	}
	else
	{
		if (thit > tin)
			tin = thit;
	}
	if (tin > tout) return 2.0;

	/* z = -1 plane : n = {0,0,-1}, D = 1 */
	/* Already eliminated parallel rays above. */
	thit = (1 + ray.start_point.z) / -ray.direction.z;
	if (ray.direction.z <  0.0)
	{
		if (thit < tout)
			tout = thit;
	}
	else
	{
		if (thit > tin)
			tin = thit;
	}
	if (tin > tout) return 2.0;

	/* Here if there is a valid intersection */
	if (tin > DELTA)
		return tin;
	
	if (tout > DELTA)
		return tout;
	

	return 2.0;

}


double
Shadow_Torus(Ray ray, Instance* obj)
/********************************************************************/
/*																	*/
/*	Intersects ray with a generic torus to determine shadowing.		*/
/*	 points and normals.											*/
/*	The quartic root solving code is adapted from Roots3And4.c,		*/
/*	 from Graphics Gems.											*/
/*																	*/
/********************************************************************/
{
	double		AA, BB, CC;	/* Coefficients for sphere eqn.*/
	double		hit[4];
	Vector		unit_dir;
	double		temp1, temp2, temp3, temp4, temp5, temp6;
	double		line_scale;	/* The scaling factor for the ray direction. */
	double		coeffs[4];	/* coefficients for quartic eqn. */
	int			num = 0;
	char		have_hit = FALSE;
	int			i;

	/* Test for some special cases.								*/
	/* First, test for rays which don't intersect the top and	*/
	/*  bottom planes of the torus.  ie the plane z=1 and z=-1.	*/
	if ((ray.start_point.z > 1.0) && (ray.direction.z > 0.0))
		return 2.0;

	if ((ray.start_point.z < -1.0) && (ray.direction.z < 0.0))
		return 2.0;

	/* Test for rays which don't intersect the bounding sphere, r=d+1. */
	AA = VDot(ray.direction, ray.direction);
	BB = VDot(ray.start_point, ray.direction);
	CC = VDot(ray.start_point, ray.start_point) -
			(obj->torus_rad + 1.0) * (obj->torus_rad + 1.0);
	if ((BB * BB - AA * CC) < 0.0) return 2.0;

	VUnit(ray.direction, line_scale, unit_dir);

	temp1 = obj->torus_rad * obj->torus_rad;
	temp2 = ray.start_point.z * ray.start_point.z;
	temp3 = unit_dir.z * unit_dir.z;
	temp4 = unit_dir.z * ray.start_point.z;
	temp5 = ray.start_point.x * ray.start_point.x + ray.start_point.y *
			ray.start_point.y + temp2 - temp1 - 1.0;
	temp6 = ray.start_point.x * unit_dir.x + ray.start_point.y *
			unit_dir.y + ray.start_point.z * unit_dir.z;

	coeffs[3] = 4.0 * temp6;
	coeffs[2] = 2.0 * (temp5 + 2.0 * temp6 * temp6 + 2.0 * temp1 * temp3);
	coeffs[1] = 4.0 * (temp6 * temp5 + 2.0 * temp1 * temp4);
	coeffs[0] = temp5 * temp5 + 4.0 * temp1 * (temp2 - 1.0);

	/* Solve the resulting quartic equation.				*/
	num = SolveQuartic(coeffs, hit);

	if (num == 0)
		return 2.0;

	temp1 = HUGE;
	/* Return the first valid hit. */
	for ( i = 0 ; i < num ; i++)
	{
		hit[i] *= line_scale;
		if ((hit[i] > DELTA) && (hit[i] < temp1))
		{
			temp1 = hit[i];
			have_hit = TRUE;
		}
	}

	if (have_hit)
		return temp1;
	else
		return 2.0;

}


#define sh_free_hit_list(list, tmp) \
	while (list) \
	{ \
		tmp = list; \
		list = list->next; \
		free_hit_elmt(tmp); \
	} \



double
Shadow_Compound(Ray ray, Instance* obj)
/****************************************************************/
/*																*/
/*	Determines whether ray hit the compound object obj.			*/
/*	t returns the hit time, n the hit vector, in_air the medium.*/
/*	Derivation returns the path needed to reach the basic		*/
/*	 object, from top to bottom.								*/
/*																*/
/****************************************************************/
{
	HitList	*hit_list;
	HitList	*obj1_list = NULL;
	HitList	*obj2_list = NULL;
	HitList	*part_list;
	HitList	*temp_list;
	HitList	*temp_list1;

	double	result;

	Ray		generic_ray;

	char	in_object, in_object1, in_object2;

	Vector	temp_v;
	Vector	*temp_m;

	/* For each sub-object, transform the ray and determine the hit	*/
	/*  times for that object.										*/

	/* Object1. */
	temp_v = (obj->object1->inverse_trans).displacement;
	VAdd(ray.start_point, temp_v, temp_v);
	temp_m = (obj->object1->inverse_trans).matrix;
	MVMul(temp_m, temp_v, generic_ray.start_point);
	MVMul(temp_m, ray.direction, generic_ray.direction);
	switch (obj->object1->type)
	{
		case sphere:
			obj1_list = Sh_Comp_Sphere(generic_ray, obj->object1);
			break;
		case plane:
			obj1_list = Sh_Comp_Plane(generic_ray, obj->object1);
			break;
		case square:
			obj1_list = Sh_Comp_Square(generic_ray, obj->object1);
			break;
		case cone:
			obj1_list = Sh_Comp_Cone(generic_ray, obj->object1);
			break;
		case cylinder:
			obj1_list = Sh_Comp_Cylinder(generic_ray, obj->object1);
			break;
		case cube:
			obj1_list = Sh_Comp_Cube(generic_ray, obj->object1);
			break;
		case torus:
			obj1_list = Sh_Comp_Torus(generic_ray, obj->object1);
			break;
		case union_op:
		case intersection_op:
		case difference_op:
			obj1_list = Sh_Comp_Compound(generic_ray, obj->object1);
			break;
	}
	

	/* Object2. */
	temp_v = (obj->object2->inverse_trans).displacement;
	VAdd(ray.start_point, temp_v, temp_v);
	temp_m = (obj->object2->inverse_trans).matrix;
	MVMul(temp_m, temp_v, generic_ray.start_point);
	MVMul(temp_m, ray.direction, generic_ray.direction);
	switch (obj->object2->type)
	{
		case sphere:
			obj2_list = Sh_Comp_Sphere(generic_ray, obj->object2);
			break;
		case plane:
			obj2_list = Sh_Comp_Plane(generic_ray, obj->object2);
			break;
		case square:
			obj2_list = Sh_Comp_Square(generic_ray, obj->object2);
			break;
		case cone:
			obj2_list = Sh_Comp_Cone(generic_ray, obj->object2);
			break;
		case cylinder:
			obj2_list = Sh_Comp_Cylinder(generic_ray, obj->object2);
			break;
		case cube:
			obj2_list = Sh_Comp_Cube(generic_ray, obj->object2);
			break;
		case torus:
			obj2_list = Sh_Comp_Torus(generic_ray, obj->object2);
			break;
		case union_op:
		case intersection_op:
		case difference_op:
			obj2_list = Sh_Comp_Compound(generic_ray, obj->object2);
			break;
	}
	

	/* Combine the lists. */
	if (!obj1_list && !obj2_list)
	{
		/* Both lists empty, no hits, return FALSE. */
		return 2.0;
	}
	if (!obj1_list)
	{
		if (obj->type == union_op)
			hit_list = obj2_list;
		else
		{
			sh_free_hit_list(obj2_list, temp_list);
			hit_list = NULL;
		}
	}
	else if (!obj2_list)
	{
		if (obj->type == intersection_op)
		{
			sh_free_hit_list(obj1_list, temp_list);
			hit_list = NULL;
		}
		else
			hit_list = obj1_list;
	}
	else
	{
		in_object = in_object1 = in_object2 = FALSE;
		part_list = hit_list = NULL;
		
		switch (obj->type)
		{
			/* First the non-cases (to stop gcc complaining) */
			case sphere:
			case cube:
			case plane:
			case square:
			case cone:
			case cylinder:
			case torus : break;

			case union_op:
				while (obj1_list && obj2_list)
				{
					if (obj1_list->t_hit < obj2_list->t_hit)
					{
						if (!in_object ||
							(in_object && in_object1 && !in_object2))
						{
							if (part_list)
							{
								part_list->next = obj1_list;
								part_list = part_list->next;
							}
							else
							{
								part_list = hit_list = obj1_list;
							}
							obj1_list = obj1_list->next;
						}
						else
						{
							temp_list = obj1_list;
							obj1_list = obj1_list->next;
							free_hit_elmt(temp_list);
						}
						in_object1 = !in_object1;
						in_object = in_object1 || in_object2;
					}
					else
					{
						if (!in_object ||
							(in_object && in_object2 && !in_object1))
						{
							if (part_list)
							{
								part_list->next = obj2_list;
								part_list = part_list->next;
							}
							else
							{
								part_list = hit_list = obj2_list;
							}
							obj2_list = obj2_list->next;
						}
						else
						{
							temp_list = obj2_list;
							obj2_list = obj2_list->next;
							free_hit_elmt(temp_list);
						}
						in_object2 = !in_object2;
						in_object = in_object1 || in_object2;
					}
				}
				/* Add the end bit */
				if (obj1_list)
					part_list->next = obj1_list;
				else
					part_list->next = obj2_list;
				break;

			case intersection_op:
				while (obj1_list && obj2_list)
				{
					if (obj1_list->t_hit < obj2_list->t_hit)
					{
						if (in_object || in_object2)
						{
							if (part_list)
							{
								part_list->next = obj1_list;
								part_list = part_list->next;
							}
							else
							{
								part_list = hit_list = obj1_list;
							}
							obj1_list = obj1_list->next;
						}
						else
						{
							temp_list = obj1_list;
							obj1_list = obj1_list->next;
							free_hit_elmt(temp_list);
						}
						in_object1 = !in_object1;
						in_object = in_object1 && in_object2;
					}
					else
					{
						if (in_object || in_object1)
						{
							if (part_list)
							{
								part_list->next = obj2_list;
								part_list = part_list->next;
							}
							else
							{
								part_list = hit_list = obj2_list;
							}
							obj2_list = obj2_list->next;
						}
						else
						{
							temp_list = obj2_list;
							obj2_list = obj2_list->next;
							free_hit_elmt(temp_list);
						}
						in_object2 = !in_object2;
						in_object = in_object1 && in_object2;
					}
				}

				/* Free the end bit */
				if (part_list)
					part_list->next = NULL;
				sh_free_hit_list(obj1_list, temp_list);
				sh_free_hit_list(obj2_list, temp_list);
				break;

			case difference_op:
				while (obj1_list && obj2_list)
				{
					if (obj1_list->t_hit < obj2_list->t_hit)
					{
						if (!in_object2)
						{
							if (part_list)
							{
								part_list->next = obj1_list;
								part_list = part_list->next;
							}
							else
							{
								part_list = hit_list = obj1_list;
							}
							obj1_list = obj1_list->next;
						}
						else
						{
							temp_list = obj1_list;
							obj1_list = obj1_list->next;
							free_hit_elmt(temp_list);
						}
						in_object1 = !in_object1;
						in_object = in_object1 && !in_object2;
					}
					else
					{
						if (in_object || in_object1)
						{
							if (part_list)
							{
								part_list->next = obj2_list;
								part_list = part_list->next;
							}
							else
							{
								part_list = hit_list = obj2_list;
							}
							obj2_list = obj2_list->next;
						}
						else
						{
							temp_list = obj2_list;
							obj2_list = obj2_list->next;
							free_hit_elmt(temp_list);
						}
						in_object2 = !in_object2;
						in_object = in_object1 && !in_object2;
					}
				}
				/* Add the end bit */
				if (obj1_list)
				{
					if (part_list)
						part_list->next = obj1_list;
					else
						part_list = hit_list = obj1_list;
				}
				else
				{
					if (part_list)
						part_list->next = NULL;
					/* Free list 2. */
					sh_free_hit_list(obj2_list, temp_list);
				}

				/* Remove duplicate hits from the list.  These cause errors. */
				while (hit_list &&
					   dequal(hit_list->t_hit, hit_list->next->t_hit))
				{
					temp_list = hit_list->next->next;
					hit_list->next->next = NULL;
					sh_free_hit_list(hit_list, temp_list1);
					hit_list = temp_list;
				}
				if (!hit_list) break;
				part_list = hit_list;
				temp_list = part_list->next->next;
				while (temp_list)
				{
					if (dequal(temp_list->t_hit, temp_list->next->t_hit))
					{
						part_list->next->next = temp_list->next->next;
						temp_list->next->next = NULL;
						sh_free_hit_list(temp_list, temp_list1);
					}
					else
						part_list = part_list->next->next;
					temp_list = part_list->next->next;
				}
				break;

		}
	}

	/* Find the first valid hit. */
	in_object = FALSE;
	for ( temp_list = hit_list ;
		  temp_list && (temp_list->t_hit < DELTA) ;
		  temp_list=temp_list->next );

	if (temp_list == NULL)
	{
		sh_free_hit_list(hit_list, temp_list1);
		return 2.0;
	}

	result = temp_list->t_hit;
	sh_free_hit_list(hit_list, temp_list1);

	return result;
}



HitList*
Sh_Comp_Sphere(Ray ray, Instance* obj)
/****************************************************************/
/*																*/
/*	Intersects a ray with a sphere for combination.				*/
/*	The hit data is returned, or NULL if no hit.				*/
/*	Taken from Hill pg 627.										*/
/*																*/
/****************************************************************/
{
	HitList	*result;
	double	AA, BB, CC;
	double	discrim, discrim_root;
	Vector	temp_v1;
	Vector	temp_v2;

	temp_v1 = ray.direction;
	AA = VDot(temp_v1, temp_v1);

	if (AA == 0.0)
	{
		fprintf(stderr, "genray: Sh_Comp_Sphere: dud ray.\n");
		exit(1);
	}

	temp_v2 = ray.start_point;
	BB = VDot(temp_v2, temp_v1);
	CC = VDot(temp_v2, temp_v2) - 1;

	discrim = BB * BB - AA * CC;

	if (discrim < 0.0)
		return NULL;
	
	result = get_hit_elmt();
	result->next = get_hit_elmt();
	result->next->next = NULL;

	discrim_root = sqrt(discrim);

	/* A hit on the way in. */
	result->t_hit = (-BB - discrim_root) / AA;

	/* A hit on the way out.  The normal is reversed. */
	result->next->t_hit = (-BB + discrim_root) / AA;

	return result;

}


HitList*
Sh_Comp_Plane(Ray ray, Instance* obj)
/****************************************************************/
/*																*/
/*	Intersects a ray with a plane for combination.				*/
/*	The hit data is returned, or NULL if no hit.				*/
/*																*/
/****************************************************************/
{
	HitList	*result;

	if (ray.direction.z == 0.0) return NULL;

	result = get_hit_elmt();
	result->next = get_hit_elmt();
	result->next->next = NULL;

	result->t_hit = result->next->t_hit = -(ray.start_point.z) /
										  ray.direction.z;

	return result;

}


HitList*
Sh_Comp_Square(Ray ray, Instance* obj)
/****************************************************************/
/*																*/
/*	Intersects a ray with a square for combination.				*/
/*	The hit data is returned, or NULL if no hit.				*/
/*																*/
/****************************************************************/
{
	HitList	*result;
	double	t;

	if (ray.direction.z == 0.0) return FALSE;

	t = -(ray.start_point.z) / ray.direction.z;

	if ((fabs(ray.start_point.x + ray.direction.x * t) <= 1.0) &&
	    (fabs(ray.start_point.y + ray.direction.y * t) <= 1.0))
	{
		result = get_hit_elmt();
		result->next = get_hit_elmt();
		result->next->next = NULL;

		return result;
	}

	return NULL;
}


HitList*
Sh_Comp_Cylinder(Ray ray, Instance* obj)
/****************************************************************/
/*																*/
/*	Intersects a ray with a cylinder for combination.			*/
/*	The hit data is returned, or NULL if no hit.				*/
/*																*/
/****************************************************************/
{
	HitList	*result;
	double	hit1 = -HUGE_VAL;	/* The hit on way in. */
	double	hit2 = -HUGE_VAL;	/* The hit on the way out. */
	double	z1, z2;
	double	AA, BB;
	double	discrim, discrim_root;

	if ((AA = ray.direction.x * ray.direction.x +
		  ray.direction.y * ray.direction.y) != 0.0)
	{
		/* There may be hits with the walls. */
		BB = ray.start_point.x * ray.direction.x +
		     ray.start_point.y * ray.direction.y;
		if ((discrim = BB * BB - AA * (ray.start_point.x * ray.start_point.x +
				ray.start_point.y * ray.start_point.y - 1)) >= 0.0)
		{
			/* There is a hit. */
			discrim_root = sqrt(discrim);

			hit1 = (-BB - discrim_root) / AA;
			z1 = ray.start_point.z + hit1 * ray.direction.z;
			hit2 = (-BB + discrim_root) / AA;
			z2 = ray.start_point.z + hit2 * ray.direction.z;

			if (((z1 > 1.0) && (z2 > 1.0)) || ((z2 < -1.0) && (z1 < -1.0)))
			{
				/* Ray entirely misses cylinder. */
				return NULL;
			}
			else
			{
				/* Must be hits. */
				result = get_hit_elmt();
				result->next = get_hit_elmt();
				result->next->next = NULL;

				if ((z1 < 1.0) && (z1 > -1.0) && (z2 < 1.0) && (z2 > -1.0))
				{
					/* Both hits on walls. */
					result->t_hit = hit1;
					result->next->t_hit = hit2;
					return result;
				}
				
				if ((z1 < 1.0) && (z1 > -1.0))
				{
					/* hit1 is valid. */
					result->t_hit = hit1;

					/* Find out where other hit is. Must be one end. */
					if (z2 > 1.0)
					{
						/* Ray must hit top. */
						result->next->t_hit = (1 - ray.start_point.z)
												/ ray.direction.z;
						return result;
					}
					else
					{
						/* Ray must hit bottom. */
						result->next->t_hit = (-1.0 - ray.start_point.z)
												/ ray.direction.z;
						return result;
					}
				}
				else if ((z2 < 1.0) && (z2 > -1.0))
				{
					/* hit2 is valid. */
					result->next->t_hit = hit2;

					/* Find out where other hit is. Must be one end. */
					if (z1 > 1.0)
					{
						/* Ray must hit top. */
						result->t_hit = (1 - ray.start_point.z)
										/ ray.direction.z;
						return result;
					}
					else
					{
						/* Ray must hit bottom. */
						result->t_hit = (-1.0 - ray.start_point.z)
										/ ray.direction.z;
						return result;
					}
				}
				else
				{
					/* Both hits on ends. */
					/* Find out where other hit is. Must be one end. */
					if (z1 > 1.0)
					{
						/* Ray must hit top first. */
						result->t_hit = (1 - ray.start_point.z)
												/ ray.direction.z;
						result->next->t_hit = (-1.0 - ray.start_point.z)
												/ ray.direction.z;
						return result;
					}
					else
					{
						/* Ray must hit bottom first. */
						result->t_hit = (-1.0 - ray.start_point.z)
										/ ray.direction.z;
						result->next->t_hit = (1 - ray.start_point.z)
												/ ray.direction.z;
						return result;
					}
				}
			}
		}
		else return NULL;		/* Ray misses entirely. */
	}
	else
	{
		/* Ray is parallel to sides. */
		hit1 = (1 - ray.start_point.z) / ray.direction.z;

		/* Test for inside end circle. */
		AA = ray.start_point.x + hit2 * ray.direction.x;
		BB = ray.start_point.y + hit2 * ray.direction.y;
		if ((AA*AA + BB*BB) > 1.0)
			return NULL;

		/* One hit is valid so must the other one be. */
		hit2 = (-1.0 - ray.start_point.z) / ray.direction.z;

		result = get_hit_elmt();
		result->next = get_hit_elmt();
		result->next->next = NULL;

		if (hit1 < hit2)
		{
			result->t_hit = hit1;
			result->next->t_hit = hit2;
		}
		else
		{
			result->t_hit = hit2;
			result->next->t_hit = hit1;
		}

		return result;
	}
}


HitList*
Sh_Comp_Cone(Ray ray, Instance* obj)
/****************************************************************/
/*																*/
/*	Intersects a ray with a cone for combination.				*/
/*	The hit data is returned, or NULL if no hit.				*/
/*	The code is entirely non obvious.  I arrived at it by		*/
/*	 drawing a pretty picture and examining cases. There are 9	*/
/*	 or so.														*/
/*																*/
/****************************************************************/
{
	HitList	*result;
	double	hit1;	/* The side hit on way in. */
	double	hit2;	/* The side hit on the way out. */
	double	z1, z2;
	double	AA, BB;
	double	temp_d;
	double	discrim, discrim_root;

	/* AA = 0 I think implies the ray lies in one of the walls of the cone. */
	if ((AA = ray.direction.x * ray.direction.x +
		  ray.direction.y * ray.direction.y -
		  ray.direction.z * ray.direction.z / 4) != 0.0)
	{
		/* There may be hits with the walls. */
		BB = ray.start_point.x * ray.direction.x +
		     ray.start_point.y * ray.direction.y +
		     ray.direction.z * ( 1 - ray.start_point.z) / 4;
		temp_d = 1 - ray.start_point.z;
		if ((discrim = BB * BB - AA * (ray.start_point.x * ray.start_point.x +
			ray.start_point.y * ray.start_point.y - temp_d*temp_d / 4)) >= 0.0)
		{
			/* There is a hit. */
			discrim_root = sqrt(discrim);

			if (AA > 0)
			{
				hit1 = (-BB - discrim_root) / AA;
				hit2 = (-BB + discrim_root) / AA;
			}
			else
			{
				hit1 = (-BB + discrim_root) / AA;
				hit2 = (-BB - discrim_root) / AA;
			}
			z1 = ray.start_point.z + hit1 * ray.direction.z;
			z2 = ray.start_point.z + hit2 * ray.direction.z;

			if (((z1 > 1.0) && ((z2 > 1.0) || (z2 < -1.0))) ||
				((z1 < -1.0) && ((z2 > 1.0) || (z2 < -1.0))))
			{
				/* Ray entirely misses cone. */
				return NULL;
			}
			else
			{
				/* Must be hits. */
				result = get_hit_elmt();
				result->next = get_hit_elmt();
				result->next->next = NULL;

				if (((z1 < 1.0) && (z1 > -1.0)) && ((z2 < 1.0) && (z2 > -1.0)))
				{
					/* Both hits on walls. */
					result->t_hit = hit1;
					result->next->t_hit = hit2;
					return result;
				}
				else if ((z1 < 1.0) && (z1 > -1.0))
				{
					/* Valid base hit as hit2. */
					hit2 = (-1.0 - ray.start_point.z) / ray.direction.z;
					if (hit2 < hit1)
					{
						result->t_hit = hit2;
						result->next->t_hit = hit1;
					}
					else
					{
						result->next->t_hit = hit2;
						result->t_hit = hit1;
					}

					return result;
				}
				else
				{
					/* hit2 must be good and a valid base hit. */
					hit1 = (-1.0 - ray.start_point.z) / ray.direction.z;
					if (hit1 < hit2)
					{
						result->t_hit = hit1;
						result->next->t_hit = hit2;
					}
					else
					{
						result->next->t_hit = hit1;
						result->t_hit = hit2;
					}

					return result;
				}
			}
		}
		else return NULL;
	}
	else return NULL;
}



HitList*
Sh_Comp_Cube(Ray ray, Instance* obj)
/****************************************************************/
/*																*/
/*	Intersects a ray with a cube for combination.				*/
/*	The hit data is returned, or NULL if no hit.				*/
/*																*/
/****************************************************************/
{
	HitList	*result;
	double	tin = -HUGE;
	double	tout = HUGE;
	double	thit;

	/* Since there are so many 0's in all this, I'm going	*/
	/*  to do it linearly.  This makes for poor readability	*/
	/*  and lots of code, but much greater speed.		*/

	/* x = 1 plane: n = {1,0,0}, D = 1 */
	if (ray.direction.x == 0.0)	/* Ray parallel.	*/
	{
		if (1 - ray.start_point.x < 0.0) return NULL;	/* Ray outside */
		/* Check x = -1 plane while we're at it. */
		if (1 + ray.start_point.x < 0.0) return NULL;
	}
	else
	{
		thit = (1 - ray.start_point.x) / ray.direction.x;
		if (ray.direction.x > 0.0)
			tout = thit;
		else
			tin = thit;
	}
	/* Not possible to eliminate ray yet */

	/* y = 1 plane: n = {0,1,0}, D = 1 */
	if (ray.direction.y == 0.0)     /* Ray parallel.        */
	{
		if (1 - ray.start_point.y < 0.0) return NULL;  /* Ray outside */
		/* Check y = -1 plane too. */
		if (1 + ray.start_point.y < 0.0) return NULL;
	}
	else
	{
		thit = (1 - ray.start_point.y) / ray.direction.y;
		if (ray.direction.y > 0.0)
		{
			if (thit < tout)
				tout = thit;
		}
		else
		{
			if (thit > tin)
				tin = thit;
		}
	}
	/* Try to eliminate intersection. */
	if (tin > tout) return NULL;

	/* z = 1 plane: n = {0,0,1}, D = 1 */
	if (ray.direction.z == 0.0)     /* Ray parallel.        */
	{
		if (1 - ray.start_point.z < 0.0) return NULL;  /* Ray outside */
		/* Check y = -1 plane too. */
		if (1 + ray.start_point.z < 0.0) return NULL;
	}
	else
	{
		thit = (1 - ray.start_point.z) / ray.direction.z;
		if (ray.direction.z > 0.0)
		{
			if (thit < tout)
				tout = thit;
		}
		else
		{
			if (thit > tin)
				tin = thit;
		}
	}
	/* Try to eliminate intersection. */
	if (tin > tout) return NULL;


	/* x = -1 plane : n = {-1,0,0}, D = 1 */
	/* Already eliminated parallel rays above. */
	thit = (1 + ray.start_point.x) / -ray.direction.x;
	if (ray.direction.x <  0.0)
	{
		if (thit < tout)
			tout = thit;
	}
	else
	{
		if (thit > tin)
			tin = thit;
	}
	if (tin > tout) return NULL;

	/* y = -1 plane : n = {0,-1,0}, D = 1 */
	/* Already eliminated parallel rays above. */
	thit = (1 + ray.start_point.y) / -ray.direction.y;
	if (ray.direction.y <  0.0)
	{
		if (thit < tout)
			tout = thit;
	}
	else
	{
		if (thit > tin)
			tin = thit;
	}
	if (tin > tout) return NULL;

	/* z = -1 plane : n = {0,0,-1}, D = 1 */
	/* Already eliminated parallel rays above. */
	thit = (1 + ray.start_point.z) / -ray.direction.z;
	if (ray.direction.z <  0.0)
	{
		if (thit < tout)
			tout = thit;
	}
	else
	{
		if (thit > tin)
			tin = thit;
	}
	if (tin > tout) return NULL;

	result = get_hit_elmt();
	result->next = get_hit_elmt();
	result->next->next = NULL;

	result->t_hit = tin;
	result->next->t_hit = tout;

	return result;

}


HitList*
Sh_Comp_Compound(Ray ray, Instance* obj)
/****************************************************************/
/*																*/
/*	Determines whether ray hit the compound object obj.			*/
/*	Return data is as above.									*/
/*																*/
/****************************************************************/
{
	HitList	*hit_list;
	HitList	*obj1_list = NULL;
	HitList	*obj2_list = NULL;
	HitList	*part_list;
	HitList	*temp_list;
	HitList	*temp_list1;

	Ray		generic_ray;

	char	in_object, in_object1, in_object2;

	Vector	temp_v;
	Vector	*temp_m;

	/* For each sub-object, transform the ray and determine the hit	*/
	/*  times for that object.										*/

	/* Object1. */
	temp_v = (obj->object1->inverse_trans).displacement;
	VAdd(ray.start_point, temp_v, temp_v);
	temp_m = (obj->object1->inverse_trans).matrix;
	MVMul(temp_m, temp_v, generic_ray.start_point);
	MVMul(temp_m, ray.direction, generic_ray.direction);
	switch (obj->object1->type)
	{
		case sphere:
			obj1_list = Sh_Comp_Sphere(generic_ray, obj->object1);
			break;
		case plane:
			obj1_list = Sh_Comp_Plane(generic_ray, obj->object1);
			break;
		case square:
			obj1_list = Sh_Comp_Square(generic_ray, obj->object1);
			break;
		case cone:
			obj1_list = Sh_Comp_Cone(generic_ray, obj->object1);
			break;
		case cylinder:
			obj1_list = Sh_Comp_Cylinder(generic_ray, obj->object1);
			break;
		case cube:
			obj1_list = Sh_Comp_Cube(generic_ray, obj->object1);
			break;
		case torus:
			obj1_list = Sh_Comp_Torus(generic_ray, obj->object1);
			break;
		case union_op:
		case intersection_op:
		case difference_op:
			obj1_list = Sh_Comp_Compound(generic_ray, obj->object1);
			break;
	}

	/* Object2. */
	temp_v = (obj->object2->inverse_trans).displacement;
	VAdd(ray.start_point, temp_v, temp_v);
	temp_m = (obj->object2->inverse_trans).matrix;
	MVMul(temp_m, temp_v, generic_ray.start_point);
	MVMul(temp_m, ray.direction, generic_ray.direction);
	switch (obj->object2->type)
	{
		case sphere:
			obj2_list = Sh_Comp_Sphere(generic_ray, obj->object2);
			break;
		case plane:
			obj2_list = Sh_Comp_Plane(generic_ray, obj->object2);
			break;
		case square:
			obj2_list = Sh_Comp_Square(generic_ray, obj->object2);
			break;
		case cone:
			obj2_list = Sh_Comp_Cone(generic_ray, obj->object2);
			break;
		case cylinder:
			obj2_list = Sh_Comp_Cylinder(generic_ray, obj->object2);
			break;
		case cube:
			obj2_list = Sh_Comp_Cube(generic_ray, obj->object2);
			break;
		case torus:
			obj2_list = Sh_Comp_Torus(generic_ray, obj->object2);
			break;
		case union_op:
		case intersection_op:
		case difference_op:
			obj2_list = Sh_Comp_Compound(generic_ray, obj->object2);
			break;
	}

	/* Combine the lists. */
	if (!obj1_list && !obj2_list)
	{
		/* Both lists empty, no hits, return FALSE. */
		return NULL;
	}
	if (!obj1_list)
	{
		if (obj->type == union_op)
			hit_list = obj2_list;
		else
		{
			sh_free_hit_list(obj2_list, temp_list1);
			hit_list = NULL;
		}
	}
	else if (!obj2_list)
	{
		if (obj->type == intersection_op)
		{
			sh_free_hit_list(obj1_list, temp_list1);
			hit_list = NULL;
		}
		else
			hit_list = obj1_list;
	}
	else
	{
		in_object = in_object1 = in_object2 = FALSE;
		part_list = hit_list = NULL;
		
		switch (obj->type)
		{
			/* First the non-cases (to stop gcc complaining) */
			case sphere:
			case cube:
			case plane:
			case square:
			case cone:
			case cylinder:
			case torus : break;

			case union_op:
				while (obj1_list && obj2_list)
				{
					if (obj1_list->t_hit < obj2_list->t_hit)
					{
						if (!in_object ||
							(in_object && in_object1 && !in_object2))
						{
							if (part_list)
							{
								part_list->next = obj1_list;
								part_list = part_list->next;
							}
							else
							{
								part_list = hit_list = obj1_list;
							}
							obj1_list = obj1_list->next;
						}
						else
						{
							temp_list = obj1_list;
							obj1_list = obj1_list->next;
							free(temp_list);
						}
						in_object1 = !in_object1;
						in_object = in_object1 || in_object2;
					}
					else
					{
						if (!in_object ||
							(in_object && in_object2 && !in_object1))
						{
							if (part_list)
							{
								part_list->next = obj2_list;
								part_list = part_list->next;
							}
							else
							{
								part_list = hit_list = obj2_list;
							}
							obj2_list = obj2_list->next;
						}
						else
						{
							temp_list = obj2_list;
							obj2_list = obj2_list->next;
							free(temp_list);
						}
						in_object2 = !in_object2;
						in_object = in_object1 || in_object2;
					}
				}
				/* Add the end bit */
				if (obj1_list)
					part_list->next = obj1_list;
				else
					part_list->next = obj2_list;
				break;

			case intersection_op:
				while (obj1_list && obj2_list)
				{
					if (obj1_list->t_hit < obj2_list->t_hit)
					{
						if (in_object || in_object2)
						{
							if (part_list)
							{
								part_list->next = obj1_list;
								part_list = part_list->next;
							}
							else
							{
								part_list = hit_list = obj1_list;
							}
							obj1_list = obj1_list->next;
						}
						else
						{
							temp_list = obj1_list;
							obj1_list = obj1_list->next;
							free(temp_list);
						}
						in_object1 = !in_object1;
						in_object = in_object1 && in_object2;
					}
					else
					{
						if (in_object || in_object1)
						{
							if (part_list)
							{
								part_list->next = obj2_list;
								part_list = part_list->next;
							}
							else
							{
								part_list = hit_list = obj2_list;
							}
							obj2_list = obj2_list->next;
						}
						else
						{
							temp_list = obj2_list;
							obj2_list = obj2_list->next;
							free(temp_list);
						}
						in_object2 = !in_object2;
						in_object = in_object1 && in_object2;
					}
				}

				/* Free the end bit */
				if (part_list)
					part_list->next = NULL;
				sh_free_hit_list(obj1_list, temp_list1);
				sh_free_hit_list(obj2_list, temp_list1);
				break;

			case difference_op:
				while (obj1_list && obj2_list)
				{
					if (obj1_list->t_hit < obj2_list->t_hit)
					{
						if (!in_object2)
						{
							if (part_list)
							{
								part_list->next = obj1_list;
								part_list = part_list->next;
							}
							else
							{
								part_list = hit_list = obj1_list;
							}
							obj1_list = obj1_list->next;
						}
						else
						{
							temp_list = obj1_list;
							obj1_list = obj1_list->next;
							free(temp_list);
						}
						in_object1 = !in_object1;
						in_object = in_object1 && !in_object2;
					}
					else
					{
						if (in_object || in_object1)
						{
							if (part_list)
							{
								part_list->next = obj2_list;
								part_list = part_list->next;
							}
							else
							{
								part_list = hit_list = obj2_list;
							}
							obj2_list = obj2_list->next;
						}
						else
						{
							temp_list = obj2_list;
							obj2_list = obj2_list->next;
							free(temp_list);
						}
						in_object2 = !in_object2;
						in_object = in_object1 && !in_object2;
					}
				}
				/* Add the end bit */
				if (obj1_list)
				{
					if (part_list)
						part_list->next = obj1_list;
					else
						part_list = hit_list = obj1_list;
				}
				else
				{
					if (part_list)
						part_list->next = NULL;
					/* Free list 2. */
					sh_free_hit_list(obj2_list, temp_list1);
				}
				/* Remove duplicate hits from the list.  These cause errors. */
				while (hit_list &&
					   dequal(hit_list->t_hit, hit_list->next->t_hit))
				{
					temp_list = hit_list->next->next;
					hit_list->next->next = NULL;
					sh_free_hit_list(hit_list, temp_list1);
					hit_list = temp_list;
				}
				if (!hit_list) break;
				part_list = hit_list;
				temp_list = part_list->next->next;
				while (temp_list)
				{
					if (dequal(temp_list->t_hit, temp_list->next->t_hit))
					{
						part_list->next->next = temp_list->next->next;
						temp_list->next->next = NULL;
						sh_free_hit_list(temp_list, temp_list1);
						temp_list = NULL;
					}
					else
						part_list = part_list->next->next;
					temp_list = part_list->next->next;
				}
				break;

		}
	}

	return hit_list;
}



HitList*
Sh_Comp_Torus(Ray ray, Instance* obj)
/********************************************************************/
/*																	*/
/*	Intersects ray with a generic torus and returns a list of hit	*/
/*	 points.														*/
/*	The quartic root solving code is adapted from Roots3And4.c,		*/
/*	 from Graphics Gems.											*/
/*																	*/
/********************************************************************/
{
	HitList	*result;
	double		AA, BB, CC;	/* Coefficients for sphere eqn.*/
	double		hit[4];
	Vector		unit_dir;
	double		temp1, temp2, temp3, temp4, temp5, temp6;
	double		line_scale;	/* The scaling factor for the ray direction. */
	double		coeffs[4];	/* coefficients for quartic eqn. */
	int			num = 0;
	int			i, j;

	/* Test for some special cases.								*/
	/* First, test for rays which don't intersect the top and	*/
	/*  bottom planes of the torus.  ie the plane z=1 and z=-1.	*/
	if ((ray.start_point.z > 1.0) && (ray.direction.z > 0.0))
		return NULL;

	if ((ray.start_point.z < -1.0) && (ray.direction.z < 0.0))
		return NULL;

	/* Test for rays which don't intersect the bounding sphere, r=d+1. */
	AA = VDot(ray.direction, ray.direction);
	BB = VDot(ray.start_point, ray.direction);
	CC = VDot(ray.start_point, ray.start_point) -
			(obj->torus_rad + 1.0) * (obj->torus_rad + 1.0);
	if ((BB * BB - AA * CC) < 0.0) return NULL;

	VUnit(ray.direction, line_scale, unit_dir);

	temp1 = obj->torus_rad * obj->torus_rad;
	temp2 = ray.start_point.z * ray.start_point.z;
	temp3 = unit_dir.z * unit_dir.z;
	temp4 = unit_dir.z * ray.start_point.z;
	temp5 = ray.start_point.x * ray.start_point.x + ray.start_point.y *
			ray.start_point.y + temp2 - temp1 - 1.0;
	temp6 = ray.start_point.x * unit_dir.x + ray.start_point.y *
			unit_dir.y + ray.start_point.z * unit_dir.z;

	coeffs[3] = 4.0 * temp6;
	coeffs[2] = 2.0 * (temp5 + 2.0 * temp6 * temp6 + 2.0 * temp1 * temp3);
	coeffs[1] = 4.0 * (temp6 * temp5 + 2.0 * temp1 * temp4);
	coeffs[0] = temp5 * temp5 + 4.0 * temp1 * (temp2 - 1.0);

	/* Solve the resulting quartic equation.				*/
	num = SolveQuartic(coeffs, hit);

	if (num == 0)
		return NULL;

	/* Sort the hits. Dead simple bubble sort. */
	for ( i = 0 ; i < num-1 ; i++ )
	{
		for ( j = i + 1 ; j < num ; j++ )
		{
			if (hit[i] > hit[j])
			{
				temp1 = hit[i];
				hit[i] = hit[j];
				hit[j] = temp1;
			}
		}
	}

	/* Allocate space. */
	if (num == 2)
	{
		result = get_hit_elmt();
		result->t_hit = hit[0] * line_scale;
		result->next = get_hit_elmt();
		result->next->t_hit = hit[1] * line_scale;
		result->next->next = NULL;
	}
	else
	{
		result = get_hit_elmt();
		result->t_hit = hit[0] * line_scale;
		result->next = get_hit_elmt();
		result->next->t_hit = hit[1] * line_scale;
		result->next->next = get_hit_elmt();
		result->next->next->t_hit = hit[2] * line_scale;
		result->next->next->next = get_hit_elmt();
		result->next->next->next->t_hit = hit[3] * line_scale;
		result->next->next->next->next = NULL;
	}

	return result;

}

