/*
**    Genray: A Generic Raytracer
**    Copyright (C) 1994  Stephen Chenney (stephen@cs.su.oz.au)
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program; if not, write to the Free Software
**    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/****************************************************************/
/*																*/
/*	File containing all the shape intersection functions		*/
/*	 for raytrace.												*/
/*	Written by Stephen Chenney.  Dec 1993.						*/
/*																*/
/****************************************************************/

#include "genray.h"


/****************************************************************/
/*																*/
/*	All the functions intersect a ray with something.			*/
/*	All return the first hit in t, and whether the ray was		*/
/*	 passing through air when it hit in in_air.  The normal at	*/
/*	 the point of contact is returned in n.  Note this normal	*/
/*	 depends on in_air.											*/
/*	If the object hit has a pix_map associated with it, the		*/
/*	 appropriate colour at the hit point is determined and put	*/
/*	 in obj->nature.colour.										*/
/*																*/
/****************************************************************/


char
Ray_Sphere(Ray ray, Instance* obj, double* t, Vector* n, char* in_air)
/****************************************************************/
/*																*/
/*	Intersects a ray with a sphere.								*/
/*	The hit time is returned in t.  The return value of the		*/
/*	 function is whether it hit at all.							*/
/*	Taken from Hill pg 627.										*/
/*																*/
/****************************************************************/
{
	double	AA, BB, CC;
	double	discrim, discrim_root;

	AA = VDot(ray.direction, ray.direction);
	BB = VDot(ray.start_point, ray.direction);
	CC = VDot(ray.start_point, ray.start_point) - 1;

	discrim = BB * BB - AA * CC;

	if (discrim < 0.0)
		return FALSE;
	
	discrim_root = sqrt(discrim);

	if (-BB > discrim_root + DELTA)
	{
		/* A hit on the way in. */
		(*t) = (-BB - discrim_root) / AA;
		n->x = ray.start_point.x + (*t) * ray.direction.x;
		n->y = ray.start_point.y + (*t) * ray.direction.y;
		n->z = ray.start_point.z + (*t) * ray.direction.z;
		*in_air = TRUE;
		return TRUE;
	}
	else if (BB < discrim_root - DELTA)
	{
		/* A hit on the way out.  The normal is reversed. */
		(*t) = (-BB + discrim_root) / AA;
		n->x = -(ray.start_point.x + (*t) * ray.direction.x);
		n->y = -(ray.start_point.y + (*t) * ray.direction.y);
		n->z = -(ray.start_point.z + (*t) * ray.direction.z);
		*in_air = FALSE;
		return TRUE;
	}
	return FALSE;

}


char
Ray_Plane(Ray ray, Instance* obj, double* t, Vector* n, char* in_air)
/****************************************************************/
/*																*/
/*	Intersects a ray with the generic plane.					*/
/*	The hit time is returned in t.  The return value of the		*/
/*	 function is whether it hit at all.							*/
/*																*/
/****************************************************************/
{
	double	x, y;

	if (ray.direction.z == 0.0) return FALSE;

	if (((*t) = -(ray.start_point.z) / ray.direction.z) > DELTA)
	{
		if (ray.direction.z > 0)
		{
			/* Ray is heading up from under plane. */
			VNew(0, 0, -1, (*n));
		}
		else
		{
			VNew(0, 0, 1, (*n));
		}
		*in_air = TRUE;

		if (obj->nature.pix_map != NULL)
		{
			x = ray.start_point.x + ray.direction.x * (*t);
	    	y = ray.start_point.y + ray.direction.y * (*t);
			obj->nature.colour =
				obj->nature.pix_map->intensity
				[(int)((y-(int)y+1.0-DELTA)*obj->nature.pix_map->height / 2.0)
				 * obj->nature.pix_map->width +
				 (int)((x-(int)x+1.0-DELTA)*obj->nature.pix_map->width / 2.0)];
		}

		return TRUE;
	}

	return FALSE;

}


char
Ray_Square(Ray ray, Instance* obj, double* t, Vector* n, char* in_air)
/****************************************************************/
/*																*/
/*	Intersects a ray with the generic square.					*/
/*	The hit time is returned in t.  The return value of the		*/
/*	 function is whether it hit at all.							*/
/*																*/
/****************************************************************/
{
	double	x, y;

	if (ray.direction.z == 0.0) return FALSE;

	if (((*t) = -(ray.start_point.z) / ray.direction.z) < DELTA) return FALSE;

	if ((fabs(x = ray.start_point.x + ray.direction.x * (*t)) <= 1.0) &&
	    (fabs(y = ray.start_point.y + ray.direction.y * (*t)) <= 1.0))
	{
		if (ray.direction.z > 0)
		{
			/* Ray is heading up from under square. */
			VNew(0, 0, -1, (*n));
		}
		else
		{
			VNew(0, 0, 1, (*n));
		}
		*in_air = TRUE;
		if (obj->nature.pix_map != NULL)
		{
			obj->nature.colour =
				obj->nature.pix_map->intensity
				[(int)((y + 1.0 - DELTA) * obj->nature.pix_map->height / 2.0) *
				 obj->nature.pix_map->width +
				 (int)((x + 1.0 - DELTA) * obj->nature.pix_map->width / 2.0)];
		}
		return TRUE;
	}

	return FALSE;
}


char
Ray_Cylinder(Ray ray, Instance* obj, double* t, Vector* n, char* in_air)
/****************************************************************/
/*																*/
/*	Intersects a ray with the generic cylinder.					*/
/*	The hit time is returned in t.  The return value of the		*/
/*	 function is whether it hit at all.							*/
/*	As per outline on Hill pg 631.								*/
/*																*/
/****************************************************************/
{
	double	hit1;	/* The hit on way in. */
	double	hit2;	/* The hit on the way out. */
	double	z1, z2;
	double	AA, BB;
	double	discrim, discrim_root;

	if ((AA = ray.direction.x * ray.direction.x +
		  ray.direction.y * ray.direction.y) != 0.0)
	{
		/* There may be hits with the walls. */
		BB = ray.start_point.x * ray.direction.x +
		     ray.start_point.y * ray.direction.y;
		if ((discrim = BB * BB - AA * (ray.start_point.x * ray.start_point.x +
				ray.start_point.y * ray.start_point.y - 1)) >= 0.0)
		{
			/* There is a hit. */
			discrim_root = sqrt(discrim);

			hit1 = (-BB - discrim_root) / AA;
			z1 = ray.start_point.z + hit1 * ray.direction.z;

			if ((hit1 > DELTA) && (z1 <= 1.0) && (z1 >= -1.0))
			{
				/* hit1 is valid */
				/* Convince yourself that this is the best hit. */
				(*t) = hit1;
				n->x = ray.start_point.x + ray.direction.x * (*t);
				n->y = ray.start_point.y + ray.direction.y * (*t);
				n->z = 0;
				*in_air = TRUE;
				return TRUE;
			}

			/* Try hit2. */
			hit2 = (-BB + discrim_root) / AA;

			if (hit2 > DELTA)
			{
				z2 = ray.start_point.z + hit2 * ray.direction.z;
				if (((z1 > 1.0) && (z2 > 1.0)) ||
					((z2 < -1.0) && (z1 < -1.0)))
				{
					/* Ray entirely misses cylinder. */
					return FALSE;
				}
				else if ((z2 <= 1.0) && (z2 >= -1.0))
				{
					/* hit2 is valid. */
					/* Test whether ray will hit either end. */
					if (z1 < -1.0)
					{
						/* Ray must hit bottom. */
						hit1 = (-1.0 - ray.start_point.z) / ray.direction.z;
						if (hit1 < DELTA)
						{
							/* hit2 is the one. */
							(*t) = hit2;
							n->x = -(ray.start_point.x + ray.direction.x * hit2);
							n->y = -(ray.start_point.y + ray.direction.y * hit2);
							n->z = 0;
							*in_air = FALSE;
							return TRUE;
						}
						else
						{
							(*t) = hit1;
							n->x = n->y = 0;
							n->z = -1;
							*in_air = TRUE;
							return TRUE;
						}
						
					}
					else if (z1 > 1.0)
					{
						/* Ray must hit top. */
						hit1 = (1 - ray.start_point.z) / ray.direction.z;
						if (hit1 < DELTA)
						{
							/* hit2 is the one. */
							(*t) = hit2;
							n->x = -(ray.start_point.x + ray.direction.x * hit2);
							n->y = -(ray.start_point.y + ray.direction.y * hit2);
							n->z = 0;
							*in_air = FALSE;
							return TRUE;
						}
						else
						{
							(*t) = hit1;
							n->x = n->y = 0;
							n->z = 1;
							*in_air = TRUE;
							return TRUE;
						}
					}
					else
					{
						/* hit2 is the one. */
						(*t) = hit2;
						n->x = -(ray.start_point.x + ray.direction.x * hit2);
						n->y = -(ray.start_point.y + ray.direction.y * hit2);
						n->z = 0;
						*in_air = FALSE;
						return TRUE;
					}

				}
				else
				{
					/* Need to test ends depending on z1. */
					if (z1 < -1)
					{
						/* Ray coming from bottom up. */
						hit2 = (1.0 - ray.start_point.z) / ray.direction.z;
						if (hit2 < DELTA) return FALSE;
						hit1 = (-1.0 - ray.start_point.z) / ray.direction.z;
						if (hit1 < DELTA)
						{
							/* hit2 is it. */
							(*t) = hit2;
							n->x = n->y = 0;
							n->z = 1.0;
							*in_air = FALSE;
							return TRUE;
						}
						else
						{
							/* hit1 is used. */
							(*t) = hit1;
							n->x = n->y = 0;
							n->z = -1.0;
							*in_air = TRUE;
							return TRUE;
						}
					}
					else
					{
						/* Ray coming from top down. */
						hit2 = (-1.0 - ray.start_point.z) / ray.direction.z;
						if (hit2 < DELTA)
							return FALSE;
						hit1 = (1.0 - ray.start_point.z) / ray.direction.z;
						if (hit1 < DELTA)
						{
							/* hit2 is it. */
							(*t) = hit2;
							n->x = n->y = 0;
							n->z = -1.0;
							*in_air = FALSE;
							return TRUE;
						}
						else
						{
							/* hit1 is used. */
							(*t) = hit1;
							n->x = n->y = 0;
							n->z = 1.0;
							*in_air = TRUE;
							return TRUE;
						}
					}
				}
			}
			else
				return FALSE;	/* If hit2 is behind eye, whole thing is. */
		}
		else return FALSE;		/* Ray misses entirely. */
	}
	else
	{
		/* Ray is parallel to sides. */
		/* Test top and bottom. */
		if (ray.direction.z > 0)
		{
			/* ray is heading up. */
			/* Get intersection with top plane. */
			hit2 = (1 - ray.start_point.z) / ray.direction.z;

			/* Test for behind eye. */
			if (hit2 < DELTA) return FALSE;

			/* Test for inside end circle. */
			AA = ray.start_point.x + hit2 * ray.direction.x;
			BB = ray.start_point.y + hit2 * ray.direction.y;
			if ((AA*AA + BB*BB) > 1.0) return FALSE;

			/* Make a decision. */
			hit1 = (-1.0 - ray.start_point.z) / ray.direction.z;
			if (hit1 < DELTA)
			{
				/* hit2 is it. */
				(*t) = hit2;
				n->x = n->y = 0;
				n->z = 1.0;
				*in_air = FALSE;
				return TRUE;
			}
			else
			{
				/* hit1 is used. */
				(*t) = hit1;
				n->x = n->y = 0;
				n->z = -1.0;
				*in_air = TRUE;
				return TRUE;
			}
		}
		else
		{
			/* ray is going down. */
			/* Get intersection with bottom plane. */
			hit2 = (-1.0 - ray.start_point.z) / ray.direction.z;

			/* Test for behind eye. */
			if (hit2 < DELTA) return FALSE;

			/* Test for inside end circle. */
			AA = ray.start_point.x + hit2 * ray.direction.x;
			BB = ray.start_point.y + hit2 * ray.direction.y;
			if ((AA*AA + BB*BB) > 1.0) return FALSE;

			/* Make a decision. */
			hit1 = (1.0 - ray.start_point.z) / ray.direction.z;
			if (hit1 < DELTA)
			{
				/* hit2 is it. */
				(*t) = hit2;
				n->x = n->y = 0;
				n->z = -1.0;
				*in_air = FALSE;
				return TRUE;
			}
			else
			{
				/* hit1 is used. */
				(*t) = hit1;
				n->x = n->y = 0;
				n->z = 1.0;
				*in_air = TRUE;
				return TRUE;
			}
		}
	}
}


char
Ray_Cone(Ray ray, Instance* obj, double* t, Vector* n, char* in_air)
/****************************************************************/
/*																*/
/*	Intersects a ray with the generic cone.						*/
/*	The hit time is returned in t.  The return value of the		*/
/*	 function is whether it hit at all.							*/
/*	As per outline on Hill pg 631-2.							*/
/*	The code is entirely non obvious.  I arrived at it by		*/
/*	 drawing a pretty picture and examining cases. There are 9	*/
/*	 or so.														*/
/*																*/
/****************************************************************/
{
	double	hit1;	/* The hit on way in. */
	double	hit2;	/* The hit on the way out. */
	double	z1, z2;
	double	AA, BB;
	double	temp_d;
	double	discrim, discrim_root;

	/* AA = 0 I think implies the ray lies in one of the walls of the cone. */
	if ((AA = ray.direction.x * ray.direction.x +
		  ray.direction.y * ray.direction.y -
		  ray.direction.z * ray.direction.z / 4) != 0.0)
	{
		/* There may be hits with the walls. */
		BB = ray.start_point.x * ray.direction.x +
		     ray.start_point.y * ray.direction.y +
		     ray.direction.z * ( 1 - ray.start_point.z) / 4;
		temp_d = 1 - ray.start_point.z;
		if ((discrim = BB * BB - AA * (ray.start_point.x * ray.start_point.x +
			ray.start_point.y * ray.start_point.y - temp_d*temp_d / 4)) >= 0.0)
		{
			/* There is a hit. */
			discrim_root = sqrt(discrim);

			if (AA > 0)
				hit1 = (-BB - discrim_root) / AA;
			else
				hit1 = (-BB + discrim_root) / AA;
			z1 = ray.start_point.z + hit1 * ray.direction.z;

			if ((hit1 > DELTA) && (z1 <= 1.0) && (z1 >= -1.0))
			{
				/* hit1 is valid */
				/* Possibly need to test against the base. */
				if (ray.direction.z > 0)
				{
					/* May be a valid hit on the bottom. */
					hit2 = (-1.0 - ray.start_point.z) / ray.direction.z;
					AA = ray.start_point.x + ray.direction.x * hit2;
					BB = ray.start_point.y + ray.direction.y * hit2;
					AA = AA*AA + BB*BB;

					if ((hit2 < DELTA) || (AA > 1.0))
					{
						/* Doesn't intersect base. */
						(*t) = hit1;
						n->x = ray.start_point.x + ray.direction.x * hit1;
						n->y = ray.start_point.y + ray.direction.y * hit1;
						n->z = 0.125 * (1 - z1) * (1- z1);
						*in_air = (AA > 1.0);
						return TRUE;
					}
					else
					{
						/* Base hit is it. */
						(*t) = hit2;
						n->x = n->y = 0;
						n->z = -1.0;
						*in_air = TRUE;
						return TRUE;
					}
				}
				else
				{
					/* Convince yourself that this is the best hit. */
					(*t) = hit1;
					n->x = ray.start_point.x + ray.direction.x * (*t);
					n->y = ray.start_point.y + ray.direction.y * (*t);
					n->z = 0.125 * (1 - z1) * (1- z1);
					*in_air = TRUE;
					return TRUE;
				}
			}

			/* Try hit2. */
			if (AA > 0)
				hit2 = (-BB + discrim_root) / AA;
			else
				hit2 = (-BB - discrim_root) / AA;

			if (hit2 > DELTA)
			{
				z2 = ray.start_point.z + hit2 * ray.direction.z;
				if ((z2 <= 1.0) && (z2 >= -1.0))
				{
					/* hit2 is valid. */
					/* Test whether ray will hit the end. */
					if (z1 < -1.0)
					{
						/* Ray must hit bottom. */
						hit1 = (-1.0 - ray.start_point.z) / ray.direction.z;
						if (hit1 < DELTA)
						{
							/* hit2 is the one. */
							(*t) = hit2;
							n->x = -(ray.start_point.x + ray.direction.x * hit2);
							n->y = -(ray.start_point.y + ray.direction.y * hit2);
							n->z = 0.125 * (1 - z2) * (1- z2);
							*in_air = FALSE;
							return TRUE;
						}
						else
						{
							(*t) = hit1;
							n->x = n->y = 0;
							n->z = -1;
							*in_air = TRUE;
							return TRUE;
						}
						
					}
					else if (z1 > 1.0)
					{
						/* hit2 must be the best hit.  Convince yourself. */
						(*t) = hit2;
						n->x = ray.start_point.x + ray.direction.x * hit2;
						n->y = ray.start_point.y + ray.direction.y * hit2;
						n->z = 0.125 * (1 - z2) * (1- z2);
						*in_air = TRUE;
						return TRUE;
					}
					else
					{
						/* hit2 is the one. */
						(*t) = hit2;
						n->x = -(ray.start_point.x + ray.direction.x * hit2);
						n->y = -(ray.start_point.y + ray.direction.y * hit2);
						n->z = -0.125 * (1 - z2) * (1- z2);
						*in_air = FALSE;
						return TRUE;
					}

				}
				else return FALSE;
				
			}
			else if (z1 > 1.0)
			{
				/* May be a valid hit on the bottom. */
				hit2 = (-1.0 - ray.start_point.z) / ray.direction.z;
				if (hit2 < DELTA) return FALSE;
				(*t) = hit2;
				n->x = n->y = 0;
				n->z = 1.0;
				*in_air = FALSE;
				return TRUE;
			}
			else return FALSE;
		}
		else return FALSE;
	}
	else return FALSE;
}



char
Ray_Cube(Ray ray, Instance* obj, double* t, Vector* n, char* in_air)
/****************************************************************/
/*																*/
/*	Intersects a ray with the generic cube.						*/
/*	The hit time is returned in t.  The return value of the		*/
/*	 function is whether it hit at all.							*/
/*	As per outline on Hill pg 632-5.							*/
/*																*/
/****************************************************************/
{
	double	tin = -HUGE;
	double	tout = HUGE;
	double	thit;
	int	fin = 0;
	int	fout = 0;

	/* Since there are so many 0's in all this, I'm going	*/
	/*  to do it linearly.  This makes for poor readability	*/
	/*  and lots of code, but much greater speed.		*/

	/* x = 1 plane: n = {1,0,0}, D = 1 */
	if (ray.direction.x == 0.0)	/* Ray parallel.	*/
	{
		if (1 - ray.start_point.x < 0.0) return FALSE;	/* Ray outside */
		/* Check x = -1 plane while we're at it. */
		if (1 + ray.start_point.x < 0.0) return FALSE;
	}
	else
	{
		thit = (1 - ray.start_point.x) / ray.direction.x;
		if (ray.direction.x > 0.0)
		{
			tout = thit;
			fout = 0;
		}
		else
		{
			tin = thit;
			fin = 0;
		}
	}
	/* Not possible to eliminate ray yet */

	/* y = 1 plane: n = {0,1,0}, D = 1 */
	if (ray.direction.y == 0.0)     /* Ray parallel.        */
	{
		if (1 - ray.start_point.y < 0.0) return FALSE;  /* Ray outside */
		/* Check y = -1 plane too. */
		if (1 + ray.start_point.y < 0.0) return FALSE;
	}
	else
	{
		thit = (1 - ray.start_point.y) / ray.direction.y;
		if (ray.direction.y > 0.0)
		{
			if (thit < tout)
			{
				fout = 1;
				tout = thit;
			}
		}
		else
		{
			if (thit > tin)
			{
				fin = 1;
				tin = thit;
			}
		}
	}
	/* Try to eliminate intersection. */
	if (tin > tout) return FALSE;

	/* z = 1 plane: n = {0,0,1}, D = 1 */
	if (ray.direction.z == 0.0)     /* Ray parallel.        */
	{
		if (1 - ray.start_point.z < 0.0) return FALSE;  /* Ray outside */
		/* Check y = -1 plane too. */
		if (1 + ray.start_point.z < 0.0) return FALSE;
	}
	else
	{
		thit = (1 - ray.start_point.z) / ray.direction.z;
		if (ray.direction.z > 0.0)
		{
			if (thit < tout)
			{
				fout = 2;
				tout = thit;
			}
		}
		else
		{
			if (thit > tin)
			{
				fin = 2;
				tin = thit;
			}
		}
	}
	/* Try to eliminate intersection. */
	if (tin > tout) return FALSE;


	/* x = -1 plane : n = {-1,0,0}, D = 1 */
	/* Already eliminated parallel rays above. */
	thit = (1 + ray.start_point.x) / -ray.direction.x;
	if (ray.direction.x <  0.0)
	{
		if (thit < tout)
		{
			fout = 3;
			tout = thit;
		}
	}
	else
	{
		if (thit > tin)
		{
			fin = 3;
			tin = thit;
		}
	}
	if (tin > tout) return FALSE;

	/* y = -1 plane : n = {0,-1,0}, D = 1 */
	/* Already eliminated parallel rays above. */
	thit = (1 + ray.start_point.y) / -ray.direction.y;
	if (ray.direction.y <  0.0)
	{
		if (thit < tout)
		{
			fout = 4;
			tout = thit;
		}
	}
	else
	{
		if (thit > tin)
		{
			fin = 4;
			tin = thit;
		}
	}
	if (tin > tout) return FALSE;

	/* z = -1 plane : n = {0,0,-1}, D = 1 */
	/* Already eliminated parallel rays above. */
	thit = (1 + ray.start_point.z) / -ray.direction.z;
	if (ray.direction.z <  0.0)
	{
		if (thit < tout)
		{
			fout = 5;
			tout = thit;
		}
	}
	else
	{
		if (thit > tin)
		{
			fin = 5;
			tin = thit;
		}
	}
	if (tin > tout) return FALSE;

	/* Here if there is a valid intersection */
	if (tin > DELTA)
	{
		(*t) = tin;
		switch (fin)
		{
			case 0 : VNew(1, 0, 0, (*n));
				break;
			case 1 : VNew(0, 1, 0, (*n));
				break;
			case 2 : VNew(0, 0, 1, (*n));
				break;
			case 3 : VNew(-1, 0, 0, (*n));
				break;
			case 4 : VNew(0, -1, 0, (*n));
				break;
			case 5 : VNew(0, 0, -1, (*n));

		}
		*in_air = TRUE;
		return TRUE;
	}
	if (tout > DELTA)
	{
		(*t) = tout;
		switch (fout)
		{
			case 0 : VNew(-1, 0, 0, (*n));
				break;
			case 1 : VNew(0, -1, 0, (*n));
				break;
			case 2 : VNew(0, 0, -1, (*n));
				break;
			case 3 : VNew(1, 0, 0, (*n));
				break;
			case 4 : VNew(0, 1, 0, (*n));
				break;
			case 5 : VNew(0, 0, 1, (*n));
		}
		*in_air = FALSE;
		return TRUE;
	}

	return FALSE;

}


char
Ray_Torus(Ray ray, Instance* obj, double* t, Vector* n, char* in_air)
/****************************************************************/
/*																*/
/*	Intersects the ray with a torus.							*/
/*	Equation of a torus slightly modified from Hill pg 636.		*/
/*	Generic torus has mean radius obj->torus_rad, tube radius 1.*/
/*	Quartic eqn solving code courtesy of Graphics Gems.			*/
/*																*/
/****************************************************************/
{
	double		AA, BB, CC;	/* Coefficients for sphere eqn.*/
	double		hit[4];
	double		temp1, temp2, temp3, temp4, temp5, temp6;
	Vector		unit_dir;
	double		line_scale;	/* The scaling factor for the ray direction. */
	double		coeffs[4];	/* coefficients for quartic eqn. */
	Vector		hit_pt;
	int			num = 0;
	int			i, j;

	/* Test for some special cases.								*/
	/* First, test for rays which don't intersect the top and	*/
	/*  bottom planes of the torus.  ie the plane z=1 and z=-1.	*/
	if ((ray.start_point.z > 1.0) && (ray.direction.z > 0.0))
		return FALSE;

	if ((ray.start_point.z < -1.0) && (ray.direction.z < 0.0))
		return FALSE;

	/* Test for rays which don't intersect the bounding sphere, r=d+1. */
	AA = VDot(ray.direction, ray.direction);
	BB = VDot(ray.start_point, ray.direction);
	CC = VDot(ray.start_point, ray.start_point) -
			(obj->torus_rad + 1.0) * (obj->torus_rad + 1.0);
	if ((BB * BB - AA * CC) < 0.0) return FALSE;

	VUnit(ray.direction, line_scale, unit_dir);

	temp1 = obj->torus_rad * obj->torus_rad;
	temp2 = ray.start_point.z * ray.start_point.z;
	temp3 = unit_dir.z * unit_dir.z;
	temp4 = unit_dir.z * ray.start_point.z;
	temp5 = ray.start_point.x * ray.start_point.x + ray.start_point.y *
			ray.start_point.y + temp2 - temp1 - 1.0;
	temp6 = ray.start_point.x * unit_dir.x + ray.start_point.y *
			unit_dir.y + ray.start_point.z * unit_dir.z;

	coeffs[3] = 4.0 * temp6;
	coeffs[2] = 2.0 * (temp5 + 2.0 * temp6 * temp6 + 2.0 * temp1 * temp3);
	coeffs[1] = 4.0 * (temp6 * temp5 + 2.0 * temp1 * temp4);
	coeffs[0] = temp5 * temp5 + 4.0 * temp1 * (temp2 - 1.0);

#ifdef TORUS_DEBUG
	for ( i = 0 ; i < 5 ; i++ )
		printf("coeff[%d] = %f : ", i, coeffs[i]);
	printf("\n");
#endif

	/* Solve the resulting quartic equation.				*/
	num = SolveQuartic(coeffs, hit);

	if (num == 0)
		return FALSE;

	/* Sort the hits. */
	for ( i = 0 ; i < num-1 ; i++ )
	{
		for ( j = i + 1 ; j < num ; j++ )
		{
			if (hit[i] > hit[j])
			{
				temp1 = hit[i];
				hit[i] = hit[j];
				hit[j] = temp1;
			}
		}
	}

#ifdef TORUS_DEBUG
	for ( i = 0 ; i < num ; i++ )
		printf("hit[%d] = %f : ", i, hit[i]);
	printf("\n");
#endif


	/* Return the first valid hit. */
	for ( i = 0 ; i < num ; i++)
	{
		if (hit[i] > DELTA)
		{
			*t = hit[i] * line_scale;
			*in_air = (i + 1) % 2;

			/* Work out the normal. */
			hit_pt.x = ray.start_point.x + hit[i] * unit_dir.x;
			hit_pt.y = ray.start_point.y + hit[i] * unit_dir.y;
			hit_pt.z = ray.start_point.z + hit[i] * unit_dir.z;

			temp1 = sqrt(hit_pt.x * hit_pt.x + hit_pt.y * hit_pt.y);
			temp2 = obj->torus_rad * hit_pt.x / temp1;
			temp3 = obj->torus_rad * hit_pt.y / temp1;
			if (*in_air)
			{
				n->x = hit_pt.x - temp2;
				n->y = hit_pt.y - temp3;
				n->z = hit_pt.z;
			}
			else
			{
				n->x = temp2 - hit_pt.x;
				n->y = temp3 - hit_pt.y;
				n->z = -hit_pt.z;
			}


			return TRUE;
		}
	}

	return FALSE;

}

